/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the classe CalculateCI used to calculate the confidence intervals of
 * an Confidence Interval session. Its is used in the J-OWAMP web interface.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 1.2
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */

import java.sql.*;  // import needed for JDBC access
import java.net.*;
import java.io.*;
import java.util.Calendar;


public class CalculateCI extends OWAMP_Base{

  // Class Variables

  static final int StreamPackets_Unauthenticated_PACKET_LENGTH = 14;


  // Constructors

  /**
   * This class contructor.
   *
   * @param serverAddress String - represents the Server address.
   * The server address should be in the following format:
   *
   *     IPv4_address:TCPportNumber    // for IPv4 address
   *     [IPv6_address]:TCPportNumber  // for IPv6 address
   *
   * The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
   * Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368)
   * the server address can be in the following format:
   *
   *     IPv4_address    // for IPv4 address
   *     [IPv6_address]  // for IPv6 address
   *
   * @param sqlServerAddress String
   * @param dbJOWAMP_CI String
   * @param userName String
   * @param passWord String
   * @param CIconfigTable String
   * @param CIsidTable String
   * @param CI_id short
   * @param CI_ResultTable String
   * @param SS_ResultTable String
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the time.
   */

  public CalculateCI(String serverAddress, String sqlServerAddress, String dbJOWAMP_CI, String userName,
                     String passWord, String CIconfigTable, String CIsidTable, short CI_id, String CI_ResultTable,
                     String SS_ResultTable, HiResTimer hiResTimer){

    String address =  serverAddress.trim();
    String addressAux = address;
    serverTCPportNumber = DEFAULT_SEVER_PORT;  // TCPportNumber to use in OWAMP-Control with Server.
    boolean IPv6 = false;
    int i=0;

    try{

      if(address.indexOf("[") != -1){  // IPv6 address
        IPv6 = true;
        if(address.indexOf("]:") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
          serverTCPportNumber = Integer.parseInt(address.substring(address.indexOf("]:")+2));
        }
        address = address.substring(address.indexOf("[")+1,address.indexOf("]"));
      }else{  // IPv4 address
        if(address.indexOf(":") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
          serverTCPportNumber = Integer.parseInt(address.substring(address.indexOf(":")+1));
          address = address.substring(0,address.indexOf(":"));
        }
      }

      if (serverTCPportNumber < 0 || serverTCPportNumber >= 65536) {
        System.err.println("Error: Sender TCP port number out of range (" +serverTCPportNumber+").");
        System.err.println();
        System.exit(1);
      }

      InetAddress inetAddress[] = InetAddress.getAllByName(address);     // The array of all the IP addresses for a given host name.

      for(i=0;i<inetAddress.length;i++){
        if(IPv6 && (inetAddress[i].getAddress().length == 16)){  // IPv6 address
          this.serverAddress = inetAddress[i];
          break;
        }

        if(!IPv6 && (inetAddress[i].getAddress().length == 4)){  // IPv4 address
          this.serverAddress = inetAddress[i];
          break;
        }

        if(i == (inetAddress.length-1)){  // last cicle.
          System.err.println("Error: There isn't a valid IPv"+(IPv6?6:4)+" address for the server address introduced ("+address+").");
          System.err.println();
          System.exit(1);
        }
      }

    }catch(NumberFormatException e){
      System.err.println("Error: The the server address read is not a valid address ("+addressAux+").");
      System.err.println("The server address should be in the following format:\n\n"+
                         "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                         "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                         "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                         "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n"+
                         "the server address can be in the following format:\n\n"+
                         "    IPv4_address    // for IPv4 address\n"+
                         "    [IPv6_address]  // for IPv6 address\n");
      System.err.println();
      System.exit(1);
    }catch(Exception e2){
      System.err.println("Error: The the server address read is not a valid address ("+addressAux+").");
      System.err.println("The server address should be in the following format:\n\n"+
                         "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                         "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                         "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                         "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n"+
                         "the server address can be in the following format:\n\n"+
                         "    IPv4_address    // for IPv4 address\n"+
                         "    [IPv6_address]  // for IPv6 address\n");
      System.err.println();
      System.exit(1);
    }

    this.sqlServerAddress = sqlServerAddress;
    this.dbJOWAMP_CI = dbJOWAMP_CI;      // Database with the CIconfigTable and CIsidTable, and where all results will
                                         // be saved.
    this.userName = userName;
    this.passWord = passWord;
    this.CIconfigTable = CIconfigTable;  // table with the CI session's configuration.
    this.CIsidTable = CIsidTable;        // table with the session's SID, senderPort, receiverPort,
                                         // real start time and total test duration for each of the individual
                                         // single session of each interval.
    this.CI_id = CI_id;                  // CI session's id in table CIconfigTable.
    this.CI_ResultTable = CI_ResultTable; // Table where all CI results will be saved.
    this.SS_ResultTable = SS_ResultTable; // Table where the statistic results of all Single Sessions, of each interval, will be saved.
    this.hiResTimer = hiResTimer;
  }


  // Instance variables

  private InetAddress serverAddress;             // The Server's address.
  private int serverTCPportNumber;               // The Server's port for TCP connection.
  private String sqlServerAddress, dbJOWAMP_CI, CIconfigTable, CIsidTable, userName, passWord,
                 CI_ResultTable, SS_ResultTable;
  private short CI_id = 0;   // CI session's id in table CIconfigTable.


  // Instance methods

  public boolean processCI(){ // process confidence interval

    // Establish a connection to the data base.

    MySQLcom MySQLclient= null;
    Connection con = null;
    try{
      MySQLclient = new MySQLcom (sqlServerAddress, dbJOWAMP_CI, userName, passWord);
      con = MySQLclient.getConnection();
    }catch(Exception e){
      System.out.println(e);
      return false;
    }

    // Read the numberOfPackets, paddingLength, startYear, startMonth, startDate, startHour, startMinute, startSecond, timeOut,
    // timeBetweenSetOfTests, timeBetweenTests, numIntervals and numOfTestsInInterval from database dbJOWAMP_CI
    // in table CIconfigTable where id = CI_id.

    ResultSet rs = null;
    int numberOfPackets = 0, paddingLength = 0, startYear = 0, startMonth = 0, startDate = 0,
        startHour = 0, startMinute = 0, startSecond = 0, numIntervals = 0, numOfTestsInInterval = 0,
        timeBetweenSetOfTests = 0, timeBetweenTests = 0;
    double timeOut = 0, slotParameter = 0;

    try{
      rs = MySQLclient.executeQuery(con,
                                    "select numberOfPackets, paddingLength, startYear, startMonth, startDate, startHour, startMinute, startSecond, timeOut, slotParameter, timeBetweenSetOfTests, timeBetweenTests, numIntervals, numOfTestsInInterval from `" +
                                    CIconfigTable +
                                    "` where id = "+CI_id);
      rs.next();
      numberOfPackets = rs.getInt("numberOfPackets");
      paddingLength = rs.getInt("paddingLength");
      startYear = rs.getInt("startYear");
      startMonth = rs.getInt("startMonth");
      startDate = rs.getInt("startDate");
      startHour = rs.getInt("startHour");
      startMinute = rs.getInt("startMinute");
      startSecond = rs.getInt("startSecond");
      timeOut = rs.getDouble("timeOut")*1000;    // in milliseconds
      slotParameter = rs.getDouble("slotParameter");    // in seconds
      timeBetweenSetOfTests = (int)rs.getDouble("timeBetweenSetOfTests")*60*1000;  // in milliseconds
      timeBetweenTests = (int)rs.getDouble("timeBetweenTests")*1000; // in milliseconds
      numIntervals = rs.getInt("numIntervals");
      numOfTestsInInterval = rs.getInt("numOfTestsInInterval");
      rs.close();
    }catch(Exception e){
      System.out.println(e);
      MySQLclient.close(con);
      return false;
    }

    // Verify if the CI_ResultTable exists. If not it should be created. If yes, another CalculateCI was made
    // so we must exit because a CalculateCI calculates all the CI of a session.

    try{
      MySQLclient.executeUpdate ( con, " CREATE TABLE `"+CI_ResultTable+"` (id int not null AUTO_INCREMENT, averageDelay DOUBLE, varDelay DOUBLE, minIntervalDelay DOUBLE, maxIntervalDelay DOUBLE, averageTimeoutLosses DOUBLE, varTimeoutLosses DOUBLE, minIntervalTimeoutLosses DOUBLE, maxIntervalTimeoutLosses DOUBLE, averageNotReceivedLosses DOUBLE, varNotReceivedLosses DOUBLE, minIntervalNotReceivedLosses DOUBLE, maxIntervalNotReceivedLosses DOUBLE,  averageNumDuplicates DOUBLE, varNumDuplicates DOUBLE, minIntervalNumDuplicates DOUBLE, maxIntervalNumDuplicates DOUBLE, averageThroughput DOUBLE, varThroughput DOUBLE, minIntervalThroughput DOUBLE, maxIntervalThroughput DOUBLE, PRIMARY KEY (id))");
    }catch(Exception e){  // If the table exists, an exception is thrown.
      try{ // Verify if the exception occurred because the table already exits in the database. If not, an error ocurred
           // and this method should return false.
        rs = MySQLclient.executeQuery(con, "SELECT COUNT(*) FROM `"+CI_ResultTable+"`");
        rs.close();
        MySQLclient.close(con);
        return true;
      }catch(Exception e2){
        System.out.println(e);
        MySQLclient.close(con);
        return false;
      }
    }

    // The table SS_ResultTable, where the statistic results of all Single Sessions will be saved, should be created.
    try{
      MySQLclient.executeUpdate ( con, " CREATE TABLE `"+SS_ResultTable+"` (id INT NOT NULL AUTO_INCREMENT, SID VARCHAR(32), numOfPacketsSent INT, numPacketsReceived INT, minDelay DOUBLE, averageDelay DOUBLE, maxDelay DOUBLE, varDelay DOUBLE, numTimeoutLosses DOUBLE, numNotReceivedLosses DOUBLE, numDuplicates DOUBLE, throughput DOUBLE, PRIMARY KEY (id))");
    }catch (Exception e) {
      // If an error ocurres we should drop the table CI_ResultTable.
      try {
        MySQLclient.executeUpdate(con, "DROP TABLE `"+CI_ResultTable+"`");
      }catch (Exception e2) {
        System.err.println(e2);
        MySQLclient.close(con);
        return false;
      }
      System.err.println(e);
      MySQLclient.close(con);
      return false;
    }


    // Deternine the CI.

    int session = 0, i=0, c=0, j=0, count=0, testPacketLength = 0, ID = -1, numPacketsReceived;
    String SID = "";
    long startTime = 0, realStartTime = 0, intervalStartTime = 0, testStartTime = 0;
    double totalTimeToWait = 0;

    double averageDelay = 0, varDelay = 0, minDelay = 0, maxDelay = 0, minIntervalDelay = 0, maxIntervalDelay = 0,
          averageTimeoutLosses = 0, varTimeoutLosses  = 0, minIntervalTimeoutLosses  = 0, maxIntervalTimeoutLosses  = 0,
          averageNotReceivedLosses  = 0, varNotReceivedLosses  = 0, minIntervalNotReceivedLosses  = 0, maxIntervalNotReceivedLosses  = 0,
          averageNumDuplicates  = 0, varNumDuplicates  = 0, minIntervalNumDuplicates  = 0, maxIntervalNumDuplicates  = 0,
          averageThroughput  = 0, varThroughput  = 0, minIntervalThroughput  = 0, maxIntervalThroughput  = 0,
          beginTime = 0, endTime = 0,
          alfaVal = (double)1.6449;  // for 90% interval
    double[] delay = new double[numOfTestsInInterval];
    double[] numTimeoutLosses = new double[numOfTestsInInterval];
    double[] numNotReceivedLosses = new double[numOfTestsInInterval];
    double[] numDuplicates = new double[numOfTestsInInterval];
    double[] throughput = new double[numOfTestsInInterval];
    boolean fetchSessionFinished = true;
    double duplicates;
    int[] sequenceNumberArray = null;
    boolean testError = false;

    try {
      Calendar sessionStartTime = Calendar.getInstance();
      sessionStartTime.set(startYear, startMonth - 1, startDate, startHour,
                           startMinute, startSecond);
      sessionStartTime.set(sessionStartTime.MILLISECOND, 0);
      startTime = sessionStartTime.getTimeInMillis();

      OWAMP_FetchClient fetch = new OWAMP_FetchClient(serverAddress, serverTCPportNumber,
                new Socket(), new BufferedReader(new InputStreamReader(System.in)),hiResTimer); // Used to process the FetchSessions

      for (int interval = 0; interval < numIntervals; interval++) { // for each interval
        intervalStartTime = startTime + interval * timeBetweenSetOfTests; // Test start time for this interval in milliseconds
        for (int num = 0; num < numOfTestsInInterval; num++, session++) { // for each test in this interval

          testStartTime = intervalStartTime + num * (timeBetweenTests); // StartTime for this session in milliseconds.

          // Wait for the startTime before processing this session because its configuration information will only
          // be in the database after its startTime. Note: If The startTime is in the past and separated from the
          // present by more than Timeout, the session will not be started. So we should compare the current time
          // with (startTime + timeout) to guarantee that the configuration information of this session will be
          // in the database.

          if ( (testStartTime + Math.round(timeOut)) > Math.round(hiResTimer.milliTime()))
            try {
              Thread.sleep( testStartTime + Math.round(timeOut - hiResTimer.milliTime()));
            }
            catch (Exception e) {}

          rs = MySQLclient.executeQuery(con,
                                        " SELECT SID, realStartTime, totalTimeToWait FROM `" +
                                        CIsidTable +
                                        "` WHERE id = "+(session+1));
          if (!rs.next()) {
            System.err.println(
                "Error: Error processing a session in confidence interval calculation.");
            MySQLclient.close(con);
            return false;
          }

          ID = (session + 1);
          SID = rs.getString("SID");
          realStartTime = rs.getLong("realStartTime");
          totalTimeToWait = rs.getDouble("totalTimeToWait");
          rs.close();

          if ( (realStartTime + Math.round(totalTimeToWait)) > Math.round(hiResTimer.milliTime())) // If this session is not already finished sleep timeToWait
            Thread.sleep( realStartTime + Math.round(totalTimeToWait - hiResTimer.milliTime()));

          try { // Verify if table SID exits
            rs = MySQLclient.executeQuery(con,
                                          "SELECT AVG(ABS(delay)) FROM `" + SID+"`");
            rs.close();
          }
          catch (Exception e2) { // If table SID doesn't exists an exception is thrown. We must make the Fecth Session
            // and wait till its end to continue.
            // Execute FetchSession

            // processFetchSessionToDataBase(String sqlServerAddress, String dataBase, String userName, String passWord,
            //                               String strSid, int numberOfPackets, double slotParameter /*in seconds*/,
            //                               long realStartTime /*in milliseconds*/, double totalTimeToWait /*in milliseconds*/){ // process Fetch-Session

            if(!fetch.processFetchSessionToDataBase(sqlServerAddress, dbJOWAMP_CI, userName, passWord, SID, numberOfPackets, slotParameter, realStartTime, totalTimeToWait)){
              System.err.println("Error: Error processing session.");
              MySQLclient.close(con);
              return false;
            }
          }

          // Check if SID entry exits in table SessionDescriptionRecords. If not it means that the fetch-Session isn't finished yet.
          // This table is used to save the Session Description Record information of all test session fetched. Each row
          // of this table represents a fetched session and is identified by the session SID. The Session Description Record
          // information of a session only should be saved in this table when the Fetch-Session is complete. So using this
          // table we can determine if the Fetch-Session of a test session is finished or not.
          fetchSessionFinished = true;
          do {
            rs = MySQLclient.executeQuery(con,
                                          " SELECT COUNT(*) FROM SessionDescriptionRecords" +
                                          " WHERE sid = '" + SID + "'"); // If the table doesn't exist, an exception is thrown
            rs.next();
            if (rs.getInt(1) != 1) { // Wait for Fetch-Session Results (There are no results yet).
              // We should wait for the Fetch-Session results only for 2xTIMEOUT after the test session is finished. If
              // actual time is less than 2xTIMEOUT from realStartTime+totalTimeToWait 'There are no results yet', else 'No results found'.
              if ( (realStartTime + Math.round(totalTimeToWait) + 2 * TIMEOUT) <
                  Math.round(hiResTimer.milliTime())) {
                System.err.println("Error: No results found for session " + ID +
                                   " with sid = " + SID + " from interval " +
                                   (i + 1) +
                                   ".");
                return false;
              }
              Thread.sleep(10000);
              fetchSessionFinished = false;
            }
            else
              fetchSessionFinished = true;

            rs.close();

          }
          while (!fetchSessionFinished);

          try {

            // Compute min/average/max delay for this test session
            rs = MySQLclient.executeQuery(con,
                                          "SELECT MIN(ABS(delay)) AS minDelay, AVG(ABS(delay)) AS averageDelay, MAX(ABS(delay)) AS maxDelay FROM `" +
                                          SID +
                                          "` WHERE receiveTimestamp != 0");
            rs.next();
            minDelay = rs.getDouble("minDelay");
            delay[i] = rs.getDouble("averageDelay");
            averageDelay = averageDelay + delay[i];
            maxDelay = rs.getDouble("maxDelay");
            rs.close();

            // Compute delay variance of this single session
            rs = MySQLclient.executeQuery(con,
                                          " SELECT ABS(delay) AS Delay FROM `" +
                                          SID +
                                          "` WHERE receiveTimestamp != 0");
            varDelay = 0;
            c = 0;
            while (rs.next()) {
              varDelay = varDelay + (double)Math.pow( (rs.getDouble("Delay") - delay[i]), 2);
              c++;
            }
            rs.close();
            varDelay = (c == 1) ? varDelay : varDelay / (c - 1);

            // Compute number of packets received (represents all packets received, including packets received after timeOut and duplicates)
            rs = MySQLclient.executeQuery(con,
                                          " SELECT COUNT(*) AS numPacketsReceived FROM `" + SID +"`");
            rs.next();
            numPacketsReceived = rs.getInt("numPacketsReceived");
            rs.close();

            // Compute number of losses (packets received after timeout) for this test session
            rs = MySQLclient.executeQuery(con,
                                          " SELECT COUNT(DISTINCT sequenceNumber) AS numTimeoutLosses FROM `" +
                                          SID +
                                          "` WHERE receiveTimestamp = 0");
            rs.next();
                        // To compute the losses for the confidence interval calculation, we should use loss ratios
                        // (number of packets lost/number of packets sent)
            numTimeoutLosses[i] = (double)rs.getInt("numTimeoutLosses")/numberOfPackets;
            averageTimeoutLosses = averageTimeoutLosses + numTimeoutLosses[i];
            rs.close();

            // Compute number of loses (packets not received) for this test session
            rs = MySQLclient.executeQuery(con,
                                          " SELECT COUNT(DISTINCT sequenceNumber) AS packetsReceived FROM `" +
                                          SID +
                                          "` WHERE sequenceNumber >= 0 AND sequenceNumber < " +
                                          numberOfPackets);
            rs.next();
                        // To compute the losses for the confidence interval calculation, we should use loss ratios
                        // (number of packets lost/number of packets sent)
            numNotReceivedLosses[i] = (double)(numberOfPackets - rs.getInt("packetsReceived"))/numberOfPackets;
            averageNotReceivedLosses = averageNotReceivedLosses +
                numNotReceivedLosses[i];
            rs.close();

            // Compute number of Duplicates for this test session
            try{
              testError = false;
              rs = MySQLclient.executeQuery(con,
                                            " SELECT COUNT(DISTINCT sequenceNumber) AS numDuplicates FROM `" +
                                            SID +
                                            "` WHERE sequenceNumber IN (SELECT sequenceNumber FROM `" +
                                            SID +
                                            "` GROUP BY sequenceNumber" +
                                            " HAVING COUNT(sequenceNumber) > 1)");
            }catch(Exception ignore){
              testError = true;
            }

            if(!testError){
              rs.next();
              // To compute the duplicates for the confidence interval calculation, we should use the ratio value
              // (number of packets duplicated/number of packets sent)
              numDuplicates[i] = (double)rs.getInt("numDuplicates")/numberOfPackets;
              averageNumDuplicates = averageNumDuplicates + numDuplicates[i];
              rs.close();
            }else{  // Some older versions of MySQL doesn't support this sql statement. So, if an error occurs
                    // we should calculate numDuplicates using the method below:

              rs = MySQLclient.executeQuery(con,"SELECT COUNT(*) AS numOfRecords FROM `"+SID+"`");
              rs.next();
              sequenceNumberArray = new int[rs.getInt("numOfRecords")];
              rs.close();
              rs = MySQLclient.executeQuery(con,"SELECT sequenceNumber FROM `"+SID+"`");
              c = 0;
              while(rs.next()){
                sequenceNumberArray[c] = rs.getInt("sequenceNumber");
                c++;
              }
              rs.close();
              duplicates = 0;
              for (j = 0; j < numberOfPackets; j++) {
                count = 0;
                for (c = 0; c < sequenceNumberArray.length; c++) {
                  if (sequenceNumberArray[c] == j) {
                    count++;
                    if (count == 2) {
                      count = 0;
                      duplicates++;  // Number of duplicates packets
                      break;
                    }
                  }
                }
              }
              // To compute the duplicates for the confidence interval calculation, we should use the ratio value
              // (number of packets duplicated/number of packets sent)
              numDuplicates[i] = (double)duplicates/numberOfPackets;
              averageNumDuplicates = averageNumDuplicates + numDuplicates[i];
            }

            // Compute Throughput for this test session
            rs = MySQLclient.executeQuery(con,
                                          "SELECT MIN(sendTimestamp) AS beginTime, MAX(sendTimestamp) AS endTime from `" +
                                          SID+"`"); // Note: Even if packet is marked as received after timeout (receiveTimestamp=0)
            rs.next(); // we should use them to compute throughput.
            beginTime = rs.getDouble("beginTime"); // Used to compute Throughput
            endTime = rs.getDouble("endTime");
            rs.close();

            testPacketLength = ( (paddingLength +
                                  StreamPackets_Unauthenticated_PACKET_LENGTH) <
                                _OWP_TESTREC_SIZE) ? (_OWP_TESTREC_SIZE) :
                (paddingLength + StreamPackets_Unauthenticated_PACKET_LENGTH);
            throughput[i] = (testPacketLength * numberOfPackets) /
                ( (endTime - beginTime) / 1000);
            averageThroughput = averageThroughput + throughput[i];

            // Save the statistic results of this single session in the SS_ResultTable
            // Note: For packet loss and duplication we should save their ratio values.
            MySQLclient.executeUpdate(con,
                                      "INSERT INTO `"+SS_ResultTable+"` ( SID, numOfPacketsSent, numPacketsReceived, minDelay, averageDelay, maxDelay, varDelay, numTimeoutLosses, numNotReceivedLosses, numDuplicates, throughput) VALUES ('" +
                                      SID +
                                      "', " + numberOfPackets + ", " +
                                      numPacketsReceived + ", " + minDelay + ", " +
                                      delay[i] + ", " + maxDelay + ", " +
                                      varDelay +
                                      ", " + numTimeoutLosses[i] + ", " +
                                      numNotReceivedLosses[i] + ", " +
                                      numDuplicates[i] + ", " + throughput[i] +
                                      ")");

          }
          catch (Exception e2) {
            System.err.println("Error: " + e2.toString());
            MySQLclient.close(con);
            return false;
          }

          if ( (i + 1) == numOfTestsInInterval) { // Calculate confidence interval
            // Compute average delay's confidence interval for this test session
            averageDelay = averageDelay / numOfTestsInInterval;
            varDelay = 0;
            for (j = 0; j < numOfTestsInInterval; j++) {
              varDelay = varDelay + (double)Math.pow( (delay[j] - averageDelay), 2);
            }
            varDelay = (numOfTestsInInterval == 1) ? varDelay :
                varDelay / (numOfTestsInInterval - 1);
            minIntervalDelay = averageDelay -
                alfaVal * (double)Math.sqrt(varDelay / numOfTestsInInterval);
            maxIntervalDelay = averageDelay +
                alfaVal * (double)Math.sqrt(varDelay / numOfTestsInInterval);

            // Compute number of losses's (packets received after timeout) confidence interval for this test session
            averageTimeoutLosses = averageTimeoutLosses / numOfTestsInInterval;
            varTimeoutLosses = 0;
            for (j = 0; j < numOfTestsInInterval; j++) {
              varTimeoutLosses = varTimeoutLosses +
                  (double)Math.pow( (numTimeoutLosses[j] - averageTimeoutLosses), 2);
            }
            varTimeoutLosses = (numOfTestsInInterval == 1) ? varTimeoutLosses :
                varTimeoutLosses / (numOfTestsInInterval - 1);
            minIntervalTimeoutLosses = averageTimeoutLosses -
                alfaVal * (double)Math.sqrt(varTimeoutLosses / numOfTestsInInterval);
            maxIntervalTimeoutLosses = averageTimeoutLosses +
                alfaVal * (double)Math.sqrt(varTimeoutLosses / numOfTestsInInterval);

            // Compute number of loses's (packets not received) confidence interval for this test session
            averageNotReceivedLosses = averageNotReceivedLosses /
                numOfTestsInInterval;
            varNotReceivedLosses = 0;
            for (j = 0; j < numOfTestsInInterval; j++) {
              varNotReceivedLosses = varNotReceivedLosses +
                  (double)Math.pow( (numNotReceivedLosses[j] - averageNotReceivedLosses),
                           2);
            }
            varNotReceivedLosses = (numOfTestsInInterval == 1) ?
                varNotReceivedLosses :
                varNotReceivedLosses / (numOfTestsInInterval - 1);
            minIntervalNotReceivedLosses = averageNotReceivedLosses -
                alfaVal * (double)Math.sqrt(varNotReceivedLosses / numOfTestsInInterval);
            maxIntervalNotReceivedLosses = averageNotReceivedLosses +
                alfaVal * (double)Math.sqrt(varNotReceivedLosses / numOfTestsInInterval);

            // Compute number of Duplicates's confidence interval for this test session
            averageNumDuplicates = averageNumDuplicates / numOfTestsInInterval;
            varNumDuplicates = 0;
            for (j = 0; j < numOfTestsInInterval; j++) {
              varNumDuplicates = varNumDuplicates +
                  (double)Math.pow( (numDuplicates[j] - averageNumDuplicates), 2);
            }
            varNumDuplicates = (numOfTestsInInterval == 1) ? varNumDuplicates :
                varNumDuplicates / (numOfTestsInInterval - 1);
            minIntervalNumDuplicates = averageNumDuplicates -
                alfaVal * (double)Math.sqrt(varNumDuplicates / numOfTestsInInterval);
            maxIntervalNumDuplicates = averageNumDuplicates +
                alfaVal * (double)Math.sqrt(varNumDuplicates / numOfTestsInInterval);

            // Compute Throughput's confidence interval for this test session
            averageThroughput = averageThroughput / numOfTestsInInterval;
            varThroughput = 0;
            for (j = 0; j < numOfTestsInInterval; j++) {
              varThroughput = varThroughput +
                  (double)Math.pow( (throughput[j] - averageThroughput), 2);
            }
            varThroughput = (numOfTestsInInterval == 1) ? varThroughput :
                varThroughput / (numOfTestsInInterval - 1);
            minIntervalThroughput = averageThroughput -
                alfaVal * (double)Math.sqrt(varThroughput / numOfTestsInInterval);
            maxIntervalThroughput = averageThroughput +
                alfaVal * (double)Math.sqrt(varThroughput / numOfTestsInInterval);

            // Record confidence interval in table CI_ResultTable

            MySQLclient.executeUpdate(con,
                                      "insert into `"+CI_ResultTable+"` (averageDelay, varDelay, minIntervalDelay, maxIntervalDelay, averageTimeoutLosses, varTimeoutLosses, minIntervalTimeoutLosses, maxIntervalTimeoutLosses, averageNotReceivedLosses, varNotReceivedLosses, minIntervalNotReceivedLosses, maxIntervalNotReceivedLosses, averageNumDuplicates, varNumDuplicates, minIntervalNumDuplicates, maxIntervalNumDuplicates, averageThroughput, varThroughput, minIntervalThroughput, maxIntervalThroughput ) values ('" +
                                      averageDelay +
                                      "', '" + varDelay + "', '" +
                                      minIntervalDelay +
                                      "', '" + maxIntervalDelay + "', '" +
                                      averageTimeoutLosses + "', '" +
                                      varTimeoutLosses + "', '" +
                                      minIntervalTimeoutLosses + "', '" +
                                      maxIntervalTimeoutLosses + "', '" +
                                      averageNotReceivedLosses + "', '" +
                                      varNotReceivedLosses + "', '" +
                                      minIntervalNotReceivedLosses + "', '" +
                                      maxIntervalNotReceivedLosses + "', '" +
                                      averageNumDuplicates + "', '" +
                                      varNumDuplicates + "', '" +
                                      minIntervalNumDuplicates + "', '" +
                                      maxIntervalNumDuplicates + "', '" +
                                      averageThroughput + "', '" + varThroughput +
                                      "', '" + minIntervalThroughput
                                      + "', '" + maxIntervalThroughput + "')");
            i = -1;
            averageDelay = 0;
            averageTimeoutLosses = 0;
            averageNotReceivedLosses = 0;
            averageNumDuplicates = 0;
            averageThroughput = 0;
          }

          i++;

        }
      }

    }catch (Exception e) {
      try {
        MySQLclient.executeUpdate(con, "drop table `"+CI_ResultTable+"`");
        MySQLclient.executeUpdate(con, "drop table `"+SS_ResultTable+"`");
      }
      catch (Exception e2) {}
      System.out.println(e);
      MySQLclient.close(con);
      return false;
    }

    MySQLclient.close(con);
    return true;
  }


  public static void main(String[] args){

    if((args.length == 10)||(args.length == 12)){

      // The user can execute this module with all the configurations parameters in the command line, to calculate
      // the confidence intervals of an confidence interval test session.
      // In this case the user must instroduce 10 parameters. The results are saved into a data base.
      // Parameters (String serverAddress, String sqlServerAddress, String dbJOWAMP_CI, String userName,
      //             String passWord, String CIconfigTable, String CIsidTable, short CI_id, String CI_ResultTable,
      //             String SS_ResultTable, String ntpServerAddress, String updateInterval). The ntpServerAddress
      // and updateInteval are optional. If not defined the system time will be used to get the time reference.

      String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
      int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

      if(args.length == 12){  // Read the NTP server's address

        ntpServerAddress = args[10].trim();
        try { // Test if the NTP server address introduced is a valid address.
          if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
            System.err.println(
                "Error: The NTP server's address read is not a valid address (" +
                InetAddress.getByName(ntpServerAddress) + ").");
            System.exit(1);
          }
        }catch (Exception e) {
          System.err.println(
              "Error: The NTP server's address read is not a valid address (" +
              ntpServerAddress + ").");
          System.err.println();
          System.exit(1);
        }

        try{
          updateInterval = Integer.parseInt(args[11].trim());
        }catch (NumberFormatException e2) {
          System.err.println("Error: " + e2 + ".");
          System.err.println();
          System.exit(1);
        }

        if (updateInterval < 0) {
          System.err.println("Error: The value of updateInterval can not be a negative number (" +
                             updateInterval + ").");
          System.err.println();
          System.exit(1);
        }

      }

      // Define the object hiResTimer used to get the system time.
      HiResTimer hiResTimer = null;
      if(updateInterval == -1){  // Use default updateInterval defined on HiResTimer class.
        hiResTimer = new HiResTimer(ntpServerAddress);
      }else
        hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

      CalculateCI calculateCI = new CalculateCI(args[0], args[1], args[2], args[3], args[4],
                                                args[5], args[6], Short.parseShort(args[7]), args[8], args[9], hiResTimer);

      if(!calculateCI.processCI()){
        System.err.println("Error: Error processing confidence intervals calculation.");
        System.exit(1);
      }

      System.exit(0);  // In this case we must terminate the execution at the end of the processing.

    }

    System.err.println("Error: Number of arguments not valid");


  }

}
