/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the Class OWAMP_Base. All other Classes
 * extends this class. It defines the basic methods, which are used by the other classes
 * defined in other sections of this project.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 1.2
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */

import java.net.*;
import java.io.*;
import java.util.*;

public class OWAMP_Base extends Thread{

  // Class Variables

  static final int DEFAULT_PACKET_LENGTH = 112;
  static final int TIMEOUT = 120*1000;                         // timeout to wait a packet to be received, in ms.
  static final int TIMEOUTTOWAIT = 1*1000;                     // timeout to wait a StopSession packet to be received, in ms.
  static final short DEFAULT_SEVER_PORT = (short)22368;        // default Server port for TCP connection.
  static final short DEFAULT_SEVER_F_PORT = (short)22368;      // default Server port for TCP connection for
                                                               // Fetch Sessions.
  static final short DEFAULT_SENDER_PORT = (short)4181;        // default Sender port for TCP connection.
  static final short DEFAULT_SENDER_UDP_PORT = (short)21164;   // default Sender port for UDP connection.
  static final short DEFAULT_RECEIVER_PORT = (short)28181;     // default Receiver port for TCP connection.
  static final short DEFAULT_RECEIVER_UDP_PORT = (short)21164; // default Receiver port for UDP connection.
  static final short numMaxRS = 500;         // num max of Request Session that a Client/Server can accept.
  static final int _OWP_TESTREC_SIZE = 24;   // default minimum test packet size.
  static final long _2POW15 = ( (long) 32768); // 2 pow 15 = 32768


  // Constructors

  public OWAMP_Base(){
  }


  // Instance variables

  byte[] buffer;        // This is a buffer to put the incoming packet.
  int mode;             // The mode value to be used in the current OWAMP control and test sessions.
  Timestamp upTime;     // Uptime is a timestamp representing the time when the current instantiation
                        // of the server started operating.  (For example, in a multi-user general
                        // purpose operating system, it could be the time when the server process was
                        // started.)
  protected HiResTimer hiResTimer;  // Used to read the system clock.
  ConnectionSetupServerGreetingReceiver dataCSSGReceiver;   // Represents the datagram CSSG to be received.
  ConnectionSetupClientResponseReceiver dataCSCRReceiver;   // Represents the datagram CSCR to be received.
  ConnectionSetupServerResponseReceiver dataCSSRReceiver;   // Represents the datagram CSSR to be received.
  ServerSocket serverSocket;                                // Used to verify if a socket connection is closed
                                                            // using method isSocketConnected.
  protected Socket socketAux = new Socket(); // used on method isSocketConnected to verify if a socket is connected.


  // Instance methods

  /**
   * This method establishes the operation Connection Setup request (CSR), between the machines where it
   * is used (Client machine) and another machine (Server machine).
   * @param mode int - Is the mode operation to be used in this session of measurements.
   * The following mode values are meaningful: 1 for unauthenticated, 2 for authenticated, 4 for
   * encrypted.  The value of the Modes field sent by the server is the bit-wise OR of the mode
   * values that it is willing to support during this session.  Thus, last three bits of the Modes
   * 32-bit value are used.  The first 29 bits MUST be zero.  A client MUST ignore the values in
   * the first 29 bits of the Modes value.  (This way, the bits are available for future protocol
   * extensions.  This is the only intended extension mechanism.)
   * @param ia InetAddress - the InetAddress represents the address of the machine who accepts requests of
   * type Connection Setup (for example a Server). Note that a Connection Setup request can be made to a
   * Server, Sender or Receiver.
   * @param port int - the port in which a machine is waiting to receive CSR.
   * @param timeout int - a value of timeout to configure the socket connection established between the two
   * machines.
   * @return Socket - this method returns the socket created to establish the connection between the two
   * machines, if no failure occurs. Otherwise it returns null.
   */

  public Socket connectionSetupRequest(int mode, InetAddress ia, int port, int timeout){    // process Connection Setup

    Socket socket = new Socket();

    // Open a TCP connection to the InetAddress ia

    socket = openSocket(ia, port, timeout);

    if(socket == null){
      System.err.println("Error: The OWAMP element in machine "+ia.toString()+", on TCP port number "+port+", doesn't exist.");
      return null;
    }

    // Receive the ConnectionSetupServerGreeting packet

    dataCSSGReceiver = new ConnectionSetupServerGreetingReceiver();

    if(!dataCSSGReceiver.receiveCSSG(socket)){
      System.err.println("Error: Error in receiving the packet ConnectionSetupServerGreeting from the Server's machine.");
      closeSocket(socket);
      return null;
    }

    if((dataCSSGReceiver.modes != 0)&&(dataCSSGReceiver.modes < mode)){  // The server doesn't support the mode requested
      System.err.println("Error: The client's desired mode is unavailable.");
      closeSocket(socket);
      return null;
    }

    if(dataCSSGReceiver.modes == 0){  // The server doesn't wish to communicate with the client
      closeSocket(socket);
      System.err.println("Error: The server doesn't wish to communicate with the client.");
      return null;
    }

    switch(mode){
      case 1:  // Unauthenticated mode. Send the ConnectionSetupClientResponse packet

        ConnectionSetupClientResponseToBeSend dataCSCRToBeSend = new ConnectionSetupClientResponseToBeSend(mode);
        dataCSCRToBeSend.sendCSCR(socket);

        break;
      case 3:
        //authenticated mode
        break;
      case 7:
        //encrypted mode
        break;

    }

    // Receive the ConnectionSetupServerResponse packet

    dataCSSRReceiver = new ConnectionSetupServerResponseReceiver();

    if(!dataCSSRReceiver.receiveCSSR(socket)){
      System.err.println("Error: Error in receiving the packet ConnectionSetupServerResponse in operation Connection Setup.");
      closeSocket(socket);
      return null;
    }

    switch(dataCSSRReceiver.accept){
      case 0:   // The server accepts the authentication
        return socket;
      case 1:   // The server doesn't accept the authentication provided by the client
                // or, for some other reason, is not willing to conduct further
                // transactions in this OWAMP-Control session.
        System.err.println("Error: The server doesn't accept the authentication provided by the client or,for some other reason, is not willing to conduct further transactions in this OWAMP-Control session.");
        closeSocket(socket);
        return null;
      default:  // Reserved values
        System.err.println("Error: The client doesn't support the accept mode received from the server.");
        closeSocket(socket);
        return null;
    }

  }


  /**
   * This method establishes the operation Connection Setup, between the client's and the Server's machine.
   * @param modes int - Is the mode operation to be used in this session of measurements.
   * The following mode values are meaningful: 1 for unauthenticated, 2 for authenticated, 4 for
   * encrypted.  The value of the Modes field sent by the server is the bit-wise OR of the mode
   * values that it is willing to support during this session.  Thus, last three bits of the Modes
   * 32-bit value are used.  The first 29 bits MUST be zero.  A client MUST ignore the values in
   * the first 29 bits of the Modes value.  (This way, the bits are available for future protocol
   * extensions.  This is the only intended extension mechanism.)
   * @param socket Socket - socket connection to client's machine.
   * @param upTime Timestamp -  Uptime is a timestamp representing the time when the current instantiation
   * of the server started operating.  (For example, in a multi-user general purpose operating system, it
   * could be the time when the server process was started.)
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean connectionSetupResponse(int modes, Socket socket, Timestamp upTime){    // process Connection Setup

    byte[] challenge = new byte[16];
    Random rand = new Random();

    // Send the ConnectionSetupServerGreeting packet

    if((modes != 0)&&(modes != 1))  // Mode authenticated or encrypted
      for(int i = 0;i < challenge.length;i++)
        challenge[i] = (byte)(rand.nextInt() & 0xFF);  // Challenge is a random sequence of octets
                                                       // generated by the server; it is used subsequently
                                                       // by the client to prove possession of a shared
                                                       // secret.

    ConnectionSetupServerGreetingToBeSend dataCSSGToBeSend = new ConnectionSetupServerGreetingToBeSend(modes,challenge);
    dataCSSGToBeSend.sendCSSG(socket);

    if(modes == 0){  // If Modes value is zero, the server doesn't wish to communicate with the client and
                     // MAY close the connection immediately.
      closeSocket(socket);
      System.err.println("Error: The server doesn't wish to communicate with the client.");
      return false;
    }

    // Receive the ConnectionSetupClientResponse packet

    dataCSCRReceiver = new ConnectionSetupClientResponseReceiver();

    if(!dataCSCRReceiver.receiveCSCR(socket)){
      System.err.println("Error: Error receiving packet ConnectionSetupClientResponse in operation Connection Setup.");
      closeSocket(socket);
      return false;
    }

    if(dataCSCRReceiver.mode == 0){  // The client will not continue with the session.
      System.err.println("Error: The client will not continue with the session.");
      closeSocket(socket);
      return false;
    }

    if(dataCSCRReceiver.mode > modes){  // The server doesn't support the mode requested
      System.err.println("Error: The server doesn't support the mode requested by the client.");
      closeSocket(socket);
      return false;
    }

    // Send the ConnectionSetupServerResponse packet

    byte[] server_IV = new byte[16];
    byte accept = 0;

    if(dataCSCRReceiver.mode != 1){    // Mode authenticated or encrypted
      // The server should first verify the client authentication
      //accept = verifyAuthentication();
      if(accept == 0)  // The server does accept the authentication
        for (int c = 0; c < 16; c++) server_IV[c] = (byte)(rand.nextInt() & 0xFF);
    }

    if(accept != 0)   // If the authentication wasn't accepted, upTime should be set to a string of zeros
      upTime = new Timestamp((double)0,(short)0);

    ConnectionSetupServerResponseToBeSend dataCSSRToBeSend = new ConnectionSetupServerResponseToBeSend(accept,server_IV,upTime);
    dataCSSRToBeSend.sendCSSR(socket);

    if(accept != 0){  // The server doesn't accept the authentication
      closeSocket(socket);
      System.err.println("Error: The server doesn't accept the authentication send by the client.");
      return false;
    }

    return true;
  }


  /**
   * This method opens a socket connection to a machine using the IP address in variable ia
   * and the port value in variable _port.
   * @param ia InetAddress - address of the machine.
   * @param _port int - the port to use to open the socket.
   * @param timeOut int - time to configure the socket timeout.
   * @return Socket - the socket connection established is returned or null if an error occurs.
   */

  public Socket openSocket(InetAddress ia, int _port, int timeOut){

    try{
      Socket socket = new Socket(ia, _port);
      socket.setSoTimeout(timeOut);
      return socket;
    }
    catch(IOException e){
      System.err.println(e);
    }

    return null;
  }


  /**
   * This method closes the socket connection of a Socket.
   * @param socket Socket - socket connection to be closed.
   */

  public static void closeSocket(Socket socket){
    try{
      if((socket!=null)&&((socket.isConnected())||(!socket.isClosed())))  // Close socket connection.
        socket.close();
    }
    catch(IOException ignored){
    }

  }


  /**
   * This method is used to receive a packet in the socket 'socket'. This
   * packet has len bytes an it is recorded in array _buffer, starting in offset off.
   * @param socket Socket - the socket to receive the packet.
   * @param _buffer byte[] - The contend of the packet received is recorded in the variable _buffer.
   * @param off int - the start offset in array _buffer at which the data is written.
   * @param len int - the maximum number of bytes to read.
   * @return int - This method returns the number of bytes read or -1 if a failure occurs.
   */

  public static int receivePacket (Socket socket, byte[] _buffer, int off, int len) {
    int readn = 0,read=0; // number of bytes read

    try{

      InputStream theInputStream = socket.getInputStream();

      do{
        read = theInputStream.read(_buffer,off,len); // receive the incoming packet
        readn=readn+read;
        off = readn;
        len=len-read;
      }while((read!=-1)&&(len>0));  // While packet is not totally received, read again.

      return readn;

    }
    catch(IOException ex){
      return -1;
    }
  }


  /**
   * Method used to save the dataRS and the arrayOfdataSSD in a file.
   * @param dataRS RequestSession
   * @param arrayOfdataSSD ScheduleSlotDescriptions[]
   * @param outputFile RandomAccessFile
   * @return boolean -  This method returns true if everything goes ok, otherwise it returns false.
   */

  public boolean saveRStoFile(RequestSession dataRS, ScheduleSlotDescriptions[] arrayOfdataSSD, RandomAccessFile outputFile){

    try {
      outputFile.writeByte(dataRS.type);
      outputFile.writeByte(dataRS.mbz_iPVN);
      outputFile.writeByte(dataRS.confSender);
      outputFile.writeByte(dataRS.confReceiver);
      outputFile.writeInt(dataRS.numberOfScheduleSlots);
      outputFile.writeInt(dataRS.numberOfPackets);
      outputFile.writeShort((short)dataRS.senderPort);
      outputFile.writeShort((short)dataRS.receiverPort);
      outputFile.write(dataRS.senderAddress);
      outputFile.write(dataRS.receiverAddress);
      outputFile.write(dataRS.sid);
      outputFile.writeInt(dataRS.paddingLenght);
      outputFile.writeInt(dataRS.startTime.integerPartOfSeconds);
      outputFile.writeInt(dataRS.startTime.fractionalPartOfSeconds);
      outputFile.writeInt(dataRS.timeout.integerPartOfSeconds);
      outputFile.writeInt(dataRS.timeout.fractionalPartOfSeconds);
      outputFile.write(dataRS.typeP_Descriptor);
      outputFile.write(dataRS.mBZ2);
      outputFile.write(dataRS.integrityZeroPadding);

      for (int j = 0; j < dataRS.numberOfScheduleSlots; j++) { // record the arrayOfdataSSD
        outputFile.writeByte(arrayOfdataSSD[j].slotType);
        outputFile.writeInt(arrayOfdataSSD[j].slotParameter.integerPartOfSeconds);
        outputFile.writeInt(arrayOfdataSSD[j].slotParameter.
                            fractionalPartOfSeconds);
      }

    }
    catch (IOException ex) {
      System.err.println(ex);
      return false;
    }

    return true;

  }



  /**
   * Method used to read the dataRS and the arrayOfdataSSD from a file.
   * @param dataRS RequestSession
   * @param inFile RandomAccessFile
   * @return ScheduleSlotDescriptions[] -  This method returns the ScheduleSlotDescriptions[] if everything goes ok,
   * otherwise it returns null.
   */

  public ScheduleSlotDescriptions[] readRSfromFile(RequestSession dataRS, RandomAccessFile inFile){

    ScheduleSlotDescriptions[] arrayOfdataSSD = null;

    try{
      dataRS.type = inFile.readByte();
      dataRS.mbz_iPVN = inFile.readByte();
      dataRS.confSender = inFile.readByte();
      dataRS.confReceiver = inFile.readByte();
      dataRS.numberOfScheduleSlots = inFile.readInt();
      dataRS.numberOfPackets = inFile.readInt();
      dataRS.senderPort = unsignedShortToInt(inFile.readShort());
      dataRS.receiverPort = unsignedShortToInt(inFile.readShort());
      inFile.read(dataRS.senderAddress, 0, 16);
      inFile.read(dataRS.receiverAddress, 0, 16);
      inFile.read(dataRS.sid, 0, 16);
      dataRS.paddingLenght = inFile.readInt();
      dataRS.startTime.integerPartOfSeconds = inFile.readInt();
      dataRS.startTime.fractionalPartOfSeconds = inFile.readInt();
      dataRS.timeout.integerPartOfSeconds = inFile.readInt();
      dataRS.timeout.fractionalPartOfSeconds = inFile.readInt();
      inFile.read(dataRS.typeP_Descriptor, 0, 4);
      inFile.read(dataRS.mBZ2, 0, 8);
      inFile.read(dataRS.integrityZeroPadding, 0, 16);

      // read the arrayOfdataSSD
      arrayOfdataSSD = new ScheduleSlotDescriptions[dataRS.numberOfScheduleSlots];
      for(int j=0;j<dataRS.numberOfScheduleSlots;j++){
        arrayOfdataSSD[j] = new ScheduleSlotDescriptions();
        arrayOfdataSSD[j].slotType = inFile.readByte();
        arrayOfdataSSD[j].slotParameter.integerPartOfSeconds = inFile.readInt();
        arrayOfdataSSD[j].slotParameter.fractionalPartOfSeconds = inFile.readInt();
      }

    }catch(IOException ex){
      System.err.println(ex);
      return null;
    }

    return arrayOfdataSSD;

  }


  /**
   * This method returns the total interval of time in ms that the Client/Server should wait before ending
   * the session. This time corresponds to the sum of the time since now to the start of the session + all
   * slotParameter + one last timeout, that receiver should wait to receive the last packet. We do this
   * because we have a schedule with a given number of `slots'. Each slots has a type and a parameter.  Two
   * types are supported: exponentially distributed pseudo-random quantity (denoted by a code of 0) and a
   * fixed quantity (denoted by a code of 1).  The parameter is expressed as a timestamp and specifies a time
   * interval.  For a type 0 slot (exponentially distributed pseudo-random quantity) this interval is the
   * mean value (or 1/lambda if the distribution density function is expressed as lambda*exp(-lambda*x) for
   * positive values of x).  For a type 1 slot, the parameter is the delay itself.  The sender starts with the
   * beginning of the schedule, and `executes' the instructions in the slots: for a slot of type 0, wait
   * exponentially distributed time with mean of the specified parameter and then send a test packet (and
   * proceed to the next slot); for a slot of type 1, wait the specified time and send a test packet (and
   * proceed to the next slot).  The schedule is circular: when there are no more slots, the sender returns to
   * the first slot.
   * @param dataRS RequestSession - represents the datagram Request Session
   * @param arrayOfdataSSD ScheduleSlotDescriptions[] - is the array of all Schedule Slot Descriptions
   * @param hiResTimer HiResTimer - variable used to read the system clock time.
   * @return double - it returns the total time to wait.
   */

  public double totalTimeToWait(RequestSession dataRS, ScheduleSlotDescriptions[] arrayOfdataSSD, HiResTimer hiResTimer){
    double totalTimeToWait = dataRS.startTime.timeInMillisecondsSince1970() - ((long)hiResTimer.milliTime());

    if(totalTimeToWait < 0)  // start time in the past.
      totalTimeToWait = 0;

    for(int i=0, j = 0;i<dataRS.numberOfPackets;i++,j++){

      totalTimeToWait = totalTimeToWait + arrayOfdataSSD[j].slotParameter.timeInMillisecondsSince1900();
      if(j == (arrayOfdataSSD.length -1))  // If we are at the end of the array of SSD we have to
        j = -1;                            // turn around.  j = -1 -> when incremented by the cycle
                                           // for, it will have the correct value (zero).
    }

    return(totalTimeToWait + dataRS.timeout.timeInMillisecondsSince1900());
  }


  /**
    * This method returns the total interval of time in ms that the Receiver should wait before ending the session (it
    * is the total test duration).
    * This time corresponds to the sum of all slotParameter + one last timeout, that receiver should wait to receive
    * the last packet. We do this because we have a schedule with a given number of `slots'. Each slots has a type
    * and a parameter.  Two types are supported: exponentially distributed pseudo-random quantity (denoted by a code
    * of 0) and a fixed quantity (denoted by a code of 1).  The parameter is expressed as a timestamp and specifies
    * a time interval.  For a type 0 slot (exponentially distributed pseudo-random quantity) this interval is the
    * mean value (or 1/lambda if the distribution density function is expressed as lambda*exp(-lambda*x) for positive
    * values of x).  For a type 1 slot, the parameter is the delay itself.  The sender starts with the beginning of
    * the schedule, and `executes' the instructions in the slots: for a slot of type 0, wait exponentially distributed
    * time with mean of the specified parameter and then send a test packet (and proceed to the next slot); for a slot
    * of type 1, wait the specified time and send a test packet (and proceed to the next slot).  The schedule is
    * circular: when there are no more slots, the sender returns to the first slot.
    * @param dataRS RequestSession - array of data RS
    * @param arrayOfdataSSD ScheduleSlotDescriptions[] - array of data SSD
    * @return double - this method returns the total time to wait.
    */

   public double totalTimeToWait(RequestSession arrayOfdataRS, ScheduleSlotDescriptions[] arrayOfdataSSD){
     double totalTimeToWait = 0;

     for(int i=0, j = 0;i<(arrayOfdataRS.numberOfPackets-1);i++,j++){
       totalTimeToWait = totalTimeToWait + arrayOfdataSSD[j].slotParameter.timeInMillisecondsSince1900();
       if(j == (arrayOfdataSSD.length -1))  // If we are at the end of the array of SSD we have to
         j = -1;                            // turn around.  j = -1 -> when incremented by the cycle
                                            // for, it will have the correct value (zero).
     }

     return(totalTimeToWait + arrayOfdataRS.timeout.timeInMillisecondsSince1900());
  }


 /**
  * This method is used to convert the byte array sid containing the SID of a session to a string of
  * characters.
  * @param sid byte[] - byte array sid containing the SID of a session.
  * @return String - string of character representing the SID.
  */

  public static String  owampHexEncode(byte[] sid){

    char hex[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
    char[] buff = new char[sid.length*2]; // buff must be at least (nbytes*2) +1 long or memory will be over-run.

    for(int i=0,j=0;i<sid.length;i++,j++){
      buff[j++] = hex[(int)((sid[i] & 0xFF) >> 4)];
      buff[j] = hex[(int)(sid[i] & 0x0f)];
    }

    String strSid = new String(buff);

    return strSid;

  }


  /**
   * This method is used to convert a string of characters containing the SID of
   * a session to the byte array sid .
   * @param str String - string of character representing the SID.
   * @return sid byte[] - byte array sid containing the SID of a session.
   */

  public byte[] owampHexDecode(String str){
    char hex[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
    int	i,j,offset, strIndex = 0, nbytes = 16;
    char a;
    byte[] sid = new byte[16];
    char[] buff = str.toCharArray();
    boolean byteset = false;

    for(i=0;i<nbytes;i++){
      sid[i] = 0;
      for(j=0;(strIndex < buff.length)&&(j<2);j++,strIndex++){
        a = buff[strIndex];
        for(offset=0;offset<hex.length;offset++){
          if(a == hex[offset]){
            sid[i] |= offset;
            if(j==0)
              sid[i] <<= 4;
            byteset = true;
            break;
          }
        }
        if(!byteset)
          System.exit(1);
        byteset = false;
      }
    }

    return sid;
  }


  /**
   * This method is used to verify if the ip InetAddress is a local address or not.
   * @param ip InetAddress
   * @return boolean - This method returns true if ip InetAddress is a local address, otherwise it returns false.
   */

  public static boolean  isAnyLocalAddress (InetAddress ip){

    try {

      Enumeration e = NetworkInterface.getNetworkInterfaces();

      while(e.hasMoreElements()) {
        NetworkInterface netface = (NetworkInterface)
        e.nextElement();

        Enumeration e2 = netface.getInetAddresses();

        while (e2.hasMoreElements()){
          if(((InetAddress) e2.nextElement()).equals(ip)){
            return true;
          }
        }
      }
    }
    catch (Exception e) {
      System.err.println ("Error: " + e);
    }

    return false;

  }


  /**
   * This method tries to get a local IP address different of 127.0.0.1, and return it as a InetAddress
   * variable. This IP will be used to form the sid variable of packet Request Session.
   * If the generating machine has any IPv4 addresses (with the exception of loopback), one of
   * them should be used, even if all communication is IPv6 based.
   * @return boolean - This method returns the IP address if it finds this IP otherwise, it returns null.
   */

  public InetAddress  getIpAddress (){

    InetAddress localIP = null;

    try {

      Enumeration e = NetworkInterface.getNetworkInterfaces();

      while(e.hasMoreElements()) {
        NetworkInterface netface = (NetworkInterface)
        e.nextElement();

        Enumeration e2 = netface.getInetAddresses();

        while (e2.hasMoreElements()){
          InetAddress ip = (InetAddress) e2.nextElement();
          if(!(ip.isLoopbackAddress())){
            localIP = ip;
            if(ip.getAddress().length == 4)  //IPv4 address
              return ip;
          }
        }
      }
    }
    catch (Exception e) {
      System.err.println ("Error: " + e);
    }

    return localIP;

  }


  /**
   * This method is used to generate the SID used in OWAMP protocol.
   * Sid should be constructed by the concatenation of 4-octet IP number belonging to the generating machine
   * 8-octet timestamp, and 4-octet random value. If this is not possible sid must be made unpredictable.
   * Gets local IP address which is different from the address 127.0.0.1.
   * @return byte[] - This method returns the a byte array that represents the SID to be used in OWAMP protocol.
   */

  public byte[] generateSid(){

    byte sid[] = new byte[16];
    Random rand = new Random();

    InetAddress localIP = getIpAddress();

    if(localIP != null){
      if(localIP.getAddress().length == 4)  // IPv4 address
        System.arraycopy(localIP.getAddress(),0,sid,0,4);
      else{                                                 // IPv6 address. In this case the last 4 octets of the
        System.arraycopy(localIP.getAddress(),12,sid,0,4);  // IPv6 address is used to form sid.
      }

      Timestamp timestamp = new Timestamp(System.currentTimeMillis(),(short)1);  // Time in milliseconds since 1Jan1970

      byte[] array = IntToByteArray(timestamp.integerPartOfSeconds);
      System.arraycopy(array,0,sid,4,4);

      array = IntToByteArray(timestamp.fractionalPartOfSeconds);
      System.arraycopy(array,0,sid,8,4);

      for(int i=12;i<16;i++)
        sid[i] = (byte)(rand.nextInt() & 0xFF);

    }
    else                              // sid must be made unpredictable
      for(int i=0;i<sid.length;i++)
        sid[i] = (byte)(rand.nextInt() & 0xFF);

    return sid;

  }


  /**
   * This method transfers the value of a int to a byte array.
   * @param val int - Int value to be transferred to a byte array.
   * @return byte[] - Reference of the byte array, which has the int value.
   */

  private byte[] IntToByteArray(int val){

    byte[] array = new byte[4];

    for (int iInd = 0; iInd < array.length; ++iInd) // for each byte
      array[array.length-1-iInd] = (byte) ( (val >> (iInd * 8)) & 0x00FF);

    return array;

  }


  /**
   * This method is used to determine the Sender's or Receiver's address of the RS. These addresses are
   * returned as an InetAddress .
   * @param mbz_iPVN byte - represents the IP version used in this session.
   * @param address byte[]  - The Sender's or Receiver's address as a byte array.
   * @return InetAddress - The Sender's or Receiver's InetAddress representation to be returned by
   * this method.
   */

  public InetAddress determineSeReAddress(byte mbz_iPVN, byte[] address){

    byte[] hostname;
    InetAddress ia = null;

    if ( ( (int) (mbz_iPVN & 0xFF) == 4)) { // IPv4
      hostname = new byte[4];
      System.arraycopy(address, 0, hostname, 0, 4);
    }
    else { // IPv6
      hostname = new byte[16];
      System.arraycopy(address, 0, hostname, 0, 16);
    }

    try{
      ia = InetAddress.getByAddress(hostname);
    }catch(IOException ex){
      System.err.println(ex);
    }

    return ia;
  }


  /**
   * This method is used to stop processing for val milliseconds.
   * @param val int - the value of milliseconds to wait before continuing the processing.
   */

  public void sleepTime(int val){

    try{
      Thread.sleep(val);
    }
    catch ( Exception x ){
      x.printStackTrace();
    }
  }


  /**
   * This method is used to verify if the socket is connected. If a exception is thrown in bind operation
   * the socket is connected (socket is already bound to socketAddress), otherwise the socket is not connected
   * because it binds the socketAddress.
   * @param socket Socket - socket to be tested.
   * @return boolean - true if socket is connected, otherwise return false.
   */

  public boolean isSocketConnected(Socket socket){
    try{
      socketAux.bind(socket.getRemoteSocketAddress());
      socketAux.close();
      socket.close();  // Socket was bound succesfully so it was closed. So we should close it again to keep its state.
      return false;
    }catch(Exception ex){
      try{
        socketAux.close();
      }catch(Exception ex2){}
      return(socket.isConnected())&&(!socket.isClosed());
    }
  }


  /**
   * Method used to delete a directory.
   * @param file File - An abstract representation of the directory pathname of the directory to be removed.
   * @throws IOException
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public static boolean deleteDir(File file) throws IOException {
    boolean success = true;
    for(boolean retry=true; retry;) {
      retry=false;
      if(file.isDirectory()) {
        File [] child = file.listFiles();
        if(child!=null && child.length>0)
          for(int i=0; i<child.length; i++)
            success &= deleteDir(child[i]);
      }
      success &= file.delete();
    }
    return success;
  }


  /**
   * Most (all?) modern systems use 2's complement representation for negative numbers. In this
   * representation the distinction between signed and unsigned types really comes down to how the
   * high order bit is interpreted. For a Java short quantity, there are 16 bits. The low order 15
   * bits represent a value between 0 and 2^15 - 1. The top bit represents -2^15 if it is set, and
   * zero if it is not. The value of the quantity is the sum of the values represented by the low
   * order 15 bits and the high order bit. For unsigned int quantities, the high order bit represents
   * 2^15 if it is set, and 0 if it is not.
   * This method converts the value of an unsigned short variable to a variable of type signed int. We use
   * a variable of type int (32 bits) because it can keep any unsigned short values without overflow.
   * @param val short - unsigned short value to convert to signed int.
   * @return int - signed int value returned.
   */

  public static int unsignedShortToInt(short val) {

    if ( (val & 0x8000) == 0x8000) { // For unsigned short quantities, the high order bit represents
      // 2^15 if it is set, and 0 if it is not.
      int aux = (int) (val & 0x7FFF); // The value of the quantity is the sum of the values
      // represented by the low order 15 bits and the high order bit.
      return ( (int) (aux + _2POW15));
    }

    return ( (int) val);
  }

}
