/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the class used to implement the OWAMP Fetch-Client.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 1.2
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */


import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.Math;
import java.math.BigDecimal;
import java.sql.*;  // import needed for JDBC access

public class OWAMP_FetchClient extends OWAMP_Base{

  // Constructors

  /**
   * Constructor for class OWAMP_FetchClient.
   * @param serverAddress InetAddress - represents the Server address.
   * @param serverTCPportNumber int - represents the Server's TCP port to use in OWAMP-Control communications.
   * @param socketToServer Socket - socket to Server to be used in OWAMP-Control communications with Server.
   * @param systemInReader BufferedReader - Used to implement the System.in.
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the system time.
   */

  public OWAMP_FetchClient(InetAddress serverAddress,int serverTCPportNumber, Socket socketToServer, BufferedReader systemInReader, HiResTimer hiResTimer){

    this.serverAddress = serverAddress;
    this.serverTCPportNumber = serverTCPportNumber;

    buffer = new byte[DEFAULT_PACKET_LENGTH];
    this.socketToServer = socketToServer;
    this.systemInReader = systemInReader;
    this.hiResTimer = hiResTimer;

    // Test fetchClientDir directory. If it doesn't exists it should be created.
    if(!fetchClientDir.exists())
      fetchClientDir.mkdir();

    // Test packet results directory. If it doesn't exists it should be created.
    if(!packetResults.exists())
      packetResults.mkdir();

    // Test test results directory. If it doesn't exist it should be created.
    if(!testResults.exists())
      testResults.mkdir();

  }


  /**
   * Constructor for class OWAMP_FetchClient used, by the OWAMP_ControlClient class, to fetch the results.
   * @param serverAddress InetAddress - represents the Server address.
   * @param serverTCPportNumber int - represents the Server's TCP port to use in OWAMP-Control communications.
   * @param socketToServer Socket - socket to Server to be used in OWAMP-Control communications with Server.
   * @param arrayOfdataRS RequestSessionToBeSend[] - When this method is used to fetch the test results in the
   * OWAMP_ControlClient all the RS are in the arrayOfdataRS variable, so there is no need to read them from the
   * clientRSdir. Also this directory may have other RS files of others instances of the OWAMP_ControlClient class.
   * When this method is used to fetch the test results in the OWAMP_ControlClient these sessions shouldn't appear
   * in the Fetch-Session's option list.
   * @param arrayOfdataSSD ScheduleSlotDescriptionsToBeSend[][]
   * @param numRS short
   * @param systemInReader BufferedReader - Used to implement the System.in.
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the time.
   */

  public OWAMP_FetchClient(InetAddress serverAddress,int serverTCPportNumber, Socket socketToServer, RequestSessionToBeSend[] arrayOfdataRS, ScheduleSlotDescriptionsToBeSend[][] arrayOfdataSSD, short numRS, BufferedReader systemInReader, HiResTimer hiResTimer){

    this.serverAddress = serverAddress;
    this.serverTCPportNumber = serverTCPportNumber;
    this.arrayOfdataRS = arrayOfdataRS;
    this.arrayOfdataSSD = arrayOfdataSSD;
    this.numRS = numRS;
    buffer = new byte[DEFAULT_PACKET_LENGTH];
    this.socketToServer = socketToServer;
    this.systemInReader = systemInReader;
    this.hiResTimer = hiResTimer;

    // Test fetchClientDir directory. If it doesn't exists it should be created.
    if(!fetchClientDir.exists())
      fetchClientDir.mkdir();

    // Test packet results directory. If it doesn't exists it should be created.
    if(!packetResults.exists())
      packetResults.mkdir();

    // Test test results directory. If it doesn't exist it should be created.
    if(!testResults.exists())
      testResults.mkdir();

    try {
      serverSocket = new ServerSocket(); // Used to verify if a socket connection is closed
                                         // using method isSocketConnected.
    }
    catch (IOException e) {
      System.err.println(e);
    }

  }

  /**
   * Constructor for class OWAMP_FetchClient used when processing Start-session and Fetch.
   * @param serverAddress InetAddress - represents the Server address.
   * @param serverTCPportNumber int - represents the Server's TCP port to use in OWAMP-Control communications.
   * @param dataRS RequestSessionToBeSend - RequestSession's datagram.
   * @param socketToServer Socket - socket to Server to be used in OWAMP-Control communications with Server.
   * @param systemInReader BufferedReader - Used to implement the System.in.
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the time.
   */

  public OWAMP_FetchClient(InetAddress serverAddress, int serverTCPportNumber, RequestSessionToBeSend dataRS, Socket socketToServer, BufferedReader systemInReader, HiResTimer hiResTimer){

    this.serverAddress = serverAddress;
    this.serverTCPportNumber = serverTCPportNumber;
    buffer = new byte[DEFAULT_PACKET_LENGTH];
    arrayOfdataRS = new RequestSession[1];
    arrayOfdataRS[0] = new RequestSession(dataRS.mbz_iPVN, dataRS.confSender, dataRS.confReceiver, dataRS.numberOfScheduleSlots, dataRS.numberOfPackets, dataRS.senderPort, dataRS.receiverPort, dataRS.senderAddress, dataRS.receiverAddress, dataRS.sid, dataRS.paddingLenght, dataRS.startTime, dataRS.timeout, dataRS.typeP_Descriptor);
    this.socketToServer = socketToServer;
    this.systemInReader = systemInReader;

    // Test fetchClientDir directory. If it doesn't exists it should be created.
    if(!fetchClientDir.exists())
      fetchClientDir.mkdir();

    // Test packet results directory. If it doesn't exists it should be created.
    if(!packetResults.exists())
      packetResults.mkdir();

    // Test test results directory. If it doesn't exist it should be created.
    if(!testResults.exists())
      testResults.mkdir();

    this.hiResTimer = hiResTimer;

    try{
      serverSocket = new ServerSocket();    // Used to verify if a socket connection is closed
                                            // using method isSocketConnected.
    }catch(IOException e){
      System.err.println(e);
    }

  }


  // Instance variables

  private Socket socketToServer;                 // The server's socket for Fetch-Session.
  private InetAddress serverAddress;             // The Server's address.
  private int serverTCPportNumber;               // The Server's port for TCP connection.
  private Control_AckReceiver dataCA = new Control_AckReceiver();
  private RequestSession[] arrayOfdataRS;           /* Array of datagrams Request Session (it has
                                                     * all request session made by the client).
                                                     * The max number of Request Session that the
                                                     * client can make is specified by the constant
                                                     * numMaxRS.
                                                     */
  private ScheduleSlotDescriptions[][] arrayOfdataSSD; // For each Request Session made by the client
                                                               // we have an array of ScheduleSlotDescriptions.
  private static File clientRSdir = new File("clientRSdir");   // The directory of all RS made using the Control-Client.

  private int option;         // The user's option.
  private short numRS;        // Number of Request Sessions made by the client. It's automatically incremented
                              // for each new Request Sessions made.
  private File fetchClientDir = new File("fetchClientDir");  // The directory of all results of this Fetch-Client.

  private File packetResults = new File("fetchClientDir"+File.separator+"fetchPacketRecords"); // The directory of packet records received from
                                                                                               // Server.
  private File testResults = new File("fetchClientDir"+File.separator+"fetchTestResults");     // The directory of file test results created with
                                                                                               // the information received in packet records.

  private boolean showMenu;           // Used to test if the Menu should be used.
  private boolean closeResultFile;
  private short resultsOption = 0;
  private short saveOption = 0;
  private boolean intervalConfianca;  // Used to determine if a 90% confidence interval is requested.
  private PrintStream resultFile;     // Is the file to save results to.
  private File resultFileName;
  private double[][] arrayDelay;      // Each element of this array represents average delay for each
                                      // session test in this interval of tests.
  private double[] arrayDuplicates;   // Each element of this array represents the number of duplicates packets in the requested interval.
  private double[] arrayLossATO;      // Each element of this array represents average loss after timeOut for each
                                      // session test in this interval of tests.
  private double[] arrayLossNR;       // Each element of this array represents average loss (not received) for each
                                      // session test in this interval of tests.
  private double[] arrayThroughput;   // Each element of this array represents average Throughput for each
                                      // session test in this interval of tests.
  private short arrayPos;
  private Calendar[][] arrayDataTime; // It will be used to represent the time of first sent and last packet sent, for a
                                      // test session.

  private BufferedReader systemInReader;    // Used to implement the System.in.


  // Instance methods

  /**
   * This Method prints the results menu on the screen.
   */

  public void showResultsMenu(){

    boolean test = true;  // To test if the option entered is ok.
    String line = "";     // Used to receive the string read from keybord.

    do{
      System.out.println("What results do you want to see? (Default 0)");
      System.out.println(" ");
      System.out.println("0  -  Individual Packets and Global statistics");
      System.out.println("1  -  Only Global statistics");
      System.out.println(" ");
      System.out.print("Choose one of the options: ");

      try {
        line = systemInReader.readLine().trim();
        option = Integer.parseInt(line);
      }catch(NumberFormatException e){
        if(line.length() == 0) // 0 is the default option  - Individual Packets and Global statistics
          option = 0;
        else{
          System.err.println(e);
          option = -1;
        }
      }catch (IOException e) {
        System.err.println(e);
        return;
      }

      if((option >= 0) && (option <= 1))
        test = false;
      else{
        System.out.println("You must choose a valid option.");
        System.out.println("Press Enter to continue...");
        try {
          systemInReader.readLine();
        }
        catch (IOException e) {
          System.err.println(e);
        }

      }
    }while(test == true);

  }


  /**
   * This Method prints the save option menu on the screen.
   */

  public void showSaveOptionMenu(){

    // Test fetchClientDir directory. To prevent errors if a user deletes it. If it doesn't exist it should be created.

    if (!fetchClientDir.exists())
      fetchClientDir.mkdir();

    // Test packetResults directory. To prevent errors if a user deletes it. If it doesn't exist it should be created.

    if (!packetResults.exists())
      packetResults.mkdir();

    // Test test results directory. To prevent errors if a user deletes it, if it doesn't exist it should be created.

    if (!testResults.exists())
      testResults.mkdir();

    boolean test = true;  // To test if the option entered is ok.
    String line = "";     // Used to receive the string read from keybord.

    do{
      System.out.println("Choose one of the options (Default 0): ");
      System.out.println(" ");
      System.out.println("0  -  Only show results on the screen.");
      System.out.println("1  -  Show results on the screen and save to a file.");
      System.out.println("2  -  Just save results to a file.");

      try {
        try {
          line = systemInReader.readLine().trim();
          option = Integer.parseInt(line);
        }catch(NumberFormatException e){
          if(line.length() == 0) // 0 is the default option  - Only show results on the screen.
            option = 0;
          else{
            System.err.println(e);
            option = -1;
          }
        }

        if((option >= 0) && (option <= 2)){
          test = false;
          if(option!=0){
            System.out.print("\nIntroduce the file name to save results to (if it exists it will be deleted): ");
            resultFileName = new File(testResults.getPath()+testResults.separator+systemInReader.readLine().trim()+".txt");
            System.out.println();
            if(resultFileName.exists())
              resultFileName.delete();
            FileOutputStream  out = new FileOutputStream(resultFileName);
            resultFile = new PrintStream( out );
          }
        }else{
          System.out.println("You must choose a valid option.");
          System.out.println("Press Enter to continue...");
          systemInReader.readLine();
        }
      }catch (IOException e) {
        System.err.println(e);
        return;
      }

    }while(test == true);

  }


  /**
   * This method is used to process a Fetch Session to the Server. It is used to ask the user which
   * session to fetch and the interval of packets to fetch.
   * @param readArrayOfRSfromFiles boolean - Test if the arrayOfRS should be read from the files in clientRSdir.
   * When this class is used to fetch the test results in the OWAMP_ControlClient all the RS are in the
   * arrayOfdataRS variable, so there is no need to read them from the clientRSdir. Also this directory may have
   * other RS files of others instances of the OWAMP_ControlClient class. When this method is used to fetch the
   * test results in the OWAMP_ControlClient these sessions shouldn't appear in the Fetch-Session's option list.
   * @return boolean -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean processFetchSession(boolean readArrayOfRSfromFiles){ // process Fetch-Session

    // Test fetchClientDir directory. To prevent errors if a user deletes it. If it doesn't exist it should be created.

    if(!fetchClientDir.exists())
      fetchClientDir.mkdir();

    // Test packetResults directory, to prevent errors if a user deletes it. If it doesn't exist it should be created.

    if(!packetResults.exists())
      packetResults.mkdir();

    // Test test results directory, to prevent errors if a user deletes it. If it doesn't exist it should be created.

    if(!testResults.exists())
      testResults.mkdir();


    // Process Fetch Session

    int i = 0, j = 0;

    option = 0;
    int c;
    String senderStrAddress = "";  // Sender's address string representation.
    int maxLength = 0;             // Length of the biggest Sender's address string representation.
    String line = "";     // Used to receive the string read from keybord.

    try{
      do{

        if(readArrayOfRSfromFiles){  // Test if the arrayOfRS should be read from the files in clientRSdir.
                                     // When this method is used to fetch the test results in the OWAMP_ControlClient
                                     // all the RS are in the arrayOfdataRS variable, so there is no need to read them
                                     // from the clientRSdir. Also this directory may have other RS files of others instances
                                     // of the OWAMP_ControlClient class. These sessions shouldn't appear in the Fetch-Session's
                                     // option list.
          // We should read the arrayOfRS again beacuse it can be different now.
          if(!arrayOfRSfromFiles()){  // Read the arrayOfdataRS from a files in clientRSdir. This directory has all the RS accepted by the
                                      // Sender and Receiver of test packets.(So, we are sore that these sessions were
                                      // started)
            System.err.println("Error: Error processing Fetch Session operation.");
            return false;
          }
        }

        if(numRS==0){
          System.out.println("There are no sessions to fetch.");
          System.out.println("Press any key to try again or F to finish...");
          option = -1;
          try {
            if(systemInReader.readLine().trim().compareToIgnoreCase("f") == 0)
              return false;
          }
          catch (IOException e) {
            System.err.println(e);
          }
        }else{

          j = 0;
          System.out.println("Choose which session do you want to fetch: ");
          System.out.println();

          maxLength = 0;
          for (i = 0; i < numRS; i++) {
            senderStrAddress = determineSeReAddress(arrayOfdataRS[i].mbz_iPVN,
                arrayOfdataRS[i].senderAddress).getHostName();
            maxLength = (senderStrAddress.length() > maxLength) ?
                senderStrAddress.length() : maxLength;
          }
          j = 0;
          System.out.print("index\tSender Address");
          for (c = 0; c < (maxLength - "Sender Address".length()); c++)
            System.out.print(" ");
          System.out.print("\t-->\tReceiver Address\n\n");
          for (i = 0; i < numRS; i++) {
            // Determine sender address
            senderStrAddress = determineSeReAddress(arrayOfdataRS[i].mbz_iPVN,
                arrayOfdataRS[i].senderAddress).getHostName();
            System.out.print( (j + 1) + "\t" + senderStrAddress);
            for (c = 0; c < (maxLength - senderStrAddress.length()); c++)
              System.out.print(" ");
            System.out.print("\t-->\t" +
                             determineSeReAddress(arrayOfdataRS[i].mbz_iPVN,
                                                  arrayOfdataRS[i].
                                                  receiverAddress).getHostName() +
                             "\n");

            j++;
          }

          if (j == 0) {
            System.err.println("Error: There is no session to fetch.");
            return false;
          }

          System.out.println( (j + 1) + "\tFetch all sessions.  ");
          System.out.println( (j + 2) +
                             "\tDetermine 90% confidence intervals. ");
          System.out.println( (j + 3) + "\tFinish.  ");

          try {
            line = systemInReader.readLine().trim();
            option = Integer.parseInt(line);
          }
          catch (NumberFormatException e) {
            if(line.length()!=0)
              System.err.println(e);
            option = -1;
          }

          if (option == (j + 3)) { // Finish
            return false;
          }

          if (option <= 0) { // Option should be greater than zero
            System.err.println("The option introduced is not valid.");
          }

          if (option > (j + 2)) { // Not a valid option
            System.err.println("Error: You must choose a valid option.");
          }
        }
      }while ((option > (j + 2))||(option <= 0));

      showMenu = true;        // Show results and Save Menu

      if(option != (j+2)){
        intervalConfianca = false;

        // Ask for the interval of packets that the user wants to fetch.

        int beginSeq = 0, endSeq = 0;

        do {
          System.out.println("which interval of packets do you want to fetch?");
          System.out.println("(Note: If Begin Seq is zero and End Seq is -1 (0xFFFFFFFF), complete session is said to  be requested.)");
          System.out.println();

          do {
            System.out.print("Begin Seq (Default 0): ");
            try {
              line = systemInReader.readLine().trim();
              beginSeq = Integer.parseInt(line);
            }catch (NumberFormatException e) {
              if (line.length() == 0) // 0 is the default option  - Default 0
                beginSeq = 0;
              else {
                System.err.println(e);
                beginSeq = -1;
              }
            }

            if (beginSeq == -1)
              System.err.println("The beginSeq introduced is not valid.");
          }while (beginSeq == -1);
          System.out.println();

          System.out.print("End Seq (Default -1): ");

          try {
            line = systemInReader.readLine().trim();
            endSeq = Integer.parseInt(line);

            if ( (beginSeq < 0) )
              System.err.println("Error: Begin Seq should be greater than zero.");

            if((endSeq < 0) && (endSeq != -1))
              System.err.println("Error: End Seq should be greater or equal to -1.");
            else
              if ( (beginSeq > endSeq) && (endSeq != -1))
                System.err.println("Error: End Seq should be greater than Begin Seq.");

          }catch(NumberFormatException e){
            if(line.length() == 0) // 0 is the default option  -  (-1)
              endSeq = -1;
            else{
              System.err.println(e);
              endSeq = -2;
              System.err.println("The endSeq introduced is not valid.");
            }
          }
          System.out.println();
        }while ( (beginSeq < 0) || ( (endSeq < 0) || (beginSeq > endSeq)) && (endSeq != -1));

        int begin = 0;
        int end = option;
        if (option != (j + 1))     // If options is different from fetch all sessions.
          begin = end - 1;
        else
          end = option - 1;


        // Start each Fetch Session

        closeResultFile = false;
        for (i = begin; i < end; i++) {
          if(i==(end-1))
            closeResultFile = true;

          if (!fetchSession(beginSeq, endSeq, i)) { // FetchSession(int beginSeq, int endSeq, int indexOfSession)
            System.err.println("Error: Error in processing Fetch Session n " + (i + 1) + ".");
            System.err.println();
            System.err.println("  Possible errors:");
            System.err.println();
            System.err.println("  i)   If a test session is still in progress and a complete Fetch Session is requested, the receiver should ");
            System.err.println("       reject the requested of FetchSession (we can not fetch the complete results of a session that is still ");
            System.err.println("       operating).");
            System.err.println("  ii)  The range of Test Packets requested doesn't exists or the receiver haven't received them yet.");
            System.err.println("  iii) SessionReceiver file doesn't exist because the receiver received a not valid Stop Session packet or a ");
            System.err.println("       StopSession packet with accept = 1. In these cases it invalidates the results of the Test Session.");
            System.err.println("  iv)  The results have been deleted because a complete session has been asked previously.");
            System.err.println("  v)   The Receiver or the Server has been restarted or doesn't exits.");
            System.err.println();

            // If all results are requested and the test session is finished, its RS file should be deleted.
            if( (socketToServer != null)&&(dataCA.accept != 0)&&(beginSeq == 0 && endSeq == -1) && ((arrayOfdataRS[i].startTime.timeInMillisecondsSince1970() + totalTimeToWait(arrayOfdataRS[i], arrayOfdataSSD[i])) < hiResTimer.milliTime())){
              File sessionFileName = new File(clientRSdir.getPath()+clientRSdir.separator+owampHexEncode(arrayOfdataRS[i].sid)+".rs");  // If session file exists, delete it.
              if(sessionFileName.exists())
                sessionFileName.delete();
            }

            if (socketToServer == null)
              return false;

            closeSocket(socketToServer);

            return true;  // We should not end now because, we have an error for this session but
                          // we can have other sessions to fetch.
          }
          showMenu = false;  // Don't Show results and Save Menu
          closeSocket(socketToServer);
        }
      }else{  // Determine the 90% confidence interval
        intervalConfianca = true;
        intervalConfianca();
      }

      closeSocket(socketToServer);

      return true;

    }
    catch (IOException e) {
      System.err.println(e);
      closeSocket(socketToServer);
      return false;
    }finally{
      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();
    }

  }


  /**
   * This method is used to process the Fetch Session and prints the result in ecran. It receives the
   * interval of packets to fetch and the index of the session in array arrayOfdataRS to be fetched. Note
   * that this array is read from a file which have all sessions that were really started.
   * @param beginSeq int - the first packet sequence number to Fetch.
   * @param endSeq int - the last packet sequence number to Fetch.
   * @param indexOfSession int - index of the RS in arrayOfdataRS representing the session to fetch.
   * @return boolean -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean fetchSession(int beginSeq, int endSeq, int indexOfSession){ // process Fetch-Session

    // Test fetchClientDir directory. To prevent errors if a user deletes it. If it doesn't exist it should be created.

    if (!fetchClientDir.exists())
      fetchClientDir.mkdir();

    // Test packetResults directory. To prevent errors if a user deletes it. If it doesn't exist it should be created.

    if (!packetResults.exists())
      packetResults.mkdir();

    // Test test results directory. To prevent errors if a user deletes it, if it doesn't exist it should be created.

    if (!testResults.exists())
      testResults.mkdir();

    // Test socket connection to Server

    if(socketToServer.isClosed() || !socketToServer.isConnected()){
      // Process ConnectionSetup with Server for fetch operation
      int mode = 1;

      // We should set the socket to server's timeout to 2 * TIMEOUT because the server waits
      // for TIMEOUT ms to receive any packet from Receiver, and then it sends it to Fetch-Client.
                                  // connectionSetup(int mode, InetAddress ia, int port, int timeout)
      socketToServer = connectionSetupRequest(mode,serverAddress, this.serverTCPportNumber,2*TIMEOUT);

      if (socketToServer == null) { // process connectionSetup(int modes)
        System.err.println("Error: Error in processing Connection Setup.");
        return false;
      }

    }


    // Send the Fetch-Session packet to Server

    byte[] sid = new byte[16];
    System.arraycopy(arrayOfdataRS[indexOfSession].sid,0,sid,0,sid.length);

    FetchSessionToBeSend dataFS = new FetchSessionToBeSend(beginSeq,endSeq,sid);
    dataFS.sendFS(socketToServer);

    // Receive the Control-Ack packet

    dataCA = new Control_AckReceiver();
    if(!dataCA.receiveCA(socketToServer)){
      System.err.println("Error: Error in receiving the packet Control-Ack");
      closeSocket(socketToServer);
      return false;
    }


    switch(dataCA.accept){    // testing the accept value
      case 1:
        closeSocket(socketToServer);
        System.err.println("Error: The Fetch-Session request was rejected.");
        System.err.println();
        return false;
      case 0:   // If Accept was 0, the server then will send the OWAMP-Test session
                // data in question, followed by 16 octets of Integrity Zero Padding

        // Receive the OWAMP-Test session data from the session receiver

        OTSessionDataReceiver dataSD = new OTSessionDataReceiver();
        if(!dataSD.receiveCompleteOTSessionData(socketToServer)){
          System.err.println("Error: Error receiving the OWAMP-Test session data from the session receiver.");
          closeSocket(socketToServer);
          return false;
        }

        // Receive each PacketRecords from Server and record them in result file, in packetResults directory.
        // All packet record must be received, otherwise the results must be invalidated.

        String strSid = owampHexEncode(dataFS.sid);

        File outputFileName = new File(packetResults.getPath()+packetResults.separator+strSid+".owp");
        if(outputFileName.exists())
          outputFileName.delete();

        OwampTestPacketRecordReceiver packetRecord = new OwampTestPacketRecordReceiver();

        try{

          RandomAccessFile outputFile = new RandomAccessFile(outputFileName,"rw");

          // Receive each PacketRecords from Server

          for (int i=0;i<dataSD.numberOfPacketRecords;i++) {    // For each one of the packet record

            if(!packetRecord.receivePR(socketToServer)){
              System.err.println("Error: Error receiving packet PacketRecord "+(i+1)+" in operation Fetch session.");
              outputFile.close();
              outputFileName.delete();
              return false;  // All the packet records must be received properly.
            }

            outputFile.writeInt(packetRecord.sequenceNumber);
            outputFile.writeInt(packetRecord.sendTimestamp.integerPartOfSeconds);
            outputFile.writeInt(packetRecord.sendTimestamp.fractionalPartOfSeconds);
            outputFile.writeByte(packetRecord.sendError.s_z_scale);
            outputFile.writeByte(packetRecord.sendError.multiplier);
            outputFile.writeInt(packetRecord.receiveTimestamp.integerPartOfSeconds);
            outputFile.writeInt(packetRecord.receiveTimestamp.fractionalPartOfSeconds);
            outputFile.writeByte(packetRecord.receiveError.s_z_scale);
            outputFile.writeByte(packetRecord.receiveError.multiplier);
            //       outputFile.writeByte(255);   // TTL value

          }

          // Receive the OWAMP-Test session data IntegrityZeroPadding

          // Receive the IntegrityZeroPadding packet

          IntegrityZeroPaddingReceiver izp2 = new IntegrityZeroPaddingReceiver(16);
          if(!izp2.receiveIZP(socketToServer)){
            System.err.println("Error: Error receiving the OWAMP-Test session data IntegrityZeroPadding in operation Fetch session.");
            outputFile.close();
            outputFileName.delete();
            return false;
          }

          outputFile.close();   // Close Packet Records results file.


          // Print Results

          // Determine sender and receiver address.

          int numberOfPacketRecords = dataSD.numberOfPacketRecords; // number of packet records that will follow.

          InetAddress senderAddress = determineSeReAddress(arrayOfdataRS[indexOfSession].mbz_iPVN, arrayOfdataRS[indexOfSession].senderAddress);
          InetAddress receiverAddress = determineSeReAddress(arrayOfdataRS[indexOfSession].mbz_iPVN, arrayOfdataRS[indexOfSession].receiverAddress);

          Calendar dataS = Calendar.getInstance(); // Get actual time. It will be used to represent sendTimestamp.
          Calendar dataR = Calendar.getInstance(); // Get actual time. It will be used to represent receiveTimestamp.
          Calendar dataRSince1900 = Calendar.getInstance(); // Get actual time. It will be used to represent receiveTimestampSince1900.

          double delay = 0;       // Delay between send and receive timestamp for each packet record.

          double sumDelay = 0;  // Sum of all delays.
          double minDelay = 0;    // Min delay.
          double maxDelay = 0;    // Max delay.
          int packetsRBTO = 0;  // Packets received before timeout
          double startTime = 0;   // Time when the first packet was sent
          double endTime = 0;     // Time when the last packet was sent


          if((!intervalConfianca)&&(showMenu)){
            showResultsMenu();
            resultsOption = (short) option;
            showSaveOptionMenu();
            saveOption = (short) option;
          }

          if ((!intervalConfianca)&&(resultsOption==0)) {// if the user wants to see Individual Packets
            if((saveOption == 0) || (saveOption == 1)){
              // if 0:  Only show results on the screen.
              // if 1: Show results on the screen and save to a file.
              System.out.println();
              System.out.println("Packets Received:" + numberOfPacketRecords);
              System.out.println();
            }

            if((saveOption == 1) || (saveOption == 2)){  // Save results to a file.
              resultFile.println ("Packets Received:" + numberOfPacketRecords);
              resultFile.println();
            }
          }

          RandomAccessFile inFile = new RandomAccessFile(outputFileName,"r");
          long filePointer = inFile.getFilePointer();  // Pointer to beginning of PacketRecord information
                                                       // in file.
          int duplicates = 0, sequenceNumber = 0;      // Used to compute the number of duplicated packets

          for(int i=0;i<numberOfPacketRecords;i++){ // For each PacketRecord
            inFile.seek(filePointer+i*packetRecord.PACKET_LENGTH);   // Put the file pointer in start of file.
            packetRecord.sequenceNumber=inFile.readInt();
            packetRecord.sendTimestamp.integerPartOfSeconds=inFile.readInt();
            packetRecord.sendTimestamp.fractionalPartOfSeconds=inFile.readInt();
            packetRecord.sendError.s_z_scale=inFile.readByte();
            packetRecord.sendError.multiplier=inFile.readByte();
            packetRecord.receiveTimestamp.integerPartOfSeconds=inFile.readInt();
            packetRecord.receiveTimestamp.fractionalPartOfSeconds=inFile.readInt();
            packetRecord.receiveError.s_z_scale=inFile.readByte();
            packetRecord.receiveError.multiplier=inFile.readByte();

            dataS.setTimeInMillis((long)packetRecord.sendTimestamp.timeInMillisecondsSince1970());
            dataR.setTimeInMillis((long)packetRecord.receiveTimestamp.timeInMillisecondsSince1970());
            dataRSince1900.setTimeInMillis((long)packetRecord.receiveTimestamp.timeInMillisecondsSince1900());

            // Store time when first and last packets were sent to compute time interval of send session.

            if (i==0)
              startTime = endTime = dataS.getTimeInMillis(); // Time when the first packet was sent;
            else {
              if (packetRecord.sendTimestamp.timeInMillisecondsSince1970() < startTime)
                startTime = packetRecord.sendTimestamp.timeInMillisecondsSince1970();
              if (packetRecord.sendTimestamp.timeInMillisecondsSince1970() > endTime)
                endTime = packetRecord.sendTimestamp.timeInMillisecondsSince1970();
            }

            if (dataRSince1900.getTimeInMillis()!=0) {       // The packet was received before timeout
              packetsRBTO++;

              // Compute delay in ms
              delay = (packetRecord.receiveTimestamp.timeInMicrosecondsSince1970()-packetRecord.sendTimestamp.timeInMicrosecondsSince1970())/1000;
              sumDelay = sumDelay + Math.abs(delay);

              if ((!intervalConfianca)&&(resultsOption==0)){ // if the user wants to see Individual Packets
                if((saveOption == 0) || (saveOption == 1)){
                  // if 0:  Only show results on the screen.
                  // if 1: Show results on the screen and save to a file.
                  System.out.println("Packet with sequence number "+packetRecord.sequenceNumber+" sent in "+ dataS.getTime()+" and received in "+dataR.getTime()+"  Delay: "+delay+" ms");
                }

                if((saveOption == 1) || (saveOption == 2)){  // Save results to a file.
                  resultFile.println("Packet with sequence number "+packetRecord.sequenceNumber+" sent in "+ dataS.getTime()+" and received in "+dataR.getTime()+"  Delay: "+delay+" ms");
                }
              }
                // Compute the minimum and max delay. We should use absolute value because delay may be negative if clock isn't synchronized.
              if (i==0)
                minDelay = maxDelay = Math.abs(delay);
              else {
                if (Math.abs(delay) <  minDelay)
                  minDelay = Math.abs(delay);
                if (Math.abs(delay) > maxDelay)
                  maxDelay = Math.abs(delay);
              }
            }

            inFile.seek(filePointer);  // Return to the beginning of file and calculate the number of duplicate packets.

            for(int c=0; c < numberOfPacketRecords;c++){
              sequenceNumber=inFile.readInt();
              if (c!=i) {
                if (sequenceNumber == packetRecord.sequenceNumber)
                  duplicates++;
              }
              inFile.skipBytes(packetRecord.PACKET_LENGTH - 4);
            }
          }


          // Calculate the number of packet received and not received

          int packetsRATO = numberOfPacketRecords - packetsRBTO; // packets lost (received after timeout)
          int packetsNR = 0;               // packets lost (not received)
          if ((dataFS.beginSeq == 0x0)) {  // a complete session was requested
            packetsNR = (((dataFS.endSeq == 0xFFFFFFFF)||((dataFS.endSeq+1) > arrayOfdataRS[indexOfSession].numberOfPackets))?(arrayOfdataRS[indexOfSession].numberOfPackets):(dataFS.endSeq+1)) - (numberOfPacketRecords-duplicates);
          }
          else {                           // only a fraction of the session was requested
            if(((dataFS.endSeq+1) < arrayOfdataRS[indexOfSession].numberOfPackets))
              packetsNR = (dataFS.endSeq+1 - dataFS.beginSeq) - (numberOfPacketRecords-duplicates);
            else
              packetsNR = (arrayOfdataRS[indexOfSession].numberOfPackets - dataFS.beginSeq) - (numberOfPacketRecords-duplicates);
          }

          // compute throughput generated by OWAMP during this interval
          int testPacketLength = ((dataSD.dataRS.paddingLenght + StreamPackets_Unauthenticated.PACKET_LENGTH) < _OWP_TESTREC_SIZE)?(_OWP_TESTREC_SIZE):(dataSD.dataRS.paddingLenght + StreamPackets_Unauthenticated.PACKET_LENGTH);
          int numberOfSentPackets = (endSeq != -1)?numberOfPacketRecords:dataSD.dataRS.numberOfPackets;
          double throughput = (testPacketLength*numberOfSentPackets)/((double)((((endTime-startTime)/1000)==0)?1:((endTime-startTime)/1000)));

          // Calculate the variance

          if ((packetsRBTO!=0)&&(!intervalConfianca)) {

            double aux = 0;   // Used in the calculation of delay variance
            double aux1 = 0;

            inFile.seek(filePointer);
            for(int i=0; i < numberOfPacketRecords;i++) {
              packetRecord.sequenceNumber=inFile.readInt();
              packetRecord.sendTimestamp.integerPartOfSeconds=inFile.readInt();
              packetRecord.sendTimestamp.fractionalPartOfSeconds=inFile.readInt();
              packetRecord.sendError.s_z_scale=inFile.readByte();
              packetRecord.sendError.multiplier=inFile.readByte();
              packetRecord.receiveTimestamp.integerPartOfSeconds=inFile.readInt();
              packetRecord.receiveTimestamp.fractionalPartOfSeconds=inFile.readInt();
              packetRecord.receiveError.s_z_scale=inFile.readByte();
              packetRecord.receiveError.multiplier=inFile.readByte();

              dataS.setTimeInMillis((long)packetRecord.sendTimestamp.timeInMillisecondsSince1970());
              dataR.setTimeInMillis((long)packetRecord.receiveTimestamp.timeInMillisecondsSince1970());
              dataRSince1900.setTimeInMillis((long)packetRecord.receiveTimestamp.timeInMillisecondsSince1900());

              if (dataRSince1900.getTimeInMillis()!=0) { // The packet was received before timeout
                // Compute delay in ms
                delay = Math.abs((packetRecord.receiveTimestamp.timeInMicrosecondsSince1970()-packetRecord.sendTimestamp.timeInMicrosecondsSince1970())/1000);

                aux = aux + Math.pow((double)(delay - (sumDelay / packetsRBTO)),(double)2);  // Used in delay variance
                aux1 = aux1 + Math.abs(delay - (sumDelay / packetsRBTO));
              }
            }

            // Compute delay variance

            double variance = aux/(((packetsRBTO-1)>0)?(packetsRBTO-1):1);    // Variance
            BigDecimal a = new BigDecimal(variance);
            double meanDev = aux1/packetsRBTO;        // Mean Deviation
            BigDecimal b = new BigDecimal(meanDev);
            double standardDev = Math.sqrt(variance); // Standard Deviation
            BigDecimal c = new BigDecimal(standardDev);

            // print statistics

            BigDecimal d = new BigDecimal(sumDelay/packetsRBTO);
            BigDecimal t = new BigDecimal(throughput);

            if((saveOption == 0) || (saveOption == 1)){
              // if 0: Only show results on the screen.
              // if 1: Show results on the screen and save to a file.
              System.out.println();
              System.out.println("Statistics between machine "+senderAddress.getHostAddress()+" and "+receiverAddress.getHostAddress()+":");
              System.out.println();
              System.out.println("Number of active measurement packets requested to be sent in this OWAMP-Test Session = "+arrayOfdataRS[indexOfSession].numberOfPackets);
              System.out.println("Interval of Packets requested: ["+dataFS.beginSeq+","+dataFS.endSeq+"]");
              System.out.println("Results for the requested interval:  Received = "+(packetsRBTO+packetsRATO)+", Lost: Received after timeout/Not received = "+packetsRATO+"/"+packetsNR+", Total lost = "+(packetsRATO+packetsNR)+", Duplicates = "+duplicates);
              System.out.println("Average delay: "+(d.setScale(3,BigDecimal.ROUND_HALF_UP))+" ms, Min delay: "+minDelay+" ms, Max delay: "+maxDelay+" ms");
              System.out.println("Delay Variation (Jitter): "+(a.setScale(1,BigDecimal.ROUND_HALF_UP))+" ms, Mean Deviation: "+(b.setScale(1,BigDecimal.ROUND_HALF_UP))+" ms, Standard Deviation: "+(c.setScale(1,BigDecimal.ROUND_HALF_UP))+" ms");
              System.out.println("Throughput generated by OWAMP: "+(t.setScale(1,BigDecimal.ROUND_HALF_UP))+" bytes/s");
              System.out.println();
            }

            if((saveOption == 1) || (saveOption == 2)){  // Save results to a file.
              resultFile.println();
              resultFile.println("Statistics between machine "+senderAddress.getHostAddress()+" and "+receiverAddress.getHostAddress()+":");
              resultFile.println();
              resultFile.println("Number of active measurement packets requested to be sent in this OWAMP-Test Session = "+arrayOfdataRS[indexOfSession].numberOfPackets);
              resultFile.println("Interval of Packets requested: ["+dataFS.beginSeq+","+dataFS.endSeq+"]");
              resultFile.println("Results for the requested interval:  Received = "+(packetsRBTO+packetsRATO)+", Lost: Received after timeout/Not received = "+packetsRATO+"/"+packetsNR+", Total lost = "+(packetsRATO+packetsNR)+", Duplicates = "+duplicates);
              resultFile.println("Average delay: "+(d.setScale(3,BigDecimal.ROUND_HALF_UP))+" ms, Min delay: "+minDelay+" ms, Max delay: "+maxDelay+" ms");
              resultFile.println("Delay Variation (Jitter): "+(a.setScale(1,BigDecimal.ROUND_HALF_UP))+" ms, Mean Deviation: "+(b.setScale(1,BigDecimal.ROUND_HALF_UP))+" ms, Standard Deviation: "+(c.setScale(1,BigDecimal.ROUND_HALF_UP))+" ms");
              resultFile.println("Throughput generated by OWAMP: "+(t.setScale(1,BigDecimal.ROUND_HALF_UP))+" bytes/s");
              if(closeResultFile)
                resultFile.close();
              System.out.println("Results saved to file "+resultFileName.getCanonicalPath());
              System.out.println();
            }

          }
          else
            if((packetsRBTO == 0)&&(!intervalConfianca)){
              if((saveOption == 0) || (saveOption == 1)){
                // if 0: Only show results on the screen.
                // if 1: Show results on the screen and save to a file.
                System.out.println("Number of active measurement packets requested to be sent in this OWAMP-Test Session = "+arrayOfdataRS[indexOfSession].numberOfPackets);
                System.out.println("Interval of Packets requested: ["+dataFS.beginSeq+","+dataFS.endSeq+"]");
                System.out.println("Results for the requested interval:  Received = "+(packetsRBTO+packetsRATO)+", Lost: Received after timeout/Not received = "+numberOfPacketRecords+"/"+0+", Total lost = "+numberOfPacketRecords);
                System.out.println();
              }
              if((saveOption == 1) || (saveOption == 2)){  // Save results to a file.
                resultFile.println("Number of active measurement packets requested to be sent in this OWAMP-Test Session = "+arrayOfdataRS[indexOfSession].numberOfPackets);
                resultFile.println("Interval of Packets requested: ["+dataFS.beginSeq+","+dataFS.endSeq+"]");
                resultFile.println("Results for the requested interval:  Received = "+(packetsRBTO+packetsRATO)+", Lost: Received after timeout/Not received = "+numberOfPacketRecords+"/"+0+", Total lost = "+numberOfPacketRecords);
                if(closeResultFile)
                  resultFile.close();
                System.out.println("Results saved to file "+resultFileName.getCanonicalPath());
                System.out.println();
              }
            }

          if(intervalConfianca){
            arrayDelay[arrayPos][0] = (packetsRBTO==0)?0:sumDelay/packetsRBTO;   // average delay
            arrayDelay[arrayPos][1] = minDelay;               // minDelay
            arrayDelay[arrayPos][2] = maxDelay;               // maxDelay
            // To compute the losses for the confidence interval calculation, we should use loss ratios
            // (number of packets lost/number of packets sent)
            arrayLossATO[arrayPos] = (double)packetsRATO/(arrayOfdataRS[indexOfSession].numberOfPackets);
            arrayLossNR[arrayPos] = (double)packetsNR/(arrayOfdataRS[indexOfSession].numberOfPackets);
            // To compute the duplicates for the confidence interval calculation, we should use the ratio value
            // (number of packets duplicated/number of packets sent)
            arrayDuplicates[arrayPos] = (double)duplicates/(arrayOfdataRS[indexOfSession].numberOfPackets);
            arrayThroughput[arrayPos] = throughput;
            arrayDataTime[arrayPos][0] = Calendar.getInstance();    // Get actual time. It will be used to represent time of first sent
            arrayDataTime[arrayPos][1] = Calendar.getInstance();    // Get actual time. It will be used to represent time of last sent
            arrayDataTime[arrayPos][0].setTimeInMillis((long)startTime);  // Time of first packet sent
            arrayDataTime[arrayPos][1].setTimeInMillis((long)endTime);    // Time of last packet sent
          }

          inFile.close();  // Close packetRecord results file.
        }
        catch(IOException ex){
          System.err.println(ex);
        }

        break;

      default:
        closeSocket(socketToServer);
        System.err.println("Error: The accept value is not valid.");
        return false;
    }

    return true;

  }


  /**
   * Method used to read the arrayOfdataRS from the clientRSdir directory to be used in Fetch Session. This directory
   * is created by the Owamp-Client and it has all the RS's files with the configuration of their corresponding sessions.
   * For each RS we have a different file, which name is the session SID as a string.
   * @return boolean -  This method returns true if everything goes ok, otherwise it returns false.
   */

   public boolean arrayOfRSfromFiles(){

     // Test clientRSdir directory, to prevent errors if a user deletes it. If it doesn't exist return false.
     if (!clientRSdir.exists()){
       System.err.println("Error: There is no Sessions to Fetch (the Client's directory with all RS files doesn't exit).");
       return false;
     }

     RandomAccessFile inFile = null;    // file with the RS made.

     try{

       File strSid[] = clientRSdir.listFiles();     // Array of all RS's files in the clientRSdir directory.

       numRS=(short)(strSid.length);   // Number of RS's files recorded in directory.

       if(strSid.length == 0)  // There are no sessions to fetch
         return true;

       arrayOfdataRS = new RequestSession[numRS];
       arrayOfdataSSD = new ScheduleSlotDescriptions[numRS][];

       // read the arrayOfdataRS
       for(int i=0;i<numRS;i++){

         inFile = new RandomAccessFile(strSid[i],"r");

         arrayOfdataRS[i] = new RequestSession();

         arrayOfdataSSD[i] = readRSfromFile(arrayOfdataRS[i],inFile);
         if(arrayOfdataSSD[i] == null){
           inFile.close();
           return false;
         }

         inFile.close();
       }

     }
     catch(IOException ex){
       System.err.println(ex);
       return false;
     }

     return true;

   }


  /**
   * This method is used to calculate the 90% confidence interval of a set of test made.
   */

  public void intervalConfianca(){

    int numTest = 0;
    try{
      do{
        System.out.print("Introduce the number of tests to be considered in each interval:");
        try {
          numTest = Integer.parseInt(systemInReader.readLine().trim());
        }catch(NumberFormatException e){
          System.err.println(e);
          numTest = -1;
        }
        if((numTest < 1)||(numTest > 3600))
          System.err.println("Error: The value of number of tests read is not valid.");
      }while((numTest < 1)||(numTest > 3600));
      System.out.println();
    }
    catch (IOException e) {
      System.err.println(e);
    }

    int totalTest = numRS/numTest;   // Total number of intervals with numTest tests.

    System.out.println("It exists "+totalTest+" interval(s) with "+numTest+" tests. ");
    System.out.println();
    if(totalTest == 0){
      System.err.println("Error: There is not enough RS for the number of tests to one interval introduced.");
      return;
    }

    showSaveOptionMenu();
    short saveOption = (short) option;

    arrayDuplicates = new double[numTest];// array of all duplication ratios (number of packets duplicated/ number of packets sent)
    arrayDelay = new double[numTest][3];  // for each interval of test arrayDelay[][0] = average delay,
                                          // arrayDelay[][1] = min delay and arrayDelay[][2] = max delay.
    arrayLossATO = new double[numTest];   // array of all loss after timeout ratios (number of losses/ number of packets sent)
    arrayLossNR = new double[numTest];    // array of all not received loss ratios (number of losses/ number of packets sent)
    arrayThroughput = new double[numTest];
    arrayDataTime = new Calendar[numTest][2];

    double averageDelay = 0;
    double averageLossATO = 0;
    double averageLossNR = 0;
    double averageDuplicates = 0;
    double averageThroughput = 0;
    double varianceDelay = 0;
    double varianceLossATO = 0;
    double varianceLossNR = 0;
    double varianceDuplicates = 0;
    double varianceThroughput = 0;

    int i= 0, j = 0, pos = 0;
    BigDecimal d = new BigDecimal(0);
    BigDecimal t = new BigDecimal(0);

    for (i = 0; i < totalTest; i++){
      arrayPos = 0;
      for (j = i*numTest; j < (i+1)*numTest; j++){
        if (!fetchSession(0, -1, j)) { // FetchSession(int beginSeq, int endSeq, int indexOfSession)
          System.err.println("Error: Error in processing Fetch Session n " +
                             (j + 1) + ".");
          System.err.println();
          System.err.println("  Possible errors:");
          System.err.println();
          System.err.println("  i)   If a test session is still in progress and a complete Fetch Session is requested, the receiver should ");
          System.err.println("       reject the requested of FetchSession (we can not fetch the complete results of a session that is still ");
          System.err.println("       operating).");
          System.err.println("  ii)  The range of Test Packets requested doesn't exists or the receiver haven't received them yet.");
          System.err.println("  iii) SessionReceiver file doesn't exist because the receiver received a not valid Stop Session packet or a ");
          System.err.println("       StopSession packet with accept = 1. In these cases it invalidates the results of the Test Session.");
          System.err.println("  iv)  The results have been deleted because a complete session has been asked previously.");
          System.err.println();

          closeSocket(socketToServer);

          return;
        }
        closeSocket(socketToServer);
        arrayPos++;
      }


      if((saveOption == 0) || (saveOption == 1)){
        // if 0:  Only show results on the screen.
        // if 1: Show results on the screen and save to a file.
        System.out.println("Statistics for test interval number "+(i+1));
        System.out.println();
        System.out.println("Number of active measurement packets requested to be sent in this OWAMP-Test Session = "+arrayOfdataRS[i*numTest].numberOfPackets);
        System.out.println();
        System.out.println("           Packets  Average delay   Min delay   Max delay   Received after timeout/Not received  Duplication    Throughput                          First Sent/Last Sent");
        System.out.println("            sent       (ms)           (ms)        (ms)                 loss ratio                   ratio ");
        System.out.println();
      }

      if((saveOption == 1) || (saveOption == 2)){  // Save results to a file.
        resultFile.println("Statistics for test interval number " + (i + 1));
        resultFile.println();
        resultFile.println("Number of active measurement packets requested to be sent in this OWAMP-Test Session = "+arrayOfdataRS[i*numTest].numberOfPackets);
        resultFile.println();
        resultFile.println("           Packets  Average delay   Min delay   Max delay   Received after timeout/Not received  Duplication    Throughput                          First Sent/Last Sent");
        resultFile.println("            sent       (ms)           (ms)        (ms)                 loss ratio                   ratio ");
        resultFile.println();
      }

      // Determine average delay

      averageDelay = 0;
      averageLossATO = 0;
      averageLossNR = 0;
      averageDuplicates = 0;
      averageThroughput = 0;

      for(pos = 0;pos<numTest;pos++){
        averageDelay = averageDelay + arrayDelay[pos][0];
        averageLossATO = averageLossATO + arrayLossATO[pos];
        averageLossNR = averageLossNR + arrayLossNR[pos];
        averageDuplicates = averageDuplicates + arrayDuplicates[pos];
        averageThroughput = averageThroughput + arrayThroughput[pos];
        d = new BigDecimal(arrayDelay[pos][0]);
        t = new BigDecimal(arrayThroughput[pos]);

        if((saveOption == 0) || (saveOption == 1)){
          // if 0: Only show results on the screen.
          // if 1: Show results on the screen and save to a file.
          System.out.println((pos+1)+"  Test     "+(arrayOfdataRS[i*numTest].numberOfPackets)+"       "+ d.setScale(3,BigDecimal.ROUND_HALF_UP)+"         "+new BigDecimal(arrayDelay[pos][1]).setScale(3,BigDecimal.ROUND_HALF_UP)+"        "+new BigDecimal(arrayDelay[pos][2]).setScale(3,BigDecimal.ROUND_HALF_UP)+"                 "+new BigDecimal(arrayLossATO[pos]).setScale(2,BigDecimal.ROUND_HALF_UP)+"/"+new BigDecimal(arrayLossNR[pos]).setScale(2,BigDecimal.ROUND_HALF_UP)+"                   "+new BigDecimal(arrayDuplicates[pos]).setScale(2,BigDecimal.ROUND_HALF_UP)+"          "+t.setScale(3,BigDecimal.ROUND_HALF_UP)+"        "+arrayDataTime[pos][0].getTime()+" / "+arrayDataTime[pos][1].getTime());
        }

        if((saveOption == 1) || (saveOption == 2)){  // Save results to a file.
          resultFile.println((pos+1)+"  Test     "+(arrayOfdataRS[i*numTest].numberOfPackets)+"       "+ d.setScale(3,BigDecimal.ROUND_HALF_UP)+"         "+new BigDecimal(arrayDelay[pos][1]).setScale(3,BigDecimal.ROUND_HALF_UP)+"        "+new BigDecimal(arrayDelay[pos][2]).setScale(3,BigDecimal.ROUND_HALF_UP)+"                 "+new BigDecimal(arrayLossATO[pos]).setScale(2,BigDecimal.ROUND_HALF_UP)+"/"+new BigDecimal(arrayLossNR[pos]).setScale(2,BigDecimal.ROUND_HALF_UP)+"                   "+new BigDecimal(arrayDuplicates[pos]).setScale(2,BigDecimal.ROUND_HALF_UP)+"          "+t.setScale(3,BigDecimal.ROUND_HALF_UP)+"        "+arrayDataTime[pos][0].getTime()+" / "+arrayDataTime[pos][1].getTime());
        }

      }

      averageDelay = averageDelay/numTest;
      averageLossATO = averageLossATO/numTest;
      averageLossNR = averageLossNR/numTest;
      averageDuplicates = averageDuplicates/numTest;
      averageThroughput = averageThroughput/numTest;

      // Determine variance

      for(pos = 0;pos<numTest;pos++){
        varianceDelay = varianceDelay + Math.pow((double)(arrayDelay[pos][0] - averageDelay),(double)2);  // Used in delay variance
        varianceLossATO = varianceLossATO + Math.pow((double)(arrayLossATO[pos] - averageLossATO),(double)2); ;
        varianceLossNR = varianceLossNR + Math.pow((double)(arrayLossNR[pos] - averageLossNR),(double)2); ;
        varianceDuplicates = varianceDuplicates + Math.pow((double)(arrayDuplicates[pos] - averageDuplicates),(double)2); ;
        varianceThroughput = varianceThroughput + Math.pow((double)(arrayThroughput[pos] - averageThroughput),(double)2); ;
      }

      varianceDelay = (numTest == 1)?varianceDelay:varianceDelay/(numTest-1);
      varianceLossATO = (numTest == 1)?varianceLossATO:varianceLossATO/(numTest-1);
      varianceLossNR = (numTest == 1)?varianceLossNR:varianceLossNR/(numTest-1);
      varianceDuplicates = (numTest == 1)?varianceDuplicates:varianceDuplicates/(numTest-1);
      varianceThroughput = (numTest == 1)?varianceThroughput:varianceThroughput/(numTest-1);

      d = new BigDecimal(averageDelay);
      t = new BigDecimal(averageThroughput);

      if((saveOption == 0) || (saveOption == 1)){
        // if 0: Only show results on the screen.
        // if 1: Show results on the screen and save to a file.

        System.out.println();
        System.out.println("Global Statistics");
        System.out.println();
        System.out.println("               Average delay (ms)    Average loss Received after timeout/Not received   Average Duplication   Average Throughput (bytes/s)");
        System.out.println("                   "+d.setScale(3,BigDecimal.ROUND_HALF_UP)+"                               "+new BigDecimal(averageLossATO).setScale(2,BigDecimal.ROUND_HALF_UP)+"/"+new BigDecimal(averageLossNR).setScale(2,BigDecimal.ROUND_HALF_UP)+"                                "+new BigDecimal(averageDuplicates).setScale(2,BigDecimal.ROUND_HALF_UP)+"                    "+t.setScale(3,BigDecimal.ROUND_HALF_UP));
        System.out.print("90% intervals");
        // delay
        d = new BigDecimal(averageDelay-1.6449*Math.sqrt(varianceDelay/numTest));
        System.out.print("   ["+d.setScale(3,BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageDelay+1.6449*Math.sqrt(varianceDelay/numTest));
        System.out.print(" "+d.setScale(3,BigDecimal.ROUND_HALF_UP)+"]");
        // LossATO
        d = new BigDecimal(averageLossATO-1.6449*Math.sqrt(varianceLossATO/numTest));
        System.out.print("               ["+d.setScale(2,BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageLossATO+1.6449*Math.sqrt(varianceLossATO/numTest));
        System.out.print(" "+d.setScale(2,BigDecimal.ROUND_HALF_UP)+"]");
        // LossNR
        d = new BigDecimal(averageLossNR-1.6449*Math.sqrt(varianceLossNR/numTest));
        System.out.print(" / ["+d.setScale(2,BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageLossNR+1.6449*Math.sqrt(varianceLossNR/numTest));
        System.out.print(" "+d.setScale(2,BigDecimal.ROUND_HALF_UP)+"]");
        // Duplicates
        d = new BigDecimal(averageDuplicates-1.6449*Math.sqrt(varianceDuplicates/numTest));
        System.out.print("                      ["+d.setScale(2,BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageDuplicates+1.6449*Math.sqrt(varianceDuplicates/numTest));
        System.out.print(" "+d.setScale(2,BigDecimal.ROUND_HALF_UP)+"]");
        // Throughput
        d = new BigDecimal(averageThroughput-1.6449*Math.sqrt(varianceThroughput/numTest));
        System.out.print("            ["+d.setScale(3,BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageThroughput+1.6449*Math.sqrt(varianceThroughput/numTest));
        System.out.print(" "+d.setScale(3,BigDecimal.ROUND_HALF_UP)+"]");

        System.out.println();
        System.out.println();
        System.out.println();
      }

      d = new BigDecimal(averageDelay);
      t = new BigDecimal(averageThroughput);

      if((saveOption == 1) || (saveOption == 2)){  // Save results to a file.

        resultFile.println();
        resultFile.println();
        resultFile.println("Global Statistics");
        resultFile.println();
        resultFile.println("               Average delay (ms)    Average loss Received after timeout/Not received   Average Duplication   Average Throughput (bytes/s)");
        resultFile.println("                   "+d.setScale(3,BigDecimal.ROUND_HALF_UP)+"                               "+new BigDecimal(averageLossATO).setScale(2,BigDecimal.ROUND_HALF_UP)+"/"+new BigDecimal(averageLossNR).setScale(2,BigDecimal.ROUND_HALF_UP)+"                                "+new BigDecimal(averageDuplicates).setScale(2,BigDecimal.ROUND_HALF_UP)+"                    "+t.setScale(3,BigDecimal.ROUND_HALF_UP));

        resultFile.print("90% intervals");
        // delay
        d = new BigDecimal(averageDelay - 1.6449 * Math.sqrt(varianceDelay / numTest));
        resultFile.print("   [" + d.setScale(3, BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageDelay + 1.6449 * Math.sqrt(varianceDelay / numTest));
        resultFile.print(" " + d.setScale(3, BigDecimal.ROUND_HALF_UP) + "]");
        // LossATO
        d = new BigDecimal(averageLossATO -
                           1.6449 * Math.sqrt(varianceLossATO / numTest));
        resultFile.print("               [" + d.setScale(2, BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageLossATO +
                           1.6449 * Math.sqrt(varianceLossATO / numTest));
        resultFile.print(" " + d.setScale(2, BigDecimal.ROUND_HALF_UP) + "]");
        // LossNR
        d = new BigDecimal(averageLossNR -
                           1.6449 * Math.sqrt(varianceLossNR / numTest));
        resultFile.print(" / [" + d.setScale(2, BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageLossNR +
                           1.6449 * Math.sqrt(varianceLossNR / numTest));
        resultFile.print(" " + d.setScale(2, BigDecimal.ROUND_HALF_UP) + "]");
        // Duplicates
        d = new BigDecimal(averageDuplicates -
                           1.6449 * Math.sqrt(varianceDuplicates / numTest));
        resultFile.print("                      [" +
                         d.setScale(2, BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageDuplicates +
                           1.6449 * Math.sqrt(varianceDuplicates / numTest));
        resultFile.print(" " + d.setScale(2, BigDecimal.ROUND_HALF_UP) + "]");
        // Throughput
        d = new BigDecimal(averageThroughput -
                           1.6449 * Math.sqrt(varianceThroughput / numTest));
        resultFile.print("            [" + d.setScale(3, BigDecimal.ROUND_HALF_UP));
        d = new BigDecimal(averageThroughput +
                           1.6449 * Math.sqrt(varianceThroughput / numTest));
        resultFile.print(" " + d.setScale(3, BigDecimal.ROUND_HALF_UP) + "]");

        resultFile.println();
        resultFile.println();
        resultFile.println();
        resultFile.println();

      }
    }

    if((saveOption == 1) || (saveOption == 2)){  // Save results to a file.
      resultFile.close();
      try{
        System.out.println("Results saved to file "+resultFileName.getCanonicalPath());
      }
      catch(IOException ex){
        System.err.println(ex);
      }

      System.out.println();
    }
  }


  /**
   * This method is used to process a Fetch Session to the Server. It is used to save the results
   * to a data base.
   *
   * @param sqlServerAddress String - it's the data base server's address.
   * @param dataBase String - it's the database used to put the table with all Request Sessions.
   * @param userName String - user name.
   * @param passWord String - user password.
   * @param strSid String - is the SID of the session to be fetched.
   * @param numberOfPackets int - it's the number of packets configured to be sent, for this session.
   * @param slotParameter double - it's the slot parameter configured for this session.
   * @param realStartTime long - it's the real start time of this session.
   * @param totalTimeToWait double - it's this session's total time duration.
   * @return boolean  -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean processFetchSessionToDataBase(String sqlServerAddress, String dataBase, String userName, String passWord, String strSid, int numberOfPackets, double slotParameter /*in seconds*/, long realStartTime /*in milliseconds*/, double totalTimeToWait /*in milliseconds*/){ // process Fetch-Session

    // Test if the session has already started. If not, return.

    if(realStartTime>(long)hiResTimer.milliTime()){
      return true;
    }

    // Establish a connection to the data base.

    MySQLcom MySQLclient= null;
    Connection con = null;
    try{
      MySQLclient = new MySQLcom (sqlServerAddress, dataBase, userName, passWord);
      con = MySQLclient.getConnection();
    }catch(Exception e){
      System.out.println(e);
      return false;
    }

    ResultSet rs = null;

    // The table SID should be created. The name of the table should be equal to the sid string representation.
    // If the table SID already exists in the dataBase it means that a previous fetchSession was
    // made. In this case we should not make a new fetchSession because a fetchSession fetchs all the results of
    // a session. In this case an exception is thrown when trying to CREATE the SID table and we should return true.
    // Using the CREATE SQL statement we can guarantee that only one FetchClient element can create this table, so it
    // is created only once on the first FetchSession.

    try{  // Create SID table.
      MySQLclient.executeUpdate ( con, "CREATE TABLE `"+strSid+"` (id INT NOT NULL AUTO_INCREMENT, sequenceNumber INT, sendTimestamp DOUBLE, sendError_s_z_scale INT, sendError_multiplier INT, receiveTimestamp DOUBLE, receiveError_s_z_scale INT, receiveError_multiplier INT, TTL INT(3), delay DOUBLE /*Packet delay in milliseconds*/, PRIMARY KEY (id))" );
    }catch(Exception e){  // If the table exists, an exception is thrown.
      try{ // Verify if the exception occurred because the table SID already exits in the database. If not, an error ocurred
           // and this method should return false.
        rs = MySQLclient.executeQuery(con, "SELECT COUNT(*) FROM `"+strSid+"`"); // If the table doesn't exist, an exception is thrown
        rs.close();
      }catch(Exception e2){
        System.err.println(e);
        MySQLclient.close(con);
        return false;
      }
      MySQLclient.close(con);
      return true;
    }

    // Check if table SessionDescriptionRecords exists. If not it should be created. This table is used to save the
    // Session Description Record information of all test session fetched. Each row of this table represents a fetched
    // session and is identified by the session SID. The Session Description Record information of a session only
    // should be saved in this table when the Fetch-Session is complete. So using this table we can determine
    // if the Fetch-Session of a test session is finished or not.
    try{
      MySQLclient.executeUpdate ( con, "CREATE TABLE SessionDescriptionRecords (sid VARCHAR(32) NOT NULL, numberOfRecords int, PRIMARY KEY (sid))" );
    }catch(Exception e){  // If the table exists, an exception is thrown.
      try{  // Verify if the exception occurred because the table SessionDescriptionRecords already exits in the database. If not, an error ocurred
            // and this method should return false.
        rs = MySQLclient.executeQuery(con, "SELECT COUNT(*) FROM SessionDescriptionRecords"); // If the table doesn't exist, an exception is thrown
        rs.close();
      }catch(Exception e2){
        System.err.println(e);
        // If an error ocurres we should drop the tables SID.
        try {
          MySQLclient.executeUpdate(con, "DROP TABLE `" + strSid+"`");
        }catch (Exception e3) {
          System.err.println(e3);
          MySQLclient.close(con);
          return false;
        }
        MySQLclient.close(con);
        return false;
      }
    }

    // While the session is active ((realStartTime+totalTimeToWait)<actualTime) make FecthSessions to dataBase.

    FetchSessionToBeSend dataFS = new FetchSessionToBeSend();
    dataCA = new Control_AckReceiver();
    OTSessionDataReceiver dataSD = new OTSessionDataReceiver();
    OwampTestPacketRecordReceiver packetRecord = new OwampTestPacketRecordReceiver();
    IntegrityZeroPaddingReceiver izp = new IntegrityZeroPaddingReceiver(16);
    double delay = 0; // Delay between send and receive timestamp for each packet record.

    int numberOfEntries = 0, i = 0;
    byte[] sid = new byte[16];
    sid = owampHexDecode(strSid);
    boolean allPacketsReceived = false;
    int period = ((int)(slotParameter*1000)*2) < 2000 ? 2000 : (int)(slotParameter*1000)*2;  // Period should be minimum 2 sec.
    int beginSeq = 0, endSeq = 0;

    try{
      do{
        // Test if the table SID exists in the dataBase and read the numberOfEntries in it.

        try {
          rs = MySQLclient.executeQuery(con,"SELECT COUNT(*) FROM `" + strSid+"`"); // If the table doesn't exist, an exception
          if (rs.next())                                                       // is thrown.
            numberOfEntries = rs.getInt(1);
          rs.close();
        }catch (Exception e) {
          System.out.println(e);
          MySQLclient.close(con);
          return false;
        }

        // Determine the interval of packets that sould be fetched.

        beginSeq = numberOfEntries;
        endSeq = numberOfPackets - 1;

        // If totalTimeToWait occurs or if the session is finished before totalTimeToWait (dataFA.finished=1 for the last FS)
        // make the last FetchSession.
        if (((realStartTime + Math.round(totalTimeToWait)) < (long) hiResTimer.milliTime())) {
          beginSeq = 0;
          endSeq = -1;
          allPacketsReceived = true;
        }

        // Process Fetch Session

        // Test socket connection to Server

        if (socketToServer.isClosed() || !socketToServer.isConnected()) {

          // Process ConnectionSetup with Server for fetch operation
          int mode = 1;

          // We should set the socket to server's timeout to 2 * TIMEOUT because the server waits
          // for TIMEOUT ms to receive any packet from Receiver, and then it sends it to Fetch-Client.
                               // connectionSetup(int mode, InetAddress ia, int port, int timeout)
          socketToServer = connectionSetupRequest(mode,serverAddress,serverTCPportNumber,2*TIMEOUT);

          if (socketToServer == null) {
            System.err.println("Error: The Server specified doesn't exist.\n");
            System.err.println("Error: Error in processing Connection Setup.");
            throw new Exception();
          }
        }

        // Send the Fetch-Session packet to Server

        dataFS.fetchSessionToBeSend(beginSeq, endSeq, sid);
        dataFS.sendFS(socketToServer);

        // Receive the Fetch-Ack packet

        if (!dataCA.receiveCA(socketToServer)) {
          System.err.println("Error: Error in receiving the packet Fetch-Ack in operation Fetch Session.");
          closeSocket(socketToServer);
          throw new Exception();
        }

        switch (dataCA.accept) { // testing the accept value
          case 1:

            // In this case we should not end the operation because this FetchClient is making periodic fecthSession
            // of packetRecords not saved to dataBase yet (during session duration). In this case, if all packets in
            // a period are lost, the receiver will not accept the fetchSession but we should continue to the others
            // periods.
            System.err.println("Error: The Fetch-Session request was rejected.");
            System.err.println();
            if(allPacketsReceived)  // If this is the last fecthSession, the table strSid should be dropped.
              throw new Exception();
            break;
          case 0: // If Accept was 0, the server then will send the OWAMP-Test session
                  // data in question, followed by 16 octets of Integrity Zero Padding

            // In the last fetch session it will receive all packet records of this session
            if (allPacketsReceived) { // so we should delete the previous records received because we will receive them again.
                                      // This code should be here because if a problem occours in the last fecth session the previous
                                      // results received are preserved.
              MySQLclient.executeUpdate(con,"DELETE FROM `" + strSid + "` WHERE id > 0");
              MySQLclient.executeUpdate(con,"ALTER TABLE `" + strSid + "` AUTO_INCREMENT = 0");

            }

            // Receive the OWAMP-Test Session Data from the Server.
            // Receive the OWAMP-Test Session Data from the Server.

            if (!dataSD.receiveCompleteOTSessionData(socketToServer)) {
              System.err.println(
                  "Error: Error receiving the OWAMP-Test session data from the session receiver.");
              closeSocket(socketToServer);
              throw new Exception();
            }

            // Receive each PacketRecords from Server and record them in data base.
            // All packet record must be received, otherwise the results must be invalidated.

            for (i = 0; (i < dataSD.numberOfPacketRecords); i++) { // For each one of the packet record

              if (!packetRecord.receivePR(socketToServer)) {
                System.err.println(
                    "Error: Error receiving packet PacketRecord " + (i + 1) +
                    " in operation Fetch session.");
                throw new Exception();
              }

              // Compute delay between send and receive timestamp for each packet record.

              delay = (packetRecord.receiveTimestamp.timeInMicrosecondsSince1970() -
                       packetRecord.sendTimestamp.timeInMicrosecondsSince1970())/1000;

              MySQLclient.executeUpdate(con,
                                        "INSERT INTO `" + strSid + "` (sequenceNumber, sendTimestamp, sendError_s_z_scale, sendError_multiplier, receiveTimestamp, receiveError_s_z_scale, receiveError_multiplier, /*ttl,*/ delay) values ('" +
                                        packetRecord.sequenceNumber + "', '" +
                                        packetRecord.sendTimestamp.
                                        timeInMillisecondsSince1970() +
                                        "', '" +
                                        (int) (packetRecord.sendError.
                                               s_z_scale &
                                               0xFF) + "', '" +
                                        (int)(packetRecord.sendError.multiplier & 0xFF) +
                                        "', '" +(((long) packetRecord.receiveTimestamp.timeInMillisecondsSince1900() != 0)?
                                        packetRecord.receiveTimestamp.
                                        timeInMillisecondsSince1970():0) +
                                        "', '" +
                                        (int)(packetRecord.receiveError.s_z_scale & 0xFF) +
                                        "', '" +
                                        (int)(packetRecord.receiveError.multiplier & 0xFF) +
                                        "', '" + /*(int)(packetRecord.ttl & 0xFF) + "', '" +*/
                                        delay +
                                        "')");
            }

            // V - 16 octets of IZP.
            // Receive the OWAMP-Test session data IntegrityZeroPadding

            if ( (!izp.receiveIZP(socketToServer))) {
              System.err.println("Error: Error receiving the OWAMP-Test session data IntegrityZeroPadding in operation Fetch session from " +
                                 socketToServer.toString());
              throw new Exception();
            }

            break;
          default:
            closeSocket(socketToServer);
            System.err.println("Error: The accept value is not valid.");
            throw new Exception();
        }

        if ( (realStartTime + Math.round(totalTimeToWait)) > (long) hiResTimer.milliTime()) { // Sleep for slotParameter seconds.
          try {
            sleep(period);
          }catch (Exception e) {}
        }

        closeSocket(socketToServer);

      }while (!allPacketsReceived);

      MySQLclient.executeUpdate(con,
                          "INSERT INTO SessionDescriptionRecords (sid, numberOfRecords) values ('" +
                          strSid+"', '" +dataSD.numberOfPacketRecords+"')");


      MySQLclient.close(con);
      return true;

    }catch (Exception e) {
      System.out.println(e);
      closeSocket(socketToServer);
      // If an error ocurres during reception we should drop all results received.
      try {
        MySQLclient.executeUpdate(con, "drop table `" + strSid+"`");
        MySQLclient.close(con);
      }catch (Exception e2) {
        System.out.println(e2);
        MySQLclient.close(con);
        return false;
      }
      return false;
    }finally{
      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();
    }

  }


  // Main method

  public static void main(String[] args){

    BufferedReader systemInReader = new BufferedReader(new InputStreamReader(System.in));

    // The user can execute this module with all the configurations parameters in the command line just
    // in the case he wants to save the results to a data base. In this case the user must instroduce 10
    // parameters (String serverAddress, String sqlServerAddress, String dataBase,
    // String userName, String passWord, String strSid, String numberOfPackets, String slotParameter /*in seconds*/,
    // String realStartTime /*in milliseconds*/, String totalTimeToWait /*in milliseconds*/, String ntpServerAddress,
    // String updateInterval). The ntpServerAddress and updateInteval are optional. If not defined the system time
    // will be used to get the time reference. This mode of execution is also used in the Web interface.

    if((args.length == 10)||(args.length == 12)){

      try {

        /* @param serverAddress String - represents the Server address.
         * The server address should be in the following format:
         *
         *     IPv4_address:TCPportNumber    // for IPv4 address
         *     [IPv6_address]:TCPportNumber  // for IPv6 address
         *
         * The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
         * Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368)
         * the server address can be in the following format:
         *
         *     IPv4_address    // for IPv4 address
         *     [IPv6_address]  // for IPv6 address
         */
        String address =  args[0].trim();
        String addressAux = address;
        InetAddress serverAddress = null;
        int serverTCPportNumber = DEFAULT_SEVER_PORT;  // TCPportNumber to use in OWAMP-Control with Server.
        boolean IPv6 = false;
        int i=0;

        try{

          if(address.indexOf("[") != -1){  // IPv6 address
            IPv6 = true;
            if(address.indexOf("]:") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
              serverTCPportNumber = Integer.parseInt(address.substring(address.indexOf("]:")+2));
            }
            address = address.substring(address.indexOf("[")+1,address.indexOf("]"));
          }else{  // IPv4 address
            if(address.indexOf(":") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
              serverTCPportNumber = Integer.parseInt(address.substring(address.indexOf(":")+1));
              address = address.substring(0,address.indexOf(":"));
            }
          }

          if (serverTCPportNumber < 0 || serverTCPportNumber >= 65536) {
            System.err.println("Error: Server TCP port number out of range (" +serverTCPportNumber+").");
            System.err.println();
            System.exit(1);
          }

          InetAddress inetAddress[] = InetAddress.getAllByName(address);     // The array of all the IP addresses for a given host name.

          for(i=0;i<inetAddress.length;i++){
            if(IPv6 && (inetAddress[i].getAddress().length == 16)){  // IPv6 address
              serverAddress = inetAddress[i];
              break;
            }

            if(!IPv6 && (inetAddress[i].getAddress().length == 4)){  // IPv4 address
              serverAddress = inetAddress[i];
              break;
            }

            if(i == (inetAddress.length-1)){  // last cicle.
              System.err.println("Error: There isn't a valid IPv"+(IPv6?6:4)+" address for the server address introduced ("+address+").");
              System.err.println();
              System.exit(1);
            }
          }

        }catch(NumberFormatException e){
          System.err.println("Error: The the server address read is not a valid address ("+addressAux+").");
          System.err.println("The server address should be in the following format:\n\n"+
                             "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                             "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                             "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                             "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n"+
                             "the server address can be in the following format:\n\n"+
                             "    IPv4_address    // for IPv4 address\n"+
                             "    [IPv6_address]  // for IPv6 address\n");
          System.err.println();
          System.exit(1);
        }catch(Exception e2){
          System.err.println("Error: The the server address read is not a valid address ("+addressAux+").");
          System.err.println("The server address should be in the following format:\n\n"+
                             "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                             "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                             "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                             "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n"+
                             "the server address can be in the following format:\n\n"+
                             "    IPv4_address    // for IPv4 address\n"+
                             "    [IPv6_address]  // for IPv6 address\n");
          System.err.println();
          System.exit(1);
        }

        String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
        int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

        if(args.length == 12){  // Read the NTP server's address

          ntpServerAddress = args[10].trim();
          try { // Test if the NTP server address introduced is a valid address.
            if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  InetAddress.getByName(ntpServerAddress) + ").");
              System.exit(1);
            }
          }catch (Exception e) {
            System.err.println(
                "Error: The NTP server's address read is not a valid address (" +
                ntpServerAddress + ").");
            System.err.println();
            System.exit(1);
          }

          updateInterval = Integer.parseInt(args[11].trim());
          if (updateInterval < 0) {
            System.err.println("Error: The value of updateInterval can not be a negative number (" +
                               updateInterval + ").");
            System.err.println();
            System.exit(1);
          }

        }

        // Define the object hiResTimer used to get the system time.
        HiResTimer hiResTimer = null;
        if(updateInterval == -1){  // Use default updateInterval defined on HiResTimer class.
          hiResTimer = new HiResTimer(ntpServerAddress);
        }else
          hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

        OWAMP_FetchClient fetch = new OWAMP_FetchClient(serverAddress, serverTCPportNumber, new Socket(), systemInReader, hiResTimer);

        // processFetchSessionToDataBase(String sqlServerAddress, String dataBase, String userName, String passWord,
        //                               String strSid, int numberOfPackets, double slotParameter /*in seconds*/,
        //                               long realStartTime /*in milliseconds*/, double totalTimeToWait /*in milliseconds*/){ // process Fetch-Session
        if (!fetch.processFetchSessionToDataBase(args[1], args[2], args[3],
                                                 args[4], args[5],
                                                 Integer.parseInt(args[6].trim()),
                                                 Double.parseDouble(args[7].trim())   /*in seconds*/,
                                                 Long.parseLong(args[8].trim())       /*in milliseconds*/,
                                                 Double.parseDouble(args[9].trim())   /*in milliseconds*/)) {
          System.err.println("Error: Error processing session.");
          System.exit(1);
        }
      }catch (NumberFormatException e) {
        System.err.println(e);
        System.err.println("Error: One argument value read is not valid.");
        System.err.println("Error: Error processing session.");
        System.exit(1);
      }

      System.exit(0);  // In this case we must terminate the execution at the end of the processing of
                       // the requested session.
    }


    // If the number of parameters introduced is not equal to 10 we must use normal operation of Fetch-Client.

    boolean testError = false;     // Used to verify if an error occurred

    InetAddress serverAddress = null;
    String address = null;
    String addressAux = null;
    int serverTCPportNumber = DEFAULT_SEVER_PORT;  // TCPportNumber to use in OWAMP-Control with Server.
    boolean IPv6 = false;
    int i=0;

    do{
      testError = false;
      /*
      * The server address should be in the following format:
      *
      *     IPv4_address:TCPportNumber    // for IPv4 address
      *     [IPv6_address]:TCPportNumber  // for IPv6 address
      *
      * The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
      * Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368)
      * the server address can be in the following format:
      *
      *     IPv4_address    // for IPv4 address
      *     [IPv6_address]  // for IPv6 address
      */
      System.out.print("Introduce the address of the Server: ");
      try{
        address = systemInReader.readLine().trim();
      }                                        // if not, an UnknownHostException is thrown.
      catch (IOException e) {
        System.err.println(e);
        System.err.println("Error: You must introduce a valid IP address.");
        testError = true;
      }

      if (!testError) {
        addressAux = address;
        serverTCPportNumber = DEFAULT_SEVER_PORT; // TCPportNumber to use in OWAMP-Control with Server.
        IPv6 = false;

        try {

          if (address.indexOf("[") != -1) { // IPv6 address
            IPv6 = true;
            if (address.indexOf("]:") != -1) { // TCPportNumber to use in OWAMP-Control with sender specified.
              serverTCPportNumber = Integer.parseInt(address.substring(address.
                  indexOf("]:") + 2));
            }
            address = address.substring(address.indexOf("[") + 1,
                                        address.indexOf("]"));
          }
          else { // IPv4 address
            if (address.indexOf(":") != -1) { // TCPportNumber to use in OWAMP-Control with sender specified.
              serverTCPportNumber = Integer.parseInt(address.substring(address.
                  indexOf(":") + 1));
              address = address.substring(0, address.indexOf(":"));
            }
          }

          if (serverTCPportNumber < 0 || serverTCPportNumber >= 65536) {
            System.err.println("Error: Server TCP port number out of range (" +
                               serverTCPportNumber + ").");
            System.err.println();
            throw new Exception();
          }

          InetAddress inetAddress[] = InetAddress.getAllByName(address); // The array of all the IP addresses for a given host name.

          for (i = 0; i < inetAddress.length; i++) {
            if (IPv6 && (inetAddress[i].getAddress().length == 16)) { // IPv6 address
              serverAddress = inetAddress[i];
              break;
            }

            if (!IPv6 && (inetAddress[i].getAddress().length == 4)) { // IPv4 address
              serverAddress = inetAddress[i];
              break;
            }

            if (i == (inetAddress.length - 1)) { // last cicle.
              System.err.println("Error: There isn't a valid IPv" + (IPv6 ? 6 : 4) +
                                 " address for the server address introduced (" +
                                 address + ").");
              System.err.println();
              testError = true;
            }
          }

        }
        catch (NumberFormatException e) {
          System.err.println(
              "Error: The the server address read is not a valid address (" +
              addressAux + ").");
          System.err.println(
              "The server address should be in the following format:\n\n" +
              "    IPv4_address:TCPportNumber    // for IPv4 address\n" +
              "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n" +
              "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n" +
              "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n" +
              "the server address can be in the following format:\n\n" +
              "    IPv4_address    // for IPv4 address\n" +
              "    [IPv6_address]  // for IPv6 address\n");
          System.err.println();
          testError = true;
        }
        catch (Exception e2) {
          System.err.println(
              "Error: The the server address read is not a valid address (" +
              addressAux + ").");
          System.err.println(
              "The server address should be in the following format:\n\n" +
              "    IPv4_address:TCPportNumber    // for IPv4 address\n" +
              "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n" +
              "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n" +
              "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n" +
              "the server address can be in the following format:\n\n" +
              "    IPv4_address    // for IPv4 address\n" +
              "    [IPv6_address]  // for IPv6 address\n");
          System.err.println();
          testError = true;
        }

      }

    }while(testError);
    System.out.println();

    // Read the NTP server's address
    String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
    int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

    try {
      String line = null;
      System.out.print(
          "Do you want to get the time reference directly from a NTP server? (Default NO) ");
      line = systemInReader.readLine().trim();
      if ((line.compareToIgnoreCase("yes") == 0)||(line.compareToIgnoreCase("y") == 0)) {

        do {
          System.out.print("Introduce the address of the NTP server: ");
          ntpServerAddress = systemInReader.readLine().trim();
          testError = false;
          try { // Test if the NTP server address introduced is a valid address.
            if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  InetAddress.getByName(ntpServerAddress) + ").\n");
              testError = true;
            }
          }
          catch (Exception e) {
            testError = true;
            System.err.println(
                "Error: The NTP server's address read is not a valid address (" +
                ntpServerAddress + ").");
            System.err.println();
          }
        }while (testError);

      }
      System.out.println();

      do {
        System.out.print("Introduce the value (in minutes) to be used for the interval time between two consecutives updates to the time reference (Default 2): ");
        try{
          testError = false;
          line = systemInReader.readLine().trim();
          updateInterval = Integer.parseInt(line);
        }catch (NumberFormatException e2) {
          if (line.length() == 0) // 0 is the default option  - Default 2
            updateInterval = 2;   // 2 minutes
          else{
            testError = true;
            System.err.println("Error: " + e2 + ".");
            System.err.println("Error: The value of the interval time between two consecutives updates to the time reference read is not a valid.");
            System.err.println();
          }
        }

        if ((updateInterval < 0)&&(!testError)) {
          testError = true;
          System.err.println(
              "Error: The value of updateInterval can not be a negative number (" +
              updateInterval + ").");
          System.err.println();
        }
      }while (testError);
      System.out.println();

    }catch (IOException e) {
      System.err.println("Error: " + e + ".");
      System.exit(1);
    }

    // Define the object hiResTimer used to get the system time.
    HiResTimer hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

    OWAMP_FetchClient fetch = new OWAMP_FetchClient(serverAddress, serverTCPportNumber, new Socket(), systemInReader, hiResTimer);

    // Test fetchClientDir directory. If it doesn't exists it should be created.
    if(!fetch.fetchClientDir.exists())
      fetch.fetchClientDir.mkdir();

    // Test packet results directory. If it doesn't exists it should be created.
    if(!fetch.packetResults.exists())
      fetch.packetResults.mkdir();

    // Test test results directory. If it doesn't exist it should be created.
    if(!fetch.testResults.exists())
      fetch.testResults.mkdir();

    while(fetch.processFetchSession(true));

    System.exit(0);  // The hiResTimer thread to frequentely update the time reference doesn't let the FC to stop,
                     // so we should use System.exit() to stop it.
  }
}
