/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the class used to implement the OWAMP Server.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 1.2
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */


import java.net.*;
import java.io.*;

public class OWAMP_Server extends OWAMP_Base{

  // Class Variables

  static final short DEFAULT_RECEIVER_F_PORT = (short)28181;   // default Receiver port for TCP connection for
                                                               // Fetch Sessions.

  // Constructors

  /**
   * Constructor for class OWAMP_Server.
   * @param socketToClient Socket - represents the socket connection to the Client.
   * @param upTime Timestamp - uptime is a timestamp representing the time when the current instantiation of
   * the server started operating, in milliseconds since 1Jan1970.
   * @param receiverTCPport int - TCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the system time.
   */

  OWAMP_Server(Socket socketToClient, Timestamp upTime, int receiverTCPport, HiResTimer hiResTimer){

    this.socketToClient = new Socket();
    this.socketToClient = socketToClient;
    try{
      this.socketToClient.setSoTimeout(TIMEOUT);
    }
    catch(IOException e){
      System.err.println(e);
    }

    buffer = new byte[DEFAULT_PACKET_LENGTH];

    arrayOfSocketToReceiver = new Socket[1];
    arrayOfdataRS= new RequestSessionReceiver[1];
    arrayOfdataSSD = new ScheduleSlotDescriptionsReceiver[1][];

    this.upTime = new Timestamp(upTime.integerPartOfSeconds, upTime.fractionalPartOfSeconds);
    this.receiverTCPport = receiverTCPport;

    this.hiResTimer = hiResTimer;

    try{
      serverSocket = new ServerSocket();    // Used to verify if a socket connection is closed
                                            // using method isSocketConnected.
    }catch(IOException e){
      System.err.println(e);
    }

  }


  // Instance variables

  private Socket socketToClient;                  // Represents the socket connection to the Client.
  private Socket socketToReceiver;                // Represents the socket connection to a Receiver.
  private Socket[] arrayOfSocketToReceiver;       // Array of sockets connections to all Receivers of a OWAMP
                                                  // test session, with more than one test session.
  private int mode;                               // The mode value to be used in the
                                                  // current OWAMP control and test sessions.

  private RequestSessionReceiver[] arrayOfdataRS;  /* Array of datagrams Request Session (it has all request
                                                    * session made by the client). The max number of Request
                                                    * Session that the client can make is specified by the
                                                    * constant numMaxRS.
                                                    */

  private short numRS;          // Number of Request Sessions made by the client. It's automatically incremented
                                // for each new Request Sessions made.

  private ScheduleSlotDescriptionsReceiver[][] arrayOfdataSSD; // For each Request Session made by the client
                                                               // we have an array of ScheduleSlotDescriptions.

  private boolean keepConnection;  // This variable specifies if the server should keep the connection with a client
                                   // in the case of a negative response in a startTestSessionResponse operation.


  private HiResTimer hiResTimer;   // Used to read the system clock.
  private static File serverRSdir = new File("serverRSdir"); // The directory of all RS received from
                                                               // Clients in this Server.
  private int receiverTCPport;       // TCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.


  // Instance methods

  /**
   * This method start a new process to implement the Server.
   */

  public void run(){

    try{

                                     // Process Connection Setup Server Greeting
      if(!connectionSetupResponse(mode, socketToClient, upTime)){    //connectionSetupResponse(int modes, Socket socket, Timestamp upTime)
        System.err.println("Error: Error in processing Connection Setup with client in socket " +socketToClient);
        closeSocket(socketToClient);
        return;
      }

      numRS = 0;    // Number of Request Session made - there are no Request Session yet.

      while(true){ // The server keeps listening for the incoming packets.
        if(socketToClient.isClosed()){
          System.out.println("End of the current Session with Client's machine in "+socketToClient.toString());
          System.out.println();
          return;
        }

        socketToClient.setSoTimeout(0);  // Configure time to wait for packets from Client to infinite.

        int readn = 0;                   // number of bytes read.

        // To receive a packet we just must read the first byte to determine which packet was received.

        readn = receivePacket(socketToClient,buffer,0,1);   // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        socketToClient.setSoTimeout(TIMEOUT);

        if(readn == -1){
          if(isSocketConnected(socketToClient))  // An error occurred reading from socket
            System.err.println("Error: Error receiving Packet.");
          else{
            // Note: A client may not fecth the session results after its end, and close the socket to server.
            // In this case the error socket to client no longer available shouldn't be shown.
            // In this case the numRS is made zero.
            if(numRS != 0){
              System.err.println("Error: The socket connection to Client on " +
                                 socketToClient +
                                 " is no longer available. All sessions will be finished.\n");
              System.out.println("End of the current Session.\n");
            }else
              System.out.println("End of the current Session with Client's machine on "+socketToClient.toString()+".\n");
          }

          if(numRS != 0)      // send the SS to Client (it will be send to Client only if the socket is connected) and to Session-Receivers.
            StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

          closeSocket(socketToClient);

          return;
        }

        if(!processPacketReceived()){
          System.err.println("Error: Error in processing packet received");
          closeSocket(socketToClient);
          for(int i=0;i<numRS;i++)
            closeSocket(arrayOfSocketToReceiver[i]);       // Close the TCP connection to the Session-Receiver
          return;
        }
      }

    }
    catch(Exception ex){     // catches any type of exception.
      ex.printStackTrace();
    }
    finally{
      try{
        socketToClient.close();
        for(int i=0;i<numRS;i++)
          closeSocket(arrayOfSocketToReceiver[i]);       // Close the TCP connection to the Session-Receiver
      }
      catch(IOException ignored){
      }

      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();
    }
  }


  /**
   *  Method to receive the packet Request Session send by a client.
   *  To each Request-Session message, an OWAMP server MUST respond with an Accept-Session message
   * @param accept short - specify if the server accepts the request. In this message, zero in the
   * Accept field means that the server is willing to conduct the session.  A value of 1 indicates
   * rejection of the request.  All other values are reserved.
   * @param port int - The meaning of Port in the response depends on the values of Conf-Sender and
   * Conf-Receiver in the query that solicited the response. If both were set, Port field is unused.
   * If only Conf-Sender was set, Port is the port to expect OWAMP-Test packets from.  If only Conf-
   * - Receiver was set, Port is the port to send OWAMP-Test packets to.
   * @return boolean - This method returns true if the request operation was ok, otherwise it returns false.
   */

  public boolean requestSessionResponse(short accept, int port){

    // In variable byte[] buffer we have the packet Request Session received

    // Process the complete Request Session reception ( RS + SSD + IZP)

    System.out.println("Request Session from Client "+socketToClient);
    System.out.println();

    arrayOfdataRS[numRS] = new RequestSessionReceiver();
    System.arraycopy(buffer,0,arrayOfdataRS[numRS].buffer,0,arrayOfdataRS[numRS].PACKET_LENGTH);

                               // Initialize the list to receive the ScheduleSlotDescriptions packets.

    int numberOfScheduleSlots = Packets.getIntFromByteArray(buffer,4);  // numberOfScheduleSlots to receive
    arrayOfdataSSD[numRS] = new ScheduleSlotDescriptionsReceiver[numberOfScheduleSlots];

    if(!arrayOfdataRS[numRS].receiveCompleteRequestSession(arrayOfdataSSD[numRS],socketToClient)){
      System.err.println("Error: Error processing complete Request Session, from Client.");

      // Send the Stop All Session to Client. In case of failure,
      // the Server should stop all sessions and stop this OWAMP-Control session.

      StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS)

      return false;
    }

    // Test confSender and confReceiver

    byte sid[] = new byte[16];

    switch(arrayOfdataRS[numRS].confReceiver){
      case 0:
        if(arrayOfdataRS[numRS].confSender == 1)
          arrayOfdataRS[numRS].senderPort = port;
        break;
      case 1:
        sid = generateSid();                // If confReceiver is set, the value of sid is generated
                                            // by the server.
        System.arraycopy(sid,0,arrayOfdataRS[numRS].sid,0,sid.length);
    }


    // Sends the RS to the receiver

                                  // Determine Session-Receiver address

    InetAddress receiverAddress = determineSeReAddress(arrayOfdataRS[numRS].mbz_iPVN, arrayOfdataRS[numRS].receiverAddress);


    // Process the connection Setup Request to the Session-Receiver

    System.out.println("connectionSetupRequest to receiver "+receiverAddress);
    System.out.println();

    socketToReceiver = connectionSetupRequest(mode,receiverAddress,receiverTCPport,TIMEOUT);


    if(socketToReceiver == null){

      System.err.println("Error: Error processing connectionSetupRequest to receiver " +receiverAddress.toString());

      // Send the AcceptSession packet

      accept = 1;
      AcceptSessionToBeSend dataAS = new AcceptSessionToBeSend(accept,port,sid);
      dataAS.sendAS(socketToClient);

      return true;  // The server should not close the TCP connection in case of error. The client
                    // may close it.

    }

    System.out.println("Request Session to Receiver "+socketToReceiver);
    System.out.println();

    if(!requestSessionToReceiver(numRS)){  // requestSessionToReceiver(int index)
      System.err.println("Error: Error processing Request Session to Session-Receiver.");
      accept = 1;

      if(!socketToReceiver.isClosed())     // Close the TCP connection to the Session-Receiver
        closeSocket(socketToReceiver);

    }else{
      accept = 0;
      arrayOfSocketToReceiver[numRS] = socketToReceiver;
    }

    /* To each Request-Session message, an OWAMP server MUST respond with an Accept-Session message.
     * If the server rejects a Request-Session command, it SHOULD not close the TCP connection.  The
     * client MAY close it if it gets negative response to Request-Session.
     * The meaning of Port in the response depends on the values of Conf-Sender and Conf-Receiver in the
     * query that solicited the response. If both were set, Port field is unused.  If only Conf-Sender was
     * set, Port is the port to expect OWAMP-Test packets from.  If only Conf- Receiver was set, Port is the
     * port to send OWAMP-Test packets to.
     * If only Conf-Sender was set, SID field in the response is unused. Otherwise, SID is a unique server-
     * generated session identifier.  It can be used later as handle to fetch the results of a session.
     */

    // Send the AcceptSession packet

    System.out.println("AcceptSession to client on "+socketToClient);
    System.out.println();

    AcceptSessionToBeSend dataAS = new AcceptSessionToBeSend(accept,arrayOfdataRS[numRS].receiverPort,arrayOfdataRS[numRS].sid);
    dataAS.sendAS(socketToClient);

    if(accept != 0){        // The server rejected the requested session.
      closeSocket(arrayOfSocketToReceiver[numRS]);       // Close the TCP connection to the Session-Receiver
      System.err.println("Error: The server rejected the requested session.");
      System.err.println();
      System.err.println("  Possible errors:");
      System.err.println();
      System.err.println("  i)   The Session-Receiver doesn't exist.");
      System.err.println("  ii)  An error occurred processing connection to Receiver.");
      System.err.println("  iii) The Session-Receiver has no free UDP port to process this test session because all\n"+
                         "       the UDP ports in the range specified when the Session-Receiver was started are binded.");
      System.err.println("  iv)  An error occurred processing Request Session with Receiver.");
      System.err.println("  v)   The receiver rejected the requested session.");
      System.err.println();
      return true;  // The server should not close the TCP connection in case of error. The client
                    // may close it.
    }else{
      // An OWAMP client may issue zero or more Request-Session messages to an OWAMP server. So we have to increment
      // the number of Request Session made (numRS) and the size of the arrays used to keep the sessions
      // configurations. The new incremented position will be used by the next session configured.

      numRS++;   // Increment the number of Request Session made

      Socket[] arrayOfSocketToReceiverAux = new Socket[numRS+1];
      System.arraycopy(arrayOfSocketToReceiver,0,arrayOfSocketToReceiverAux,0,numRS);
      arrayOfSocketToReceiver = arrayOfSocketToReceiverAux;

      RequestSessionReceiver[] arrayOfdataRSAux = new RequestSessionReceiver[numRS+1];
      System.arraycopy(arrayOfdataRS,0,arrayOfdataRSAux,0,numRS);
      arrayOfdataRS = arrayOfdataRSAux;

      ScheduleSlotDescriptionsReceiver[][] arrayOfdataSSDAux = new ScheduleSlotDescriptionsReceiver[numRS+1][];
      for(int i=0;i<numRS;i++){
        arrayOfdataSSDAux[i] = new ScheduleSlotDescriptionsReceiver[arrayOfdataSSD[i].length];
        System.arraycopy(arrayOfdataSSD[i],0,arrayOfdataSSDAux[i],0,arrayOfdataSSD[i].length);
      }
      arrayOfdataSSD = arrayOfdataSSDAux;
    }

    return true;

  }


  /**
   * Method to send the packet Request Session to the Session-Receiver.
   * To each Request-Session an OWAMP Session-Receiver MUST respond with an Accept-Session message.
   * @param index int - The variable index represents the current Request Session in array arrayOfdataRS.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean requestSessionToReceiver(int index){

    // Test socket connection

    if((!socketToReceiver.isConnected())||(socketToReceiver.isClosed())){
      System.err.println("Error: You should connect to a Session-Receiver first, before trying this operation.");
      return false;
    }

    // Process the complete RequestSession command

    RequestSessionToBeSend auxDataRS = new RequestSessionToBeSend(arrayOfdataRS[index]);
    ScheduleSlotDescriptionsToBeSend[] auxArrayOfdataSSD = new ScheduleSlotDescriptionsToBeSend[auxDataRS.numberOfScheduleSlots];
    for(int i=0;i<auxDataRS.numberOfScheduleSlots;i++)
      auxArrayOfdataSSD[i] = new ScheduleSlotDescriptionsToBeSend(arrayOfdataSSD[index][i]);

    auxDataRS.sendCompleteRequestSession(auxArrayOfdataSSD,socketToReceiver);

    //  To each Request-Session message, an OWAMP Receiver MUST respond with an Accept-Session message.

    // Receive the AcceptSession packet

    AcceptSessionReceiver dataAS = new AcceptSessionReceiver();

    if(!dataAS.receiveAS(socketToReceiver)){
      System.err.println("Error: Error receiving packet AcceptSession from receiver, in operation Request session.");
      return false;
    }

    System.out.println("AcceptSession from receiver "+socketToReceiver);
    System.out.println();

    if(dataAS.accept != 0){
      System.err.println("Error: The receiver rejected the requested session.");
      closeSocket(socketToReceiver);
      return false;
    }

    arrayOfdataRS[index].receiverPort = dataAS.port;

    return true;

  }


  /**
   * Having requested one or more test sessions and received affirmative Accept-Session responses, an OWAMP
   * client may start the execution of the requested test sessions by sending a Start-Sessions message to
   * the server. The Server should send the StartingTestSession command to each Session-Receiver. If all
   * receiver accept these requests, the Server sends an Control_Ack packet to Client with accept equal to
   * 0, otherwise accept value is 1.
   * @param accept short - If Accept is 1, the Start-Sessions request was rejected; zero means that the
   * command was accepted.  All other values are reserved.  The server MAY and the client SHOULD close the
   * connection in the case of a negative response.
   * @return boolean -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean startingTestSessionsResponse(byte accept){ // process Starting Test Sessions Response

    // In variable byte[] buffer we have the packet startingTestSessions received
    // Test the startingTestSessions packet

    StartingTestSessions dataSTSR = new StartingTestSessions();
    if(!dataSTSR.testIntegrityZeroPadding(buffer,(short)16,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: Error receiving packet startingTestSessions.");

      // Send the Stop All Session to Client. In case of failure,
      // the Server should stop all sessions and stop this OWAMP-Control session.

      StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS)

      return false;
    }

    System.out.println("Starting Test Sessions from Client "+socketToClient);
    System.out.println();

    if(numRS < 1){  // The client didn't ask for a Request Session yet.
      System.err.println("Error: There is no sessions to start.");
      accept = 1;
    }

    Control_AckToBeSend dataCAtoClient;

    switch(accept){
      case 0:
        // The server SHOULD start all OWAMP-Test streams immediately after it sends the response or
        // immediately after their specified start times, whichever is later.
        // Send the StartingTestSessions packet to each receiver to start the test sessions. Each request
        // is considered separately in a different socket connection.

        System.out.println("Starting Test Sessions to Receivers.");
        System.out.println();

        for(int i = 0; i<numRS;i++){     // For each RS made.

          // Test socket connection

          if((!arrayOfSocketToReceiver[i].isConnected())||(arrayOfSocketToReceiver[i].isClosed())){
            System.err.println("Error: Error in Starting Test Session operation.");
            System.err.println("Error: The socket connection to receiver "+(i+1)+" - "+arrayOfSocketToReceiver[i]+" - is closed.");
            System.err.println("Error: All sessions requested so far will be finished.");
            System.err.println();
            accept = 1;
            break;
          }

          // Send StartingTestSessions packet to the Receiver

          StartingTestSessionsToBeSend dataSTS = new StartingTestSessionsToBeSend();
          dataSTS.sendSTS(arrayOfSocketToReceiver[i]);

          // Receive the Control-Ack packet

          Control_AckReceiver dataCA = new Control_AckReceiver();

          System.out.println("Control_Ack from Receiver "+(i+1)+": "+arrayOfSocketToReceiver[i]);
          System.out.println();

          if(!dataCA.receiveCA(arrayOfSocketToReceiver[i])){
            System.err.println("Error: Error in Starting Test Session operation.");
            System.err.println("Error: Error receiving packet Control-Ack in operation Starting Test Session, from receiver "+(i+1)+": "+arrayOfSocketToReceiver[i]+".");
            System.err.println("Error: All sessions requested so far will be finished.");
            System.err.println();
            accept = 1;
            break;
          }

          if(dataCA.accept == 1){

            System.err.println("Error: Error in Starting Test Session operation.");
            System.err.println("Error: The Session-Receiver "+(i+1)+" - "+arrayOfSocketToReceiver[i]+" - has rejected the Starting Test Session request.");
            System.err.println("Error: All sessions requested so far will be finished.");
            System.err.println();
            accept = 1;
            break;
          }

        }

        // If a receiver didn't accept a request session, the Server should inform the Client sending the
        // Control_Ack packet with accept equal to 1 and terminate all sessions.

        // Send datagram Control_AckToBeSend to Client

        dataCAtoClient = new Control_AckToBeSend(accept);
        dataCAtoClient.sendCA(socketToClient);

        System.out.println("Control_Ack to Client "+socketToClient);
        System.out.println();

                       // Stop all sessions started till now.

        if(accept == 1){

          Socket socketAux = socketToClient;

          System.err.println("Error: The Start-Sessions request was rejected.");
          System.err.println("Error: All sessions requested so far will be finished.");
          System.err.println();
          System.err.println("  Possible errors:");
          System.err.println();
          System.err.println("  i)   A Session-Receiver doesn't exists.");
          System.err.println("  ii)  A failure occurred in processing Starting Test Sessions between the Server and a Receiver.");
          System.err.println("  iii) A Session-Receiver has rejected the Starting Test Session request.");
          System.err.println("  iv)  A failure occurred receiving the Starting Test Sessions command from Client "+socketToClient);
          System.err.println();

          // Stop all sessions started till now. In case of failure,
          // the Server should stop all sessions and stop this OWAMP-Control session.

          StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS)

          if(keepConnection){            // This variable is used by the Server to verify if the it should keep the connection
            socketToClient = socketAux;  // with a client in the case of a negative response in a startTestSessionResponse operation.
            return true;
          }
          else
            return false;

        }

        long startTime = (long)hiResTimer.milliTime();


                                        // Calculate the total time to wait
        double totalTimeToWait = 0;
        double ti;

        for(int i = 0; i<numRS;i++){
          ti = totalTimeToWait(arrayOfdataRS[i], arrayOfdataSSD[i], hiResTimer);  // totalTimeToWait(RequestSession dataRS, ScheduleSlotDescriptions[] arrayOfdataSSD, HiResTimer hiResTimer)
          if(totalTimeToWait < ti)
            totalTimeToWait = ti;
        }

        arrayOfRStoFiles();  // Record the RS that have been accepted and started to a file. This information
                             // will be used in Fetch-Sessions.

        /* During total time to wait test if the client didn't request a Stop Session or if a Stop Session
         * haven't been received from a Receiver.
         */

        if(!testStopSession(startTime, totalTimeToWait)) // testStopSession(long startTime, long totalTimeToWait)
          return false;

        break;
      case 1:
        // The server didn't accept the command Start-Sessions request.
        System.err.println("Error: The server didn't accept the command Start-Sessions request.");

        // Send datagram Control_AckToBeSend to Client

        dataCAtoClient = new Control_AckToBeSend(accept);
        dataCAtoClient.sendCA(socketToClient);

                      // Stop all sessions started till now.

        // Stop all sessions started till now. In case of failure,
        // the Server should stop all sessions and stop this OWAMP-Control session.

        StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS)

        break;
      default:
        // Send datagram Control_AckToBeSend to Client.

        dataCAtoClient = new Control_AckToBeSend(accept);
        dataCAtoClient.sendCA(socketToClient);

                       // Stop all sessions started till now.

        // Stop all sessions started till now. In case of failure,
        // the Server should stop all sessions and stop this OWAMP-Control session.

        StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS)

        System.err.println("Error: The accept value returned by the server is not valid.");
        return false;
    }

    return true;

  }


  /**
   * During total time to wait, this method is used to test if the user didn't request a Stop Session.
   * @param startTime long - this variable represents the time that the current session started operating.
   * @param totalTimeToWait long - the time the Client should wait to finish the current session.
   * @return boolean -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean testStopSession(long startTime, double totalTimeToWait){

    // Wait for command Stop Session or end of total time to wait for the end of all sessions.

    // Wait for the StopSessions packets during totalTimeToWait

    try{

      int stopVal = 0;  // Auxiliary variable used to verify if a session was stopped.
      totalTimeToWait = totalTimeToWait + 5*1000;  // The Server should wait for totalTimeToWait + 5 sec
                                                   // to guaranty that, at the end of totalTimeToWait the
                                                   // Server and the Client don't send the Stop Session
                                                   // at same time.
      double totalTimeToWaitAux = totalTimeToWait;


                        // wait for Stop Session

      StopSessionsReceiver dataSS = new StopSessionsReceiver();

      do {

        // Test if a Stop Session was send by the Client and, if yes, process it.

        stopVal = dataSS.stopSessionFromSocket(socketToClient,arrayOfdataRS,arrayOfSocketToReceiver,numRS,totalTimeToWaitAux,true);

        if(stopVal == -1){    // If stopSession's accept val = 1 stop operation.
          System.err.println("Error: Error in StopSession operation from Client "+socketToClient);
          System.err.println();
          System.err.println("  Possible errors:");
          System.err.println();
          System.err.println("  i)   Failure of some sort. (Ex:connection reset by Client)");
          System.err.println("  ii)  Accept val received in StopSession from Client = 1.");
          System.err.println("  iii) Errors in StopSession packet received.");
          System.err.println();
          System.err.println("Error: All sessions will be finished.");
          return false;
        }

        if(stopVal == 1){  // If there is a StopSession

          System.out.println("StopSession from Client "+socketToClient);
          System.out.println();

                   // Test if not all session has been finished
          int j=0;
          for(int i=0;i<numRS;i++)
            if(!arrayOfSocketToReceiver[i].isClosed()){
              j++;
              break;
            }

          if (j == 0){  // all session has been finished
            System.out.println("End of all session between this Server and the Client "+socketToClient);
            System.out.println();
            if(dataSS.numberOfSessions!=0){   // If number of Sessions is zero it means that this socket will be used
              if(!socketToClient.isClosed()){ // to fetch the results of this session.
                closeSocket(socketToClient);
              }
            }
            numRS = 0; // It will be used to verify if all sessions are finished in case the socket from client is closed before
                       // the fetch-session, on method run. Note: A client may not fecth the session results after its end and close
                       // the socket to server. In this case the error socket to client no longer available shouldn't be shown.
            return true;
          }

                    // If there is a stopped session, test total time to wait again because the current
                    // TotalTimeToWait maybe has respect to a stopped RS.

          double ti = 0;
          totalTimeToWait = 0;
          for(int i = 0; i<numRS;i++){
            if(!arrayOfSocketToReceiver[i].isClosed()){
              ti = totalTimeToWait(arrayOfdataRS[i], arrayOfdataSSD[i], hiResTimer);  // totalTimeToWait(RequestSession dataRS, ScheduleSlotDescriptions[] arrayOfdataSSD, HiResTimer hiResTimer)
              if(totalTimeToWait < ti)
                totalTimeToWait = ti;
            }
          }

          // TotalTimeToWait is shorter now

          totalTimeToWaitAux = totalTimeToWait - ((long)hiResTimer.milliTime() - startTime);

        }

        // Verify the socketToClient.If it is closed before totalTimeToWait
        // It means that a failure has occurred, possibly a loss of connection
        // to Client. The session must be finished.

        if(!isSocketConnected(socketToClient)){
          System.err.println();
          System.err.println("Error: The socket connection to Client is no longer available. All sessions will be finished.");
          StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);
          return false;
        }


      }while ((startTime + totalTimeToWait) > ((long)hiResTimer.milliTime()));


      // If totalTimeToWait occurred
      // Send a StopSession to Client

      System.out.println("TotalTimetOwait occurred");
      System.out.println("StopSession to Client "+socketToClient);

      StopSessionsToBeSend dataSSS = new StopSessionsToBeSend( (short)0, 0);
      dataSSS.sendSS(socketToClient);

                    // Receive the StopSession response from Server.

      System.out.println("StopSession response from Client "+socketToClient);

      if(dataSS.stopSessionFromSocket(socketToClient,arrayOfdataRS,arrayOfSocketToReceiver,numRS,TIMEOUT,false)==-1){
        System.err.println("Error: The Client didn't response the request of StopSession or the packet received contains errors.");
        System.err.println("Error: All sessions will be terminated.");
        closeSocket(socketToClient);
        return false;
      }


      for(int i=0;i<numRS;i++){  // Close all sockets to Sessions-Receivers
          if(!arrayOfSocketToReceiver[i].isClosed())
            arrayOfSocketToReceiver[i].close();
      }

      return true;

    }
    catch(IOException e){
      System.err.println(e);
      return false;
    }

  }


  /**
   * This Class is used to process the Fetch Session as a Thread. When a user requests a Fetch Session
   * this thread is executed. Implementing the Fetch Session as a thread is useful because it permits
   * minimizing the delay in processing Stop Session, during Test Sessions.
   */

  public class ProcessFetchSession extends Thread{

    public void run(){

      if (!processFetchSession()){
        System.err.println("Error: Error processing Fetch session data operation.");
      }
    }


    /* Method to process the Server's Fetch Session response. The server will be the intermediate between
     * the Control-Client and the Session-Receiver of the requested session. It will determine the address
     * of that Session-Receiver and will forward the fetch session request to it. Then it will forward
     * the controlAck message sent by the session receiver to the client. Finally, if Session-Receiver sends
     * any results, the Server will forward results to the client.
     * @return boolean - It returns true if everything goes ok, otherwise it returns false.
     */

    public boolean processFetchSession(){

      Control_AckToBeSend dataCAs;

      // In variable byte[] buffer we have the packet FetchSession received

      FetchSessionReceiver dataFS = new FetchSessionReceiver();
      System.arraycopy(buffer,0,dataFS.buffer,0,dataFS.PACKET_LENGTH);
      if(!dataFS.receiveFS()){
        System.err.println("Error: Error in receiving FetchSession packet.");

        // Send the Control_Ack packet to the client

        dataCAs = new Control_AckToBeSend((byte)1);
        dataCAs.sendCA(socketToClient);

        closeSocket(socketToClient);
        return false;
      }

      System.out.println("Fetch Session from Client "+socketToClient);
      System.out.println();

      ScheduleSlotDescriptions[] arrayOfdataSSDAux = dataRSfromFile(dataFS.sid);
      if(arrayOfdataSSDAux == null){     // Read the dataRS from a file and record it in arrayOfdataRS[0].
                                         // This file, if it exits, it's name is a string representation
                                         // of the session SID and it should has the RS accepted by the
                                         // Sender and Receiver of test packets.(So, we are sore that
                                         // this session was started)

        System.err.println("Error: This Session doesn't exist.");

        // Send the Control_Ack packet to the client

        dataCAs = new Control_AckToBeSend((byte)1);
        dataCAs.sendCA(socketToClient);


        System.out.println("Control-Ack to Client on "+socketToClient);
        System.out.println();
        System.err.println("Error: The Fetch Session command was rejected by the Server.");
        System.err.println();
        closeSocket(socketToClient);

        return false;
      }

      // Determine the receiver's address that has the results of the session required

      InetAddress receiverAddress = determineSeReAddress(arrayOfdataRS[0].mbz_iPVN, arrayOfdataRS[0].receiverAddress);

      // Open a TCP connection to the Session-Receiver

      socketToReceiver = null;

      System.out.println("connectionSetupRequest to receiver"+receiverAddress);
      System.out.println();

                           // connectionSetup(int mode, InetAddress ia, int port, int timeout)
      socketToReceiver = connectionSetupRequest(mode,receiverAddress,receiverTCPport,TIMEOUT);

      if(socketToReceiver == null){
        System.err.println("Error: Error opening socket to receiver " +receiverAddress.toString());

        // Send the Control_Ack packet to the client

        dataCAs = new Control_AckToBeSend((byte)1);
        dataCAs.sendCA(socketToClient);

        closeSocket(socketToClient);

        return false;
      }

      // Send packet Fetch-Session to receiver

      // Send the dataFS packet to the receiver ( In variable byte[] dataFS.buffer we have the packet dataFS received from the client,
      // and is the same packet that will be send to the session receiver)

      System.out.println("Fetch Session to Receiver "+socketToReceiver);
      System.out.println();

      dataFS.sendTCP(socketToReceiver,dataFS.PACKET_LENGTH);    // send the fetch session message to the session receiver.

      // Receive the Control_Ack packet

      Control_AckReceiver dataCA = new Control_AckReceiver();

      if(!dataCA.receiveCA(socketToReceiver)){
        System.err.println("Error: Error in receiving the packet Control-Ack");
        closeSocket(socketToReceiver);
        dataCA.accept = (byte)1;
      }

      System.out.println("Control-Ack from Receiver "+socketToReceiver);
      System.out.println();

      // Send the Control_Ack packet to the client

      System.out.println("Control-Ack to Client "+socketToClient);
      System.out.println();

      dataCAs = new Control_AckToBeSend(dataCA.accept);
      dataCAs.sendCA(socketToClient);

      switch(dataCA.accept){    // Test the accept value.
        case 1:   // Fetch-Session request was rejected
          closeSocket(socketToClient);
          closeSocket(socketToReceiver);
          System.err.println("Error: The Fetch-Session request was rejected.");
          return false;
        case 0:   // If Accept was 0, the server sends the control ack to the client. The session receiver
                  // then will send the OWAMP-Test session data in question, followed by 16 octets of
                  // Integrity Zero Padding.

          // Receive the OWAMP-Test session data from the session receiver

          OTSessionDataReceiver dataSD = new OTSessionDataReceiver();
          if(!dataSD.receiveCompleteOTSessionData(socketToReceiver)){
            System.err.println("Error: Error receiving the OWAMP-Test session data from the session receiver.");
            closeSocket(socketToClient);
            closeSocket(socketToReceiver);
            return false;
          }

          System.out.println("OWAMP-Test session data from receiver "+socketToReceiver);
          System.out.println();

          // Send the OTSessionData packet to the client with the reproduction of request session

          System.out.println("OWAMP-Test session data to Client "+socketToClient);
          System.out.println();

          OTSessionDataToBeSend dataSDtoBeSend = new OTSessionDataToBeSend(arrayOfdataRS[0],arrayOfdataSSDAux,dataSD.numberOfPacketRecords);

          dataSDtoBeSend.sendCompleteOTSD(socketToClient);

          // Receive each PacketRecords from Receiver and send them to Client

          OwampTestPacketRecordReceiver packetRecordReceiver = new OwampTestPacketRecordReceiver();
          OwampTestPacketRecordToBeSend packetRecordToBeSend = new OwampTestPacketRecordToBeSend();

          for (int i=0;i<dataSD.numberOfPacketRecords;i++) {    // For each one of the ScheduleSlots Packets
             if(!packetRecordReceiver.receivePR(socketToReceiver)){
               closeSocket(socketToClient);
               closeSocket(socketToReceiver);
               System.err.println("Error: Error receiving packet PacketRecord "+(i+1)+" in operation Fetch session.");
               return false;  // All the packet records must be received properly.
             }else{
               packetRecordToBeSend.owampTestPacketRecordToBeSend(packetRecordReceiver);
               packetRecordToBeSend.sendOTPR(socketToClient);
             }
           }


           // Receive the OWAMP-Test session data IntegrityZeroPadding

           // Receive the IntegrityZeroPadding packet

           IntegrityZeroPaddingReceiver izp2 = new IntegrityZeroPaddingReceiver(16);
           if(!izp2.receiveIZP(socketToReceiver)){
             closeSocket(socketToClient);
             closeSocket(socketToReceiver);
             System.err.println("Error: Error receiving the OWAMP-Test session data IntegrityZeroPadding in operation Fetch session.");
             return false;
           }

           // Send integrityZeroPadding packet to Client

           IntegrityZeroPaddingToBeSend izpToBeSend = new IntegrityZeroPaddingToBeSend(16);
           izpToBeSend.sendIZP(socketToClient);

          break;
        default:
          closeSocket(socketToClient);
          closeSocket(socketToReceiver);
          System.err.println("Error: The accept value in Control-Ack is not valid.");
          return false;
      }

      closeSocket(socketToClient);

      closeSocket(socketToReceiver);

      return true;

    }
  }


  /**
   * This method is used to determine which packet was received and process it.
   * The first byte read represents the type of the packet.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean processPacketReceived(){

    int readn;   // number of bytes read
    switch((int)(buffer[0] & 0xFF)){  // First byte read, it represents the type of the packet.
      case 1:  //Request Session
        readn = receivePacket(socketToClient, buffer, 1, RequestSession.PACKET_LENGTH -1);   // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        if(readn == (RequestSession.PACKET_LENGTH -1)){   // Packet Request Session
          short accept = 0;
          int port = DEFAULT_RECEIVER_UDP_PORT;
          if (!requestSessionResponse(accept, port)) { // requestSession(short accept,int port)
            System.err.println("Error: Error in Request-Session operation.");
            return false;
          }
          return true;
        }
        break;
      case 2:  // Starting Test Sessions
        readn = receivePacket(socketToClient,buffer,1,StartingTestSessions.PACKET_LENGTH -1);   // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        if(readn == (StartingTestSessions.PACKET_LENGTH -1)){   // Packet Starting Test Sessions

          byte accept = 0;
          if (!startingTestSessionsResponse(accept)) {  // startingTestSessionsResponse(short accept)
            System.err.println("Error: Error in Starting Test Sessions operation.");
            return false;
          }
          return true;
        }
        break;
      case 3:  // Stop Sessions. If a SS is received before the Starting Test session command, this session should be finished.

        // Stop Sessions
        readn = receivePacket(socketToClient,buffer,1,StopSessions.PACKET_LENGTH -1);   // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        if(readn == (StopSessions.PACKET_LENGTH -1)){   // Packet Stop Sessions

          // In variable byte[] buffer we have the packet stopSessions received

          // Create the stopSessions packet

          StopSessionsReceiver dataSS = new StopSessionsReceiver();
          System.arraycopy(buffer,0,dataSS.buffer,0,dataSS.PACKET_LENGTH);

                                  // Initialize the list to receive the SessionPacketsSentRecord packets.

          int numberOfSessions = Packets.getIntFromByteArray(buffer,4);  // number Of SessionPacketsSentRecord to receive
          SessionPacketsSentRecordReceiver[] arrayOfdataSPSR = new SessionPacketsSentRecordReceiver[numberOfSessions];
          arrayOfdataSPSR = dataSS.receiveCompleteSS(socketToClient);    // Receive complete stopSessions

          if (arrayOfdataSPSR == null) {
            if(dataSS.accept != 1){
              System.err.println("Error: Error in Stop Sessions operation.");
              System.err.println("Error: Packet received is not a Stop Session packet.");

              // In case of failure the Server should send the stop all sessions to Client, stop all
              // sessions requested so far and stop this OWAMP-Control session.

              StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

              closeSocket(socketToClient);
              return false;
            }else
              System.err.println("Error: Stop Session, with  accept = 1, received from "+socketToClient);

            System.err.println("All sessions requested so far will be finished.");

            // In case of failure the Server should stop all sessions and stop this OWAMP-Control session.

            StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

            closeSocket(socketToClient);
            return true;

          }

          // At this point we can not receive a StopSession with accept equal to zero, so we should
          // send the Stop All Session to Client and stop this OWAMP-Control session.

          System.err.println("Error: Error in Stop Sessions operation.");
          System.err.println("Error: At this point we can not receive a StopSession with accept equal to zero.");
          System.err.println("All sessions requested so far will be finished.");

          StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

          return true;
        }
        break;
      case 4:   // Fetch-Session
        readn = receivePacket(socketToClient,buffer,1,FetchSession.PACKET_LENGTH -1);   // number of bytes read
        if(readn == (FetchSession.PACKET_LENGTH -1)){   // Packet Fetch Session
          ProcessFetchSession fetch = new ProcessFetchSession();
          fetch.run();
          return true;
        }

    }

    // If we reach this point it means that the packet received doesn't respects the protocol.
    // So, we should send the Stop All Session to Client and stop this OWAMP-Control session.

    StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

    return false;

  }


  /**
   * This method is used to find the index in arrayOfdataRS of a session that has the same sid as the sid
   * that it was given.
   * @param sid byte[] - sid of the session to be find.
   * @return int - this method returns the index of the session if one of the elements of arrayOfdataRS has
   * the same sid, otherwise it returns -1.
   */

  public int findSession(byte[] sid){

    String str1 = new String(sid);  // Auxiliary string
    for(int i = 0; i<numRS; i++){
      if (str1.compareTo(new String(arrayOfdataRS[i].sid)) == 0){
        return i;
      }
    }

    return -1;

  }


  /**
   * Method used to record the arrayOfdataRS and the arrayOfdataSSD, for each RS, in a file to be used in
   * Fetch Session. For each RS we have a different file, which name is the session SID as a string.
   * In this case we can not have the same file to all sessions because, if the Servers is communicating
   * with more then one Client at same time, they would use the same file. Which can not be possible. The
   * only way to resolve this problem is to have different files, with different names.
   * @return boolean -  This method returns true if everything goes ok, otherwise it returns false.
   */

  public boolean arrayOfRStoFiles(){

    try{

      // Test serverRSdir directory, to prevent errors if a user deletes it. If it doesn't exist it should be created.
      if (!serverRSdir.exists())
        serverRSdir.mkdir();

      for(int i=0;i<numRS;i++){
        if (!arrayOfSocketToReceiver[i].isClosed()) {  // It means that this session was started
                                                       // record the arrayOfdataRS

          File outputFileName = new File(serverRSdir.getPath()+serverRSdir.separator+owampHexEncode(arrayOfdataRS[i].sid)+".rs");  // If outputFile already exists, delete it
          if(outputFileName.exists())
            outputFileName.delete();

          RandomAccessFile outputFile = new RandomAccessFile(outputFileName,"rw");  // Create a new outputFile
          if(!saveRStoFile(arrayOfdataRS[i],arrayOfdataSSD[i],outputFile)){
            outputFile.close();
            return false;
          }

          outputFile.close();
        }
      }

    }
    catch(IOException ex){
      System.err.println(ex);
      return false;
    }

    return true;
  }


  /**
   * Method used to read the dataRS and the arrayOfdataSSD from a file to be used in Fetch Session.
   * For each started session we have a file which name is the SID of the session.
   * @return ScheduleSlotDescriptions[] -  This method returns the ScheduleSlotDescriptions[] if everything goes ok,
   * otherwise it returns null.
   */

  public ScheduleSlotDescriptions[] dataRSfromFile(byte[] sid){

    // Test serverRSdir directory, to prevent errors if a user deletes it. If it doesn't exist return false.
    if (!serverRSdir.exists())
      return null;

    String strSid = owampHexEncode(sid);
    File inputFileName = new File(serverRSdir.getPath()+serverRSdir.separator+strSid+".rs");  // Test if inputFile exists.
    if(!inputFileName.exists()){
      System.err.println("Error: There is no Sessions to Fetch (the Server's file of RS doesn't exits).");
      return null;
    }

    ScheduleSlotDescriptions[] arrayOfdataSSDAux = null;
    RandomAccessFile inFile = null;    // file with the RS made (RS is the combination of the arrayOfdataRS and the arrayOfdataSSD).

    try{

      inFile = new RandomAccessFile(inputFileName,"r");

      // read the arrayOfdataRS

      arrayOfdataRS[0] = new RequestSessionReceiver();

      arrayOfdataSSDAux = readRSfromFile(arrayOfdataRS[0],inFile);
      if(arrayOfdataSSDAux == null){
        inFile.close();
        return null;
      }

      inFile.close();

    }catch(IOException ex){
      System.err.println(ex);
      return null;
    }finally{
      try {
        inFile.close();
      }catch (IOException ex) {
        System.err.println(ex);
      }
    }

    return arrayOfdataSSDAux;

  }


  // Main function

  public static void main(String[] args) throws IOException{

    InetAddress serverAddress = null; // The local IP address to bind to. The serverAddress can be used on a
                                      // multi-homed host for a ServerSocket that will only accept connect requests
                                      // to one of its addresses. If serverAddress is null, it will default
                                      // accepting connections on any/all local addresses.
    int serverTCPportNumber = -1,     // TCP port number to be used in OWAMP-Control.
        receiverTCPport = -1;         // TCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.
    String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
                                      // If not defined the system time will be used to get the time reference.
    int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

    if((args.length == 1)&&((args[0].compareToIgnoreCase("-?") == 0)||(args[0].compareToIgnoreCase("-help") == 0))){  // Print the help message
      System.out.println("\nUsage: OWAMP_Server [-options]\n");
      System.out.println("where options include:");
      System.out.println("\n   -a | -address             \tThe local IP address to bind to. This address can be used on a multi-homed host");
      System.out.println("                               \tfor a Server that will only accept connection requests to one of its addresses.");
      System.out.println("                               \tIf not defined, it will by default accept connections on any/all local addresses.");
      System.out.println("\n   -p | -port port           \tThe TCP port number to be used in OWAMP-Control.");
      System.out.println("\n   -r | -receiverTCPport port\tTCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.");
      System.out.println("\n   -n | -ntpServer           \tThe address of the NTP server to be used to get the time reference. If not defined,");
      System.out.println("                               \tthe system time will be used to get the time reference.");
      System.out.println("\n   -u | -updateInterval      \tThe value (in minutes) to be used for the time interval between two consecutives");
      System.out.println("                               \tupdates of the time reference.");
      System.out.println("\n   -? | -help                \tPrints this help message.");
      System.exit(1);
    }

    if(args.length == 0){
      System.out.println("\nDefault port will be used on OWAMP-Control to this Server ("+DEFAULT_SEVER_PORT+").");
      System.out.println("\nDefault port will be used on OWAMP-Control between this Server and all OWAMP Session-Receivers ("+DEFAULT_RECEIVER_PORT+").");
      serverTCPportNumber = DEFAULT_SEVER_PORT;
      receiverTCPport = DEFAULT_RECEIVER_PORT;
    }else{
      try{

        int numOfArg = 0; // Used to determine if all arguments are ok.

        // option -address (It's the local address to bind to)
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-a") == 0)||(args[i].compareToIgnoreCase("-address") == 0)) {
            try{ // Test if the server address introduced is a valid address.
              serverAddress = InetAddress.getByName(args[i + 1]);
              if(!OWAMP_Base.isAnyLocalAddress(serverAddress)){
                System.err.println("Error: The OWAMP server's address read is not a valid local address (" + args[i + 1] + ").");
                System.exit(1);
              }
            }catch (Exception e) {
              System.err.println(
                  "Error: The OWAMP server's address read is not a valid address (" +
                  args[i + 1] + ").");
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -port
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-p") == 0)||(args[i].compareToIgnoreCase("-port") == 0)) {
            serverTCPportNumber = Integer.parseInt(args[i + 1]);
            if (serverTCPportNumber < 0 || serverTCPportNumber >= 65536) {
              System.err.println("Error: Server TCP port number out of range (" +
                                 serverTCPportNumber + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -receiverTCPport
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-r") == 0)||(args[i].compareToIgnoreCase("-receiverTCPport") == 0)) {
            receiverTCPport = Integer.parseInt(args[i + 1]);
            if (receiverTCPport < 0 || receiverTCPport >= 65536) {
              System.err.println("Error: Receiver TCP port number out of range (" +
                                 receiverTCPport + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -ntpServer
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-n") == 0)||(args[i].compareToIgnoreCase("-ntpServer") == 0)) {
            ntpServerAddress = args[i + 1];
            try{ // Test if the NTP server address introduced is a valid address.
              if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
                System.err.println(
                    "Error: The NTP server's address read is not a valid address (" +
                    InetAddress.getByName(ntpServerAddress) + ").");
                System.exit(1);
              }
            }catch (Exception e) {
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  ntpServerAddress + ").");
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -updateInterval
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-u") == 0)||(args[i].compareToIgnoreCase("-updateInterval") == 0)) {
            updateInterval = Integer.parseInt(args[i + 1]);
            if (updateInterval < 0) {
              System.err.println("Error: The value of updateInterval can not be a negative number (" +
                                 updateInterval + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        if((numOfArg<(double)args.length/2)||(serverAddress == null)&&(serverTCPportNumber==-1)&&(receiverTCPport==-1)&&(ntpServerAddress==null)&&(updateInterval==-1))
          throw new IllegalArgumentException();

        if(serverTCPportNumber==-1){ // Only receiverTCPport option was used.
          System.out.println("\nDefault port will be used on OWAMP-Control to this Server ("+DEFAULT_SEVER_PORT+").");
          serverTCPportNumber = DEFAULT_SEVER_PORT;
        }

        if(receiverTCPport==-1){ // Only port option was used.
          System.out.println("\nDefault port will be used on OWAMP-Control between this Server and all OWAMP Session-Receivers ("+DEFAULT_RECEIVER_PORT+").");
          receiverTCPport = DEFAULT_RECEIVER_PORT;
        }else
          System.out.println("\nTCP port "+receiverTCPport+" will be used on OWAMP-Control between this Server and all OWAMP Session-Receivers.");

      }catch (Exception e){
        System.err.println("\nIllegalArgumentException in thread main: Usage: OWAMP_Server [-options]\n");
        System.err.println("where options include:");
        System.err.println("\n   -a | -address             \tThe local IP address to bind to. This address can be used on a multi-homed host");
        System.err.println("                               \tfor a Server that will only accept connection requests to one of its addresses.");
        System.err.println("                               \tIf not defined, it will by default accept connections on any/all local addresses.");
        System.err.println("\n   -p | -port port           \tThe TCP port number to be used in OWAMP-Control.");
        System.err.println("\n   -r | -receiverTCPport port\tTCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.");
        System.err.println("\n   -n | -ntpServer           \tThe address of the NTP server to be used to get the time reference. If not defined,");
        System.err.println("                               \tthe system time will be used to get the time reference.");
        System.err.println("\n   -u | -updateInterval      \tThe value (in minutes) to be used for the time interval between two consecutives");
        System.err.println("                               \tupdates of the time reference.");
        System.err.println("\n   -? | -help                \tPrints this help message.");
        System.exit(1);
      }
    }

    System.out.println("\nStarting on port "+serverTCPportNumber);
    System.out.println();

    try{
      ServerSocket serverSocket = null;
      if(serverAddress == null)
        serverSocket = new ServerSocket(serverTCPportNumber); // initialize a server socket on port 'port'
      else
        serverSocket = new ServerSocket(serverTCPportNumber,0,serverAddress); // initialize a server socket on port 'port'

      // The hiResTimer should be definied here to guarante that
      // all Server instantiation has same time reference. It will
      // be passed has an argument of the Server constructor.
      HiResTimer hiResTimer = null;
      if(updateInterval == -1){  // Use default updateInterval defined on HiResTimer class.
        hiResTimer = new HiResTimer(ntpServerAddress);
      }else
        hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

      Timestamp upTime = new Timestamp(hiResTimer.milliTime(), (short)1);
                                             // Uptime is a timestamp representing the time when the
                                             // current instantiation of the server started operating, in
                                             // milliseconds since 1Jan1970.

      // Test serverRSdir directory. If it exists it should be cleaned (delete it and create it again). This
      // directory will have files, each with a RS made to this Server during a test session.

      if(OWAMP_Server.serverRSdir.exists())
        try{
          OWAMP_Server.deleteDir(OWAMP_Server.serverRSdir);
        }
        catch(IOException ex){
          ex.printStackTrace();
        }

      OWAMP_Server.serverRSdir.mkdir();

      // wait for requests connections

      while(true){
        Socket client = serverSocket.accept();                                              // Accept the client socket connection.
        OWAMP_Server server = new OWAMP_Server(client,upTime,receiverTCPport, hiResTimer);  // deliver this socket to the server
                                                                                            // OWAMP_Server(Socket socketToClient, Timestamp upTime, int receiverTCPport, HiResTimer hiResTimer)
        server.mode = 1;          // Mode of authentication.
        server.start();           // initialize thread.
      }
    }
    catch (SocketException se){
      System.err.println(se);
    }

  }

}

