/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the class used to implement the OWAMP Session-Sender.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 1.2
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */

import java.net.*;
import java.io.*;
import java.util.*;

class OWAMP_SessionSender extends OWAMP_SessionTerminal{

  // Constructors

  /**
   * Constructor for class OWAMP_SessionSender.
   * @param socket Socket - represents the socket connection to the Client.
   * @param upTime Timestamp - uptime is a timestamp representing the time when the current instantiation of
   * the sender started operating, in milliseconds since 1Jan1970.
   * @param portRange int[] - the UDP port range for test sessions.
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the system time.
   */

  OWAMP_SessionSender(Socket socket,  Timestamp upTime, int[] portRange, HiResTimer hiResTimer){
    super(portRange);
    try{

      this.socketToCS = socket;
      socketToCS.setSoTimeout(TIMEOUT);
    }
    catch(IOException ex){
      System.err.println(ex);
    }
    dataRS = null;
    buffer = new byte[DEFAULT_PACKET_LENGTH];
    this.upTime = new Timestamp(upTime.integerPartOfSeconds, upTime.fractionalPartOfSeconds);
    dataSS = new StopSessionsReceiver();
    element = "Client";                  // The Sender communicates directly with the Client.
    this.hiResTimer = hiResTimer;

    try{
      serverSocket = new ServerSocket();    // Used to verify if a socket connection is closed
                                            // using method isSocketConnected.
    }catch(IOException e){
      System.err.println(e);
    }

  }

  OWAMP_SessionSender(){super(null);}


  // Instance methods


  /**
   * This method start a new process to send the test packets.
   */

  public void run(){

    try{
                                     // Process Connection Setup Server Greeting

      if(!connectionSetupResponse(mode, socketToCS, upTime)){    //connectionSetupResponse(int modes, Socket socket, Timestamp upTime)

        System.err.println("Error: Error in processing Connection Setup with "+element+" in socket " +socketToCS);
        System.out.println();

        closeSocket(socketToCS);
        return;
      }

      // After the connection Setup process, a sender should receive the RS packet. It will be used to read the
      // sender IP address to be used to open the datagram socket, used to send the test packets to receiver.
      dataRS = new RequestSessionReceiver();
      socketToCS.setSoTimeout(TIMEOUT);
      System.out.println("Request-Session packet from Client on " +socketToCS+"\n");
      if(!dataRS.receiveRS(socketToCS)){
        System.err.println("Error: Error receiving the Request-Session packet from Client " +socketToCS);
        closeSocket(socketToCS);
        return;
      }

      InetAddress senderAddress = determineSeReAddress(dataRS.mbz_iPVN, dataRS.senderAddress);

      if(!isAnyLocalAddress(senderAddress)){
        System.err.println("Error: The Sender address in the Request-Session packet received from Client, in socket " +socketToCS+", is not a local address.");
        closeSocket(socketToCS);
        return;
      }

      dataRS = null; // It will be received again on RS command, with the receiver port to be used in the sending process.

      // After this process, a sender should send an AcceptSession packet to the ControlClient.
      // This packet has the possible UDP port to be used to send test packets to Receiver, in the port field
      // of the AcceptSession packet. This port will only be used if confSender is zero and confReceiver is 1.
      // To get this port number we should create a datagram socket, bound to a specified local address (the sender
      // address in RS packet) and a free port.
      // Note: We should use the sender address on RS packet as the local address to the datagram socket because
      // this address will be tested by the Receiver, to verify if the packet received was sent by this sender.
      // If not, the packet is dropped.
      byte accept = 0;
      int port = 0;
      if (this.portRange != null) {  // A port range was definied
        int i = 0;
        for (i = this.portRange[0]; i <= this.portRange[1]; i++) // Find a free port in the port range
          try {
            ds = new DatagramSocket(i, senderAddress); // initialize a datagram socket on a free port of port range to be used.
            port = ds.getLocalPort();
            break;
          }
          catch (Exception e) {
          }
        if (i > this.portRange[1]) { // All ports in the specified range are binded. We should reject the session.
          System.err.println(
              "Error: The Session-Sender has no free UDP port to process this test session.");
          System.err.println("Error: All the UDP ports in the range specified when the Session-Sender was started are binded.");
          System.err.println();
          accept = 1;
          port = 0;
        }
      }else{  // No port range was definied. So a free UDP port should be used.
        // To get a free UDP port an auxiliary datagram socket should be created and then its port should be
        // used to create the datagram socket to be used by this Sender. As we have to close the auxiliary datagram
        // socket first and then create the new one, another process can use the port we want during this interval.
        // So, if an exception ocurres when trying to create the new datagram socket we should try a new free port again.
        port = -1;
        short numberOfAttempts = 100, count = 0;  // To define the number of attempts that it should be processed.
        do{
          try {
            ds = new DatagramSocket();  // initialize a datagram socket on a free port. This free port will be used
                                        // as the port of the datagram socket to be used by this Sender.
            port = ds.getLocalPort();
            ds.close();
            ds = new DatagramSocket(port, senderAddress); // initialize a datagram socket on the free port.
            accept = 0;
          }catch (Exception e) {
            accept = 1;
            port = -1;
            count++;
            System.err.println(e);
          }
        }while((port == -1)&&(count < numberOfAttempts));
      }

      AcceptSessionToBeSend dataAS = new AcceptSessionToBeSend(accept,
          port, new byte[16]);
      dataAS.sendAS(socketToCS);
      if (accept == 0){
        System.out.println("AcceptSession, with sender UDP port, to client on " +
                           socketToCS);
        System.out.println();
      }else{
        System.out.println("AcceptSession to client on " + socketToCS);
        System.out.println();
        closeSocket(socketToCS);
        if(port == 0){
          System.err.println(
              "Error: The Session-Sender has no free UDP port to process this test session.");
          System.err.println("Error: All the UDP in the range specified when the Session-Sender was started are binded.");
        }else
          System.err.println(
              "Error: The Session-Sender has no free UDP port to process this test session or couldn't bind the UDP socket to the requested address.");
        System.err.println();
        return;
      }

      int readn = 0;
      while(true){ // The Sender keeps listening for the incoming packets
        if(socketToCS.isClosed()){
          if((ds != null) && (!ds.isClosed()))
            ds.close();
          System.out.println("End of the current Session.\n");
          return;
        }

        socketToCS.setSoTimeout(0);  // Configure time to wait for packets from Client to infinite

        readn = 0;

        readn = receivePacket(socketToCS, buffer, 0, 1);   // number of bytes read
        socketToCS.setSoTimeout(TIMEOUT);

        if(readn == -1){
          if(isSocketConnected(socketToCS))  // An error occurred reading from socket
            System.err.println("Error: Error receiving Packet.");
          else{
            System.err.println("Error: The socket connection to Client" +
                               " on "+socketToCS+" is no longer available. All sessions will be finished.\n");
            System.out.println("End of the current Session.\n");
          }
          closeSocket(socketToCS);
          if((ds != null) && (!ds.isClosed()))
            ds.close();
          return;
        }

        if(!processPacketReceived()){
          System.err.println("Error: Error in processing packet received");
          if((ds != null) && (!ds.isClosed()))
            ds.close();
          if(socketToCS.isClosed())
            return;
                                 // Send a StopSession to Client

          StopSessionsToBeSend dataSSS = new StopSessionsToBeSend((byte)1, 1);
          byte[] sid = new byte[16];
          if(dataRS != null)
            System.arraycopy(dataRS.sid,0,sid,0,dataRS.sid.length);
          System.out.println("Stop Sessions to "+element+" "+socketToCS);
          System.out.println();
          if(!dataSSS.sendCompleteSS(sid,0,socketToCS)) // sendCompleteSS(byte[] sid, int sessionPacketsSent, Socket socket)
            System.err.println("Error: Error sending StopSession command to "+element+" in socket " +socketToCS);

          closeSocket(socketToCS);

          return;
        }
      }

    }
    catch(IOException ex){
      ex.printStackTrace();
    }
    finally{
      try{
        socketToCS.close();
        if((ds != null) && (!ds.isClosed()))
          ds.close();
      }
      catch(IOException ignored){
      }

      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();
    }
  }


  /**
   * Method to receive the packet Request Session send by a client.
   * To each Request-Session message, an OWAMP Session-Sender MUST respond with an Accept-Session message.
   * @param accept short - specify if the Session-Sender accepts the request. In this message, zero in the
   * Accept field means that the Session-Sender is willing to conduct the session.  A value of 1 indicates
   * rejection of the request.  All other values are reserved.
   * @return boolean - This method returns true if the request operation was ok, otherwise it returns false.
   */

  public boolean requestSessionResponse(byte accept){

    AcceptSessionToBeSend dataAS;

    if(dataRS != null){   // if a RS was already received refuse the RS and return
      // We should receive this RS to unfill the socket.
      // In variable byte[] buffer we have the packet Request Session received
      // Process the complete Request Session reception ( RS + SSD + IZP)
      int numberOfScheduleSlots = Packets.getIntFromByteArray(buffer,4);  // numberOfScheduleSlots to receive
      ScheduleSlotDescriptionsReceiver[] arrayOfdataSSDAux = new ScheduleSlotDescriptionsReceiver[numberOfScheduleSlots];
      new RequestSessionReceiver().receiveCompleteRequestSession(arrayOfdataSSDAux,socketToCS);

      dataAS = new AcceptSessionToBeSend((short)1,0,new byte[16]);
      dataAS.sendAS(socketToCS);
      System.out.println("Accept Session to "+element+" "+socketToCS);
      System.err.println("Error: A second RS was received in the same socket from "+element+".");
      System.out.println();
      return false;
    }

    // In variable byte[] buffer we have the packet Request Session received

    // Process the complete Request Session reception ( RS + SSD + IZP)

    dataRS = new RequestSessionReceiver();
    System.arraycopy(buffer,0,dataRS.buffer,0,dataRS.PACKET_LENGTH);

                               // Initialize the list to receive the ScheduleSlotDescriptions packets.

    int numberOfScheduleSlots = Packets.getIntFromByteArray(buffer,4);  // numberOfScheduleSlots to receive
    arrayOfdataSSD = new ScheduleSlotDescriptionsReceiver[numberOfScheduleSlots];

    if(!dataRS.receiveCompleteRequestSession(arrayOfdataSSD,socketToCS)){
      System.err.println("Error: Error processing complete Request Session, from "+element+".");
      return false;
    }

    //  To each Request-Session message, an OWAMP Sender MUST respond with an Accept-Session message.

    // Send the AcceptSession packet

    dataAS = new AcceptSessionToBeSend(accept,ds.getLocalPort(),dataRS.sid);
    dataAS.sendAS(socketToCS);
    System.out.println("Accept Session to "+element+" on "+socketToCS);
    System.out.println();

    if(accept == 1){
      System.err.println();
      System.err.println("Error: The Session-Sender rejected the Request-Session.");
      System.err.println();
      closeSocket(socketToCS);
      return false;
    }

    return true;

  }


  /**
   * Having requested one or more test sessions and received affirmative Accept-Session responses, an OWAMP
   * client may start the execution of the requested test sessions by sending a Start-Sessions message to
   * the server.
   * @param accept byte - If Accept is 1, the Start-Sessions request was rejected. Zero means that the
   * command was accepted.  All other values are reserved.  The server MAY and the client SHOULD close the
   * connection in the case of a negative response.
   * @return boolean -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean startingTestSessionsResponse(byte accept){ // process Starting Test Sessions Response

    Control_AckToBeSend dataCA;

    if(dataRS == null){ // if a RS wasn't received yet refuse the STS and return.
      System.err.println("Error: A StartingTestSession command was received before a RequestSession command.");
      System.err.println("Error: This session will be terminated.");

                              // Send datagram Control_AckToBeSend.

      dataCA = new Control_AckToBeSend((byte)1);
      dataCA.sendCA(socketToCS);
      System.out.println("Control Ack to "+element+" "+socketToCS);
      System.out.println();

      closeSocket(socketToCS);

      return false;
    }

    // In variable byte[] buffer we have the packet startingTestSessions received.
    // Test the startingTestSessions packet.

    StartingTestSessions dataSTSR = new StartingTestSessions();
    if(!dataSTSR.testIntegrityZeroPadding(buffer,(short)16,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: Error receiving packet startingTestSessions.");
      return false;
    }

    // Send datagram Control_AckToBeSend

    dataCA = new Control_AckToBeSend(accept);
    dataCA.sendCA(socketToCS);
    System.out.println("Control Ack to "+element+" "+socketToCS);
    System.out.println();

    if(accept == 0){
      switch(mode){
        case 1:
          if(!process_StreamPackets_Unauthenticated())
            System.err.println("Error: Error processing Stream Packets Unauthenticated.");
          break;
  /*    case 2:
        process_StreamPackets_Authenticated();
        break;
      case 4:
        process_StreamPackets_Encrypted();
        break;*/
      }
    }else{
      System.err.println("Error: The StartingTestSession command was rejected.");
      closeSocket(socketToCS);
    }

    closeSocket(socketToCS);   // At this point the session is already finished.

    return true;

  }


  /**
   * This method is used to process the sending of the Stream Packets Unauthenticated.
   * @return boolean - This method returns true if the request operation was ok, otherwise it returns false.
   */

  public boolean process_StreamPackets_Unauthenticated(){

    try{

      int sequenceNumber = 0; // Test Packet sequence number.
      Timestamp timestamp = new Timestamp((double)0,(short)0);
      int packetPaddingLenght = ((dataRS.paddingLenght + StreamPackets_Unauthenticated.PACKET_LENGTH) < _OWP_TESTREC_SIZE)?(_OWP_TESTREC_SIZE - StreamPackets_Unauthenticated.PACKET_LENGTH):dataRS.paddingLenght;
      byte[] packetPadding = new byte[packetPaddingLenght];
      Random rand = new Random();
      double delay = 0;
      byte s_z_scale = (byte)0x3F, multiplier = (byte)0xFF;
      ErrorEstimate errorEstimate = new ErrorEstimate(s_z_scale,multiplier);

      // Determine receiver address

      InetAddress receiverAddress = determineSeReAddress(dataRS.mbz_iPVN, dataRS.receiverAddress);

      // Create packet padding

      if(dataRS.paddingLenght != 0){
        for (int i = 0; i < dataRS.paddingLenght; i++)
          packetPadding[i] = (byte) (rand.nextInt() & 0xFF);
      }

      // Verify start time

      // Get Session start time.
      double startTime = dataRS.startTime.timeInMillisecondsSince1970();

      waitTimeToWait(1000 , 0);  // wait 1 second. It is used to give the Starting  test Sessions packet time to
                                 // reach every receiver in case of StartTime in the pass.

      switch(waitTimeToWait(startTime - (long)hiResTimer.milliTime(), 0)){   // waitTimeToWait(long timeToWait, int sessionPacketsSent)
        case 0:
          break;
        case 1:
          System.err.println("This Session will not be started.");
          closeSocket(socketToCS);
          ds.close();
          return true;
        case -1:
          System.err.println("Error: The Session on "+socketToCS+" will not be started.\n");
          closeSocket(socketToCS);
          ds.close();
          return false;
        default:
          System.err.println("Error: The Session on "+socketToCS+" will not be started.\n");
          closeSocket(socketToCS);
          ds.close();
          return false;
      }

      // Send packets StreamPackets_Unauthenticated

      System.out.println("Test packets to receiver "+new InetSocketAddress(receiverAddress, dataRS.receiverPort)+".");
      System.out.println();

      if(dataRS.senderPort != ds.getLocalPort()){  // In case confSender is not zero, the Sender will have a different port.
        ds.close();
        ds = new DatagramSocket(dataRS.senderPort,determineSeReAddress(dataRS.mbz_iPVN, dataRS.senderAddress)); // creates a UDP connection to send test packets to the Receiver.
      }

      StreamPackets_UnauthenticatedToBeSend dataSPU = new
        StreamPackets_UnauthenticatedToBeSend(sequenceNumber,
                                              timestamp, errorEstimate,
                                              packetPadding);

      for(int i = 0, indexOfSSD = -1;i<dataRS.numberOfPackets;i++, sequenceNumber++, indexOfSSD++){

        if(indexOfSSD > (arrayOfdataSSD.length - 1))
          indexOfSSD = 0;

        if(i != 0){          // i = 0 - first packet
          delay = arrayOfdataSSD[indexOfSSD].slotParameter.timeInMillisecondsSince1900();

          switch(waitTimeToWait(delay,i)){   // waitTimeToWait(long timeToWait, int sessionPacketsSent)
            case 0:
              break;
            case 1:
              closeSocket(socketToCS);
              ds.close();
              return true;
            case -1:
              closeSocket(socketToCS);
              ds.close();
              return false;
            default:
              closeSocket(socketToCS);
              ds.close();
              return false;
          }
        }

        // Verify the socketToCS (Client).If it is closed before totalTimeToWait, it means that a failure has
        // occurred, possibly a loss of connection to client. The session must be finished.

        if(!isSocketConnected(socketToCS)){
          System.err.println();
          System.err.println("Error: The socket connection to Client is no longer available. This session will be finished.");
          ds.close();
          return false;
        }

        timestamp.timestamp((long)hiResTimer.milliTime(),(short)1);  // Time in milliseconds since 1Jan1970
        dataSPU.streamPackets_UnauthenticatedToBeSend(sequenceNumber, timestamp, errorEstimate, packetPadding);
        dataSPU.sendSPU(receiverAddress, ds, dataRS.receiverPort, hiResTimer);  // void sendSPU(InetAddress ia, DatagramSocket ds, int port, HiResTimer hiResTimer)
      }

    }
    catch(IOException e){
      System.err.println(e);
      if(ds.getLocalPort()!=dataRS.senderPort){
        StopSessionsToBeSend dataSSS = new StopSessionsToBeSend( (short) 1, 0);
        dataSSS.sendSS(socketToCS);
        System.out.println("StopSessions to Client "+socketToCS);
        System.out.println();
      }else
        ds.close();
      return false;
    }

    ds.close();
    closeSocket(socketToCS);

    return true;
  }


  /**
   * This method is used to stop the processing and wait the time in ms specified by variable timeToWait.
   * During this time the Sender test if a StopSession has been sent by the client, and if yes it is processed.
   * If accept field of StopSession packet received is 0, it means normal but possibly premature completion. So
   * the Sender should send back to client a StopSession packet with the accept field indicating if this
   * session is still running and the number of session packets sent in field number of sessions. If accept
   * is zero, the session is still running. The Sender should stop sending packets.
   * If accept field of StopSession packet received is 1, it means failure of some sort. In this case the session
   * to Client should be closed and the sending of test packets stopped.
   * @param timeToWait long - the time to wait before continuing.
   * @param sessionPacketsSent int - the number of test packets already send.
   * @return short - It returns -1 if a failure occures or a SS with accept = 1 is received, 0 if none SS is received
   * and 1 if a SS with accept = 0 is received.
   */

  public short  waitTimeToWait(double timeToWait, int sessionPacketsSent){

    if(timeToWait < 0)
      timeToWait = 1;

    // During timeToWait wait to receive StopSession from Client

    int stopVal = dataSS.receiveSS(socketToCS,timeToWait);

    if( stopVal == -1){
      System.err.println("Error: Error receiving packet StopSession from Client in operation Stop Sessions.");

      StopSessionsToBeSend dataSSS = new StopSessionsToBeSend( (short) 1, sessionPacketsSent);
      dataSSS.sendSS(socketToCS);
      System.out.println("StopSessions to Client "+socketToCS);
      System.out.println();

      closeSocket(socketToCS);  // Sender should close connection to Client and stop sending test packets.

      return -1;

    }

    if(stopVal == 0)  // None StopSessions were received
      return 0;

    // If a Stop Sessions was received

    System.out.println("Stop Sessions from Client "+socketToCS);

    if (dataSS.accept == 0) {
         // Send the stopSessions packet with the number of session packets sent in field number of sessions.
         // After this, the Sender should close connection to Client and stop sending test packets.

      StopSessionsToBeSend dataSSS = new StopSessionsToBeSend( (short) 0, sessionPacketsSent);
      dataSSS.sendSS(socketToCS);
      System.out.println("StopSessions to Client "+socketToCS);
      System.out.println();

      closeSocket(socketToCS);  // Sender should close connection to Client and stop sending test packets.

      return 1;

    }else{  // if accept is 1 the Sender should close connection to Client and stop sending test packets.
      System.err.println("Error: Stop session from Client with accept = 1 (failure of some sort occurred).");
      closeSocket(socketToCS);
      return -1;
    }

  }


  // Main function

  public static void main(String[] args) throws IOException{

    InetAddress senderAddress = null;   // The local IP address to bind to. The senderAddress can be used on a
                                        // multi-homed host for a ServerSocket that will only accept connect requests
                                        // to one of its addresses. If serverAddress is null, it will default
                                        // accepting connections on any/all local addresses.
    int port = -1;            // TCP port number to be used in OWAMP-Control.
    int[] portRange = null;   // Represents the UDP port range to be used during test sessions.
    String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
                                      // If not defined the system time will be used to get the time reference.
    int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

    if((args.length == 1)&&((args[0].compareToIgnoreCase("-?") == 0)||(args[0].compareToIgnoreCase("-help") == 0))){  // Print the help message
      System.out.println("\nUsage: OWAMP_SessionSender [-options]\n");
      System.out.println("where options include:");
      System.out.println("\n   -a | -address                     \tThe local IP address to bind to. This address can be used on a multi-homed host");
      System.out.println("                                       \tfor a Session-Sender that will only accept connection requests to one of its addresses.");
      System.out.println("                                       \tIf not defined, it will by default accept connections on any/all local addresses.");
      System.out.println("\n   -p | -port port                   \tThe TCP port number to be used in OWAMP-Control.");
      System.out.println("\n   -r | -portRange [lowport,highport]\tThe UDP port range to be used in OWAMP-Test sessions.");
      System.out.println("\n   -n | -ntpServer                   \tThe address of the NTP server to be used to get the time reference. If not defined,");
      System.out.println("                                       \tthe system time will be used to get the time reference.");
      System.out.println("\n   -u | -updateInterval              \tThe value (in minutes) to be used for the time interval between two consecutives");
      System.out.println("                                       \tupdates of the time reference.");
      System.out.println("\n   -? | -help                        \tPrints this help message.");
      System.exit(1);
    }

    if(args.length == 0){
      System.out.println("\nDefault ports will be used.");
      port = DEFAULT_SENDER_PORT;
    }else{
      try{
        int numOfArg = 0; // Used to determine if all arguments are ok.

        // option -address (It's the local address to bind to)
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-a") == 0)||(args[i].compareToIgnoreCase("-address") == 0)) {
            try{ // Test if the sender address introduced is a valid address.
              senderAddress = InetAddress.getByName(args[i + 1]);
              if(!OWAMP_Base.isAnyLocalAddress(senderAddress)){
                System.err.println("Error: The OWAMP Session-Sender's address read is not a valid local address (" + args[i + 1] + ").");
                System.exit(1);
              }
            }catch (Exception e) {
              System.err.println(
                  "Error: The OWAMP Session-Sender's address read is not a valid address (" +
                  args[i + 1] + ").");
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -port
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-p") == 0)||(args[i].compareToIgnoreCase("-port") == 0)) {
            port = Integer.parseInt(args[i + 1]);
            if (port < 0 || port >= 65536) {
              System.err.println("Error: Sender TCP port number out of range (" +
                                 port + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -portRange
        for (int i = 0; i < args.length; ) {

          if ((args[i].compareToIgnoreCase("-r") == 0)||(args[i].compareToIgnoreCase("-portRange") == 0)) {
            int upperIndex = args[i + 1].indexOf(",");
            int lowRange = Integer.parseInt(args[i + 1].substring(1, upperIndex)),
                upperRange = Integer.parseInt(args[i + 1].substring(upperIndex + 1,args[i + 1].indexOf("]")));
            if (lowRange < 0 || lowRange >= 65536) {
              System.err.println("Error: Lower range TCP port number out of range (" +
                                 lowRange + ").");
              System.err.println();
              System.exit(1);
            }
            if (upperRange < 0 || upperRange >= 65536) {
              System.err.println("Error: Upper range TCP port number out of range (" +
                                 upperRange + ").");
              System.err.println();
              System.exit(1);
            }
            if(upperRange - lowRange < 0){
              System.err.println("\nError: Upper port range should be greather than lower port range.");
              throw new IllegalArgumentException();
            }

            portRange = new int[2];
            portRange[0] = lowRange;
            portRange[1] = upperRange;
            numOfArg++;
            break;
          }

          i += 2;
        }

        // option -ntpServer
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-n") == 0)||(args[i].compareToIgnoreCase("-ntpServer") == 0)) {
            ntpServerAddress = args[i + 1];
            try{ // Test if the NTP server address introduced is a valid address.
              if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
                System.err.println(
                    "Error: The NTP server's address read is not a valid address (" +
                    InetAddress.getByName(ntpServerAddress) + ").");
                System.exit(1);
              }
            }catch (Exception e) {
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  ntpServerAddress + ").");
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -updateInterval
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-u") == 0)||(args[i].compareToIgnoreCase("-updateInterval") == 0)) {
            updateInterval = Integer.parseInt(args[i + 1]);
            if (updateInterval < 0) {
              System.err.println("Error: The value of updateInterval can not be a negative number (" +
                                 updateInterval + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        if((numOfArg<(double)args.length/2)||(senderAddress == null)&&(port==-1)&&(portRange==null)&&(ntpServerAddress==null)&&(updateInterval==-1))
          throw new IllegalArgumentException();

      }catch (Exception e){
        System.err.println("\nIllegalArgumentException in thread main: Usage: OWAMP_SessionSender [-options]\n");
        System.err.println("where options include:");
        System.err.println("\n   -a | -address                     \tThe local IP address to bind to. This address can be used on a multi-homed host");
        System.err.println("                                       \tfor a Session-Sender that will only accept connection requests to one of its addresses.");
        System.err.println("                                       \tIf not defined, it will by default accept connections on any/all local addresses.");
        System.err.println("\n   -p | -port port                   \tThe TCP port number to be used in OWAMP-Control.");
        System.err.println("\n   -r | -portRange [lowport,highport]\tThe UDP port range to be used in OWAMP-Test sessions.");
        System.err.println("\n   -n | -ntpServer                   \tThe address of the NTP server to be used to get the time reference. If not defined,");
        System.err.println("                                       \tthe system time will be used to get the time reference.");
        System.err.println("\n   -u | -updateInterval              \tThe value (in minutes) to be used for the time interval between two consecutives");
        System.err.println("                                       \tupdates of the time reference.");
        System.err.println("\n   -? | -help                        \tPrints this help message.");
        System.exit(1);
      }

      if(port == -1){
        port = DEFAULT_SENDER_PORT;
        System.out.println("\nDefault port for OWAMP-Control will be used ("+DEFAULT_SENDER_PORT+").");
      }

      if(portRange==null){
        System.out.println("\nDefault portRange for OWAMP-Test will be used.");
      }

    }

    System.out.println("\nStarting on port "+port);
    System.out.println();

    try{
      ServerSocket senderSocket = null;
      if(senderAddress == null)
        senderSocket = new ServerSocket(port); // initialize a server socket on port 'port'
      else
        senderSocket = new ServerSocket(port,0,senderAddress); // initialize a server socket on port 'port'

      // The hiResTimer should be definied here to guarante that
      // all Sender instantiation has same time reference. It will
      // be passed has an argument of the Sender constructor.
      HiResTimer hiResTimer = null;
      if(updateInterval == -1){  // Use default updateInterval defined on HiResTimer class.
        hiResTimer = new HiResTimer(ntpServerAddress);
      }else
        hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

      Timestamp upTime = new Timestamp(hiResTimer.milliTime(), (short)1);
                                             // Uptime is a timestamp representing the time when the
                                             // current instantiation of the Sender started operating, in
                                             // milliseconds since 1Jan1970.

      while(true){
        Socket client = senderSocket.accept();                                // accept the client socket connection
        OWAMP_SessionSender sender = new OWAMP_SessionSender(client, upTime, portRange, hiResTimer); // deliver this socket to the sender

        sender.mode = 1;                                                      // Mode of authentication.
        sender.start();                                                       // initialize thread

      }
    }
    catch (SocketException se){
        System.err.println(se);
    }
  }

}
